'use strict';

var anchor = require('@project-serum/anchor');
var token = require('@project-serum/anchor/dist/cjs/spl/token');
var serum = require('@project-serum/serum');
var client = require('@pythnetwork/client');
var web3_js = require('@solana/web3.js');
var BN = require('bn.js');
var lodash = require('lodash');
var Big = require('big.js');
var splToken = require('@solana/spl-token');
var bytes = require('@project-serum/anchor/dist/cjs/utils/bytes');
var SwitchboardProgram = require('@switchboard-xyz/sbv2-lite');
var tokenInstructions = require('@project-serum/serum/lib/token-instructions');
var bs58 = require('bs58');
var market = require('@project-serum/serum/lib/market');
var fetch = require('cross-fetch');
var NodeWallet = require('@project-serum/anchor/dist/cjs/nodewallet');

const RUST_U64_MAX = () => {
    return new anchor.BN('18446744073709551615');
};
const RUST_I64_MAX = () => {
    return new anchor.BN('9223372036854775807');
};
const RUST_I64_MIN = () => {
    return new anchor.BN('-9223372036854775807');
};
const OPENBOOK_PROGRAM_ID = {
    devnet: new web3_js.PublicKey('EoTcMgcDRTJVZDMZWBoU6rhYHZfkNTVEAfz3uUJRcYGj'),
    'mainnet-beta': new web3_js.PublicKey('srmqPvymJeFKQ4zGQed1GFppgkRHL9kaELCbyksJtPX'),
};
const MANGO_V4_ID = {
    testnet: new web3_js.PublicKey('4MangoMjqJ2firMokCjjGgoK8d4MXcrgL7XJaL3w6fVg'),
    devnet: new web3_js.PublicKey('4MangoMjqJ2firMokCjjGgoK8d4MXcrgL7XJaL3w6fVg'),
    'mainnet-beta': new web3_js.PublicKey('4MangoMjqJ2firMokCjjGgoK8d4MXcrgL7XJaL3w6fVg'),
};

class I80F48Dto {
    val;
    constructor(val) {
        this.val = val;
    }
}
// TODO - this whole class is inefficient; consider optimizing
class I80F48 {
    /**
    This is represented by a 128 bit signed integer underneath
    The first 80 bits are treated as an integer and last 48 bits are treated as fractional part after binary point
    It's possible to think of an I80F48 as an i128 divided by 2 ^ 40
  
    Read up on how fixed point math works: https://inst.eecs.berkeley.edu/~cs61c/sp06/handout/fixedpt.html
    Read up on how 2s complement works: https://en.wikipedia.org/wiki/Two%27s_complement
     */
    static MAX_SIZE = 128;
    static FRACTIONS = 48;
    static MULTIPLIER_BIG = new Big(2).pow(I80F48.FRACTIONS);
    static MULTIPLIER_BN = new BN(2).pow(new BN(I80F48.FRACTIONS));
    static MULTIPLIER_NUMBER = Math.pow(2, I80F48.FRACTIONS);
    static MAX_BN = new BN(2)
        .pow(new BN(I80F48.MAX_SIZE))
        .div(new BN(2))
        .sub(new BN(1));
    static MIN_BN = new BN(2)
        .pow(new BN(I80F48.MAX_SIZE))
        .div(new BN(2))
        .neg();
    data; // This is i128 => array of 16 bytes
    static from(dto) {
        return new I80F48(dto.val);
    }
    constructor(data) {
        if (data.lt(I80F48.MIN_BN) || data.gt(I80F48.MAX_BN)) {
            throw new Error('Number out of range');
        }
        this.data = data;
    }
    static fromNumber(x) {
        const int_part = Math.trunc(x);
        const v = new BN(int_part.toFixed(0)).iushln(48);
        v.iadd(new BN((x - int_part) * I80F48.MULTIPLIER_NUMBER));
        return new I80F48(v);
    }
    static fromNumberOrUndef(x) {
        return x === undefined ? undefined : I80F48.fromNumber(x);
    }
    static fromOptionalString(x) {
        return x ? I80F48.fromString(x) : undefined;
    }
    static fromString(x) {
        const initialValue = new Big(x).times(I80F48.MULTIPLIER_BIG);
        const fixedPointValue = new BN(initialValue.round().toFixed());
        return new I80F48(fixedPointValue);
    }
    static fromI64(x) {
        return new I80F48(x.ushln(48));
    }
    static fromU64(x) {
        return new I80F48(x.ushln(48));
    }
    toTwos() {
        return this.data.toTwos(I80F48.MAX_SIZE);
    }
    toString() {
        return this.toBig().toFixed();
    }
    /**
     * The number will be rounded first for UI sensibilities, then toFixed
     */
    toFixed(decimals) {
        return this.toBig().round(14).toFixed(decimals);
    }
    toLocaleString(locales, options) {
        return this.toNumber().toLocaleString(locales, options);
    }
    toBig() {
        return new Big(this.data.toString()).div(I80F48.MULTIPLIER_BIG);
    }
    static fromBig(x) {
        return new I80F48(new BN(x.mul(I80F48.MULTIPLIER_BIG).round().toFixed()));
    }
    toNumber() {
        return this.toBig().toNumber();
    }
    static fromArray(src) {
        if (src.length !== 16) {
            throw new Error('Uint8Array must be of length 16');
        }
        return new I80F48(new BN(src, 'le').fromTwos(I80F48.MAX_SIZE));
    }
    toArray() {
        return new Uint8Array(this.data.toTwos(I80F48.MAX_SIZE).toArray('le', 16));
    }
    toArrayLike(ArrayType, endian, length) {
        return this.data
            .toTwos(I80F48.MAX_SIZE)
            .toArrayLike(ArrayType, endian, length);
    }
    getData() {
        return this.data;
    }
    getBinaryLayout() {
        return this.data
            .toTwos(I80F48.MAX_SIZE)
            .toString(2, I80F48.MAX_SIZE)
            .replace(/-/g, '');
    }
    add(x) {
        return new I80F48(this.data.add(x.getData()));
    }
    sub(x) {
        return new I80F48(this.data.sub(x.getData()));
    }
    iadd(x) {
        this.data.iadd(x.getData());
        return this;
    }
    isub(x) {
        this.data.isub(x.getData());
        return this;
    }
    floor() {
        // Low IQ method
        return I80F48.fromBig(this.toBig().round(undefined, 0));
        // return new I80F48(this.data.shrn(I80F48.FRACTIONS).shln(I80F48.FRACTIONS));
    }
    ceil() {
        // Low IQ method, 3 -> round up
        return I80F48.fromBig(this.toBig().round(undefined, 3));
        // const frac = this.data.maskn(I80F48.FRACTIONS);
        // if (frac.eq(ZERO_BN)) {
        //   return this;
        // } else {
        //   return this.floor().add(ONE_I80F48);
        // }
    }
    frac() {
        // TODO verify this works for negative numbers
        return new I80F48(this.data.maskn(I80F48.FRACTIONS));
    }
    /**
     * Multiply the two and shift
     */
    mul(x) {
        return new I80F48(this.data.mul(x.data).iushrn(I80F48.FRACTIONS));
    }
    imul(x) {
        this.data.imul(x.getData()).iushrn(I80F48.FRACTIONS);
        return this;
    }
    div(x) {
        return new I80F48(this.data.ushln(I80F48.FRACTIONS).div(x.data));
    }
    idiv(x) {
        this.data = this.data.iushln(I80F48.FRACTIONS).div(x.data);
        return this;
    }
    gt(x) {
        return this.data.gt(x.getData());
    }
    lt(x) {
        return this.data.lt(x.getData());
    }
    gte(x) {
        return this.data.gte(x.getData());
    }
    lte(x) {
        return this.data.lte(x.getData());
    }
    eq(x) {
        // TODO make sure this works when they're diff signs or 0
        return this.data.eq(x.getData());
    }
    cmp(x) {
        // TODO make sure this works when they're diff signs or 0
        return this.data.cmp(x.getData());
    }
    neg() {
        return this.mul(_NEG_ONE_I80F48);
    }
    isPos() {
        return this.gt(_ZERO_I80F48);
    }
    isNeg() {
        return this.data.isNeg();
    }
    isZero() {
        return this.eq(_ZERO_I80F48);
    }
    min(x) {
        return this.lte(x) ? this : x;
    }
    max(x) {
        return this.gte(x) ? this : x;
    }
    abs() {
        if (this.isNeg()) {
            return this.neg();
        }
        else {
            return this;
        }
    }
}
/** @internal */
const _ZERO_I80F48 = I80F48.fromNumber(0);
/** @internal */
const _NEG_ONE_I80F48 = I80F48.fromNumber(-1);
function ONE_I80F48() {
    return I80F48.fromNumber(1);
}
function ZERO_I80F48() {
    return I80F48.fromNumber(0);
}
function HUNDRED_I80F48() {
    return I80F48.fromNumber(100);
}
function MAX_I80F48() {
    return new I80F48(I80F48.MAX_BN);
}

///
/// numeric helpers
///
const U64_MAX_BN = new BN('18446744073709551615');
const I64_MAX_BN = new BN('9223372036854775807').toTwos(64);
function toNativeI80F48ForQuote(uiAmount) {
    return I80F48.fromNumber(uiAmount * Math.pow(10, 6));
}
function toNativeI80F48(uiAmount, decimals) {
    return I80F48.fromNumber(uiAmount * Math.pow(10, decimals));
}
function toNative(uiAmount, decimals) {
    return new BN((uiAmount * Math.pow(10, decimals)).toFixed(0));
}
function toUiDecimals(nativeAmount, decimals) {
    // TODO: remove BN and upgrade to bigint https://github.com/solana-labs/solana/issues/27440
    if (nativeAmount instanceof BN) {
        nativeAmount = I80F48.fromU64(nativeAmount);
    }
    if (nativeAmount instanceof I80F48) {
        return nativeAmount
            .div(I80F48.fromNumber(Math.pow(10, decimals)))
            .toNumber();
    }
    return nativeAmount / Math.pow(10, decimals);
}
const QUOTE_DECIMALS = 6;
function toUiDecimalsForQuote(nativeAmount) {
    return toUiDecimals(nativeAmount, QUOTE_DECIMALS);
}
function toUiI80F48(nativeAmount, decimals) {
    return nativeAmount.div(I80F48.fromNumber(Math.pow(10, decimals)));
}
///
/// web3js extensions
///
/**
 * Get the address of the associated token account for a given mint and owner
 *
 * @param mint                     Token mint account
 * @param owner                    Owner of the new account
 * @param allowOwnerOffCurve       Allow the owner account to be a PDA (Program Derived Address)
 * @param programId                SPL Token program account
 * @param associatedTokenProgramId SPL Associated Token program account
 *
 * @return Address of the associated token account
 */
async function getAssociatedTokenAddress(mint, owner, allowOwnerOffCurve = false, programId = splToken.TOKEN_PROGRAM_ID, associatedTokenProgramId = splToken.ASSOCIATED_TOKEN_PROGRAM_ID) {
    if (!allowOwnerOffCurve && !web3_js.PublicKey.isOnCurve(owner.toBuffer()))
        throw new Error('TokenOwnerOffCurve!');
    const [address] = await web3_js.PublicKey.findProgramAddress([owner.toBuffer(), programId.toBuffer(), mint.toBuffer()], associatedTokenProgramId);
    return address;
}
async function createAssociatedTokenAccountIdempotentInstruction(payer, owner, mint) {
    const account = await getAssociatedTokenAddress(mint, owner);
    return new web3_js.TransactionInstruction({
        keys: [
            { pubkey: payer, isSigner: true, isWritable: true },
            { pubkey: account, isSigner: false, isWritable: true },
            { pubkey: owner, isSigner: false, isWritable: false },
            { pubkey: mint, isSigner: false, isWritable: false },
            {
                pubkey: web3_js.SystemProgram.programId,
                isSigner: false,
                isWritable: false,
            },
            { pubkey: splToken.TOKEN_PROGRAM_ID, isSigner: false, isWritable: false },
        ],
        programId: splToken.ASSOCIATED_TOKEN_PROGRAM_ID,
        data: Buffer.from([0x1]),
    });
}
async function buildVersionedTx(provider, ix, additionalSigners = [], alts = []) {
    const message = web3_js.MessageV0.compile({
        payerKey: provider.wallet.publicKey,
        instructions: ix,
        recentBlockhash: (await provider.connection.getLatestBlockhash()).blockhash,
        addressLookupTableAccounts: alts,
    });
    const vTx = new web3_js.VersionedTransaction(message);
    vTx.sign([
        provider.wallet.payer,
        ...additionalSigners,
    ]);
    return vTx;
}

class Bank {
    publicKey;
    group;
    mint;
    vault;
    oracle;
    stablePriceModel;
    indexLastUpdated;
    bankRateLastUpdated;
    tokenIndex;
    mintDecimals;
    bankNum;
    minVaultToDepositsRatio;
    netBorrowLimitWindowSizeTs;
    lastNetBorrowsWindowStartTs;
    netBorrowLimitPerWindowQuote;
    netBorrowsInWindow;
    borrowWeightScaleStartQuote;
    depositWeightScaleStartQuote;
    reduceOnly;
    name;
    oracleConfig;
    depositIndex;
    borrowIndex;
    indexedDeposits;
    indexedBorrows;
    avgUtilization;
    adjustmentFactor;
    maxRate;
    rate0;
    rate1;
    util0;
    util1;
    _price;
    _uiPrice;
    _oracleLastUpdatedSlot;
    collectedFeesNative;
    loanFeeRate;
    loanOriginationFeeRate;
    initAssetWeight;
    maintAssetWeight;
    initLiabWeight;
    maintLiabWeight;
    liquidationFee;
    dust;
    static from(publicKey, obj) {
        return new Bank(publicKey, obj.group, obj.name, obj.mint, obj.vault, obj.oracle, obj.oracleConfig, obj.stablePriceModel, obj.depositIndex, obj.borrowIndex, obj.indexedDeposits, obj.indexedBorrows, obj.indexLastUpdated, obj.bankRateLastUpdated, obj.avgUtilization, obj.adjustmentFactor, obj.util0, obj.rate0, obj.util1, obj.rate1, obj.maxRate, obj.collectedFeesNative, obj.loanOriginationFeeRate, obj.loanFeeRate, obj.maintAssetWeight, obj.initAssetWeight, obj.maintLiabWeight, obj.initLiabWeight, obj.liquidationFee, obj.dust, obj.flashLoanTokenAccountInitial, obj.flashLoanApprovedAmount, obj.tokenIndex, obj.mintDecimals, obj.bankNum, obj.minVaultToDepositsRatio, obj.netBorrowLimitWindowSizeTs, obj.lastNetBorrowsWindowStartTs, obj.netBorrowLimitPerWindowQuote, obj.netBorrowsInWindow, obj.borrowWeightScaleStartQuote, obj.depositWeightScaleStartQuote, obj.reduceOnly == 1);
    }
    constructor(publicKey, group, name, mint, vault, oracle, oracleConfig, stablePriceModel, depositIndex, borrowIndex, indexedDeposits, indexedBorrows, indexLastUpdated, bankRateLastUpdated, avgUtilization, adjustmentFactor, util0, rate0, util1, rate1, maxRate, collectedFeesNative, loanOriginationFeeRate, loanFeeRate, maintAssetWeight, initAssetWeight, maintLiabWeight, initLiabWeight, liquidationFee, dust, flashLoanTokenAccountInitial, flashLoanApprovedAmount, tokenIndex, mintDecimals, bankNum, minVaultToDepositsRatio, netBorrowLimitWindowSizeTs, lastNetBorrowsWindowStartTs, netBorrowLimitPerWindowQuote, netBorrowsInWindow, borrowWeightScaleStartQuote, depositWeightScaleStartQuote, reduceOnly) {
        this.publicKey = publicKey;
        this.group = group;
        this.mint = mint;
        this.vault = vault;
        this.oracle = oracle;
        this.stablePriceModel = stablePriceModel;
        this.indexLastUpdated = indexLastUpdated;
        this.bankRateLastUpdated = bankRateLastUpdated;
        this.tokenIndex = tokenIndex;
        this.mintDecimals = mintDecimals;
        this.bankNum = bankNum;
        this.minVaultToDepositsRatio = minVaultToDepositsRatio;
        this.netBorrowLimitWindowSizeTs = netBorrowLimitWindowSizeTs;
        this.lastNetBorrowsWindowStartTs = lastNetBorrowsWindowStartTs;
        this.netBorrowLimitPerWindowQuote = netBorrowLimitPerWindowQuote;
        this.netBorrowsInWindow = netBorrowsInWindow;
        this.borrowWeightScaleStartQuote = borrowWeightScaleStartQuote;
        this.depositWeightScaleStartQuote = depositWeightScaleStartQuote;
        this.reduceOnly = reduceOnly;
        this.name = bytes.utf8.decode(new Uint8Array(name)).split('\x00')[0];
        this.oracleConfig = {
            confFilter: I80F48.from(oracleConfig.confFilter),
            maxStalenessSlots: oracleConfig.maxStalenessSlots,
        };
        this.depositIndex = I80F48.from(depositIndex);
        this.borrowIndex = I80F48.from(borrowIndex);
        this.indexedDeposits = I80F48.from(indexedDeposits);
        this.indexedBorrows = I80F48.from(indexedBorrows);
        this.avgUtilization = I80F48.from(avgUtilization);
        this.adjustmentFactor = I80F48.from(adjustmentFactor);
        this.maxRate = I80F48.from(maxRate);
        this.util0 = I80F48.from(util0);
        this.rate0 = I80F48.from(rate0);
        this.util1 = I80F48.from(util1);
        this.rate1 = I80F48.from(rate1);
        this.collectedFeesNative = I80F48.from(collectedFeesNative);
        this.loanFeeRate = I80F48.from(loanFeeRate);
        this.loanOriginationFeeRate = I80F48.from(loanOriginationFeeRate);
        this.maintAssetWeight = I80F48.from(maintAssetWeight);
        this.initAssetWeight = I80F48.from(initAssetWeight);
        this.maintLiabWeight = I80F48.from(maintLiabWeight);
        this.initLiabWeight = I80F48.from(initLiabWeight);
        this.liquidationFee = I80F48.from(liquidationFee);
        this.dust = I80F48.from(dust);
        this._price = undefined;
        this._uiPrice = undefined;
        this._oracleLastUpdatedSlot = undefined;
    }
    toString() {
        return ('Bank ' +
            '\n public key - ' +
            this.publicKey.toBase58() +
            '\n token index - ' +
            this.tokenIndex +
            '\n token name - ' +
            this.name +
            '\n vault - ' +
            this.vault.toBase58() +
            '\n mintDecimals - ' +
            this.mintDecimals +
            '\n oracle - ' +
            this.oracle.toBase58() +
            '\n price - ' +
            this._price?.toString() +
            '\n uiPrice - ' +
            this._uiPrice +
            '\n deposit index - ' +
            this.depositIndex.toString() +
            '\n borrow index - ' +
            this.borrowIndex.toString() +
            '\n indexedDeposits - ' +
            this.indexedDeposits.toString() +
            '\n indexedBorrows - ' +
            this.indexedBorrows.toString() +
            '\n indexLastUpdated - ' +
            new Date(this.indexLastUpdated.toNumber() * 1000) +
            '\n bankRateLastUpdated - ' +
            new Date(this.bankRateLastUpdated.toNumber() * 1000) +
            '\n avgUtilization - ' +
            this.avgUtilization.toString() +
            '\n adjustmentFactor - ' +
            this.adjustmentFactor.toString() +
            '\n maxRate - ' +
            this.maxRate.toString() +
            '\n util0 - ' +
            this.util0.toString() +
            '\n rate0 - ' +
            this.rate0.toString() +
            '\n util1 - ' +
            this.util1.toString() +
            '\n rate1 - ' +
            this.rate1.toString() +
            '\n loanFeeRate - ' +
            this.loanFeeRate.toString() +
            '\n loanOriginationFeeRate - ' +
            this.loanOriginationFeeRate.toString() +
            '\n maintAssetWeight - ' +
            this.maintAssetWeight.toString() +
            '\n initAssetWeight - ' +
            this.initAssetWeight.toString() +
            '\n maintLiabWeight - ' +
            this.maintLiabWeight.toString() +
            '\n initLiabWeight - ' +
            this.initLiabWeight.toString() +
            '\n liquidationFee - ' +
            this.liquidationFee.toString() +
            '\n uiDeposits() - ' +
            this.uiDeposits() +
            '\n uiBorrows() - ' +
            this.uiBorrows() +
            '\n getDepositRate() - ' +
            this.getDepositRate().toString() +
            '\n getBorrowRate() - ' +
            this.getBorrowRate().toString());
    }
    scaledInitAssetWeight(price) {
        const depositsQuote = this.nativeDeposits().mul(price);
        if (this.depositWeightScaleStartQuote >= Number.MAX_SAFE_INTEGER ||
            depositsQuote.lte(I80F48.fromNumber(this.depositWeightScaleStartQuote))) {
            return this.initAssetWeight;
        }
        return this.initAssetWeight.mul(I80F48.fromNumber(this.depositWeightScaleStartQuote).div(depositsQuote));
    }
    scaledInitLiabWeight(price) {
        const borrowsQuote = this.nativeBorrows().mul(price);
        if (this.borrowWeightScaleStartQuote >= Number.MAX_SAFE_INTEGER ||
            borrowsQuote.lte(I80F48.fromNumber(this.borrowWeightScaleStartQuote))) {
            return this.initLiabWeight;
        }
        return this.initLiabWeight.mul(borrowsQuote.div(I80F48.fromNumber(this.borrowWeightScaleStartQuote)));
    }
    get price() {
        if (!this._price) {
            throw new Error(`Undefined price for bank ${this.publicKey} with tokenIndex ${this.tokenIndex}!`);
        }
        return this._price;
    }
    get uiPrice() {
        if (!this._uiPrice) {
            throw new Error(`Undefined uiPrice for bank ${this.publicKey} with tokenIndex ${this.tokenIndex}!`);
        }
        return this._uiPrice;
    }
    get oracleLastUpdatedSlot() {
        if (!this._oracleLastUpdatedSlot) {
            throw new Error(`Undefined oracleLastUpdatedSlot for bank ${this.publicKey} with tokenIndex ${this.tokenIndex}!`);
        }
        return this._oracleLastUpdatedSlot;
    }
    nativeDeposits() {
        return this.indexedDeposits.mul(this.depositIndex);
    }
    nativeBorrows() {
        return this.indexedBorrows.mul(this.borrowIndex);
    }
    uiDeposits() {
        return toUiDecimals(this.indexedDeposits.mul(this.depositIndex), this.mintDecimals);
    }
    uiBorrows() {
        return toUiDecimals(this.indexedBorrows.mul(this.borrowIndex), this.mintDecimals);
    }
    /**
     *
     * @returns borrow rate, 0 is 0% where 1 is 100%
     */
    getBorrowRate() {
        const totalBorrows = this.nativeBorrows();
        const totalDeposits = this.nativeDeposits();
        if (totalDeposits.isZero() && totalBorrows.isZero()) {
            return ZERO_I80F48();
        }
        if (totalDeposits.lte(totalBorrows)) {
            return this.maxRate;
        }
        const utilization = totalBorrows.div(totalDeposits);
        if (utilization.gt(this.util1)) {
            const extraUtil = utilization.sub(this.util1);
            const slope = this.maxRate
                .sub(this.rate1)
                .div(I80F48.fromNumber(1).sub(this.util1));
            return this.rate1.add(slope.mul(extraUtil));
        }
        else if (utilization.gt(this.util0)) {
            const extraUtil = utilization.sub(this.util0);
            const slope = this.maxRate
                .sub(this.rate0)
                .div(I80F48.fromNumber(1).sub(this.util0));
            return this.rate0.add(slope.mul(extraUtil));
        }
        else {
            const slope = this.rate0.div(this.util0);
            return slope.mul(utilization);
        }
    }
    /**
     *
     * @returns borrow rate percentage
     */
    getBorrowRateUi() {
        return this.getBorrowRate().toNumber() * 100;
    }
    /**
     *
     * @returns deposit rate, 0 is 0% where 1 is 100%
     */
    getDepositRate() {
        const borrowRate = this.getBorrowRate();
        const totalBorrows = this.nativeBorrows();
        const totalDeposits = this.nativeDeposits();
        if (totalDeposits.isZero() && totalBorrows.isZero()) {
            return ZERO_I80F48();
        }
        else if (totalDeposits.isZero()) {
            return this.maxRate;
        }
        const utilization = totalBorrows.div(totalDeposits);
        return utilization.mul(borrowRate);
    }
    /**
     *
     * @returns deposit rate percentage
     */
    getDepositRateUi() {
        return this.getDepositRate().toNumber() * 100;
    }
}
class MintInfo {
    publicKey;
    group;
    tokenIndex;
    mint;
    banks;
    vaults;
    oracle;
    registrationTime;
    groupInsuranceFund;
    static from(publicKey, obj) {
        return new MintInfo(publicKey, obj.group, obj.tokenIndex, obj.mint, obj.banks, obj.vaults, obj.oracle, obj.registrationTime, obj.groupInsuranceFund);
    }
    constructor(publicKey, group, tokenIndex, mint, banks, vaults, oracle, registrationTime, groupInsuranceFund) {
        this.publicKey = publicKey;
        this.group = group;
        this.tokenIndex = tokenIndex;
        this.mint = mint;
        this.banks = banks;
        this.vaults = vaults;
        this.oracle = oracle;
        this.registrationTime = registrationTime;
        this.groupInsuranceFund = groupInsuranceFund;
    }
    firstBank() {
        return this.banks[0];
    }
    firstVault() {
        return this.vaults[0];
    }
    toString() {
        const res = 'mint ' +
            this.mint.toBase58() +
            '\n oracle ' +
            this.oracle.toBase58() +
            '\n banks ' +
            this.banks
                .filter((pk) => pk.toBase58() !== web3_js.PublicKey.default.toBase58())
                .toString() +
            '\n vaults ' +
            this.vaults
                .filter((pk) => pk.toBase58() !== web3_js.PublicKey.default.toBase58())
                .toString();
        return res;
    }
}

const SBV1_DEVNET_PID = new web3_js.PublicKey('7azgmy1pFXHikv36q1zZASvFq5vFa39TT9NweVugKKTU');
const SBV1_MAINNET_PID = new web3_js.PublicKey('DtmE9D2CSB4L5D6A15mraeEjrGMm6auWVzgaD8hK2tZM');
let sbv2DevnetProgram;
let sbv2MainnetProgram;
class StubOracle {
    publicKey;
    group;
    mint;
    price;
    lastUpdated;
    static from(publicKey, obj) {
        return new StubOracle(publicKey, obj.group, obj.mint, obj.price, obj.lastUpdated);
    }
    constructor(publicKey, group, mint, price, lastUpdated) {
        this.publicKey = publicKey;
        this.group = group;
        this.mint = mint;
        this.price = I80F48.from(price);
        this.lastUpdated = lastUpdated;
    }
}
// https://gist.github.com/microwavedcola1/b741a11e6ee273a859f3ef00b35ac1f0
function parseSwitchboardOracleV1(accountInfo) {
    const price = accountInfo.data.readDoubleLE(1 + 32 + 4 + 4);
    const lastUpdatedSlot = parseInt(accountInfo.data.readBigUInt64LE(1 + 32 + 4 + 4 + 8).toString());
    return { price, lastUpdatedSlot };
}
function parseSwitchboardOracleV2(program, accountInfo) {
    const price = program.decodeLatestAggregatorValue(accountInfo).toNumber();
    const lastUpdatedSlot = program
        .decodeAggregator(accountInfo)
        .latestConfirmedRound.roundOpenSlot.toNumber();
    if (!price || !lastUpdatedSlot)
        throw new Error('Unable to parse Switchboard Oracle V2');
    return { price, lastUpdatedSlot };
}
/**
 *
 * @param accountInfo
 * @returns ui price
 */
async function parseSwitchboardOracle(accountInfo, connection) {
    if (accountInfo.owner.equals(SwitchboardProgram.devnetPid)) {
        if (!sbv2DevnetProgram) {
            sbv2DevnetProgram = await SwitchboardProgram.loadDevnet(connection);
        }
        return parseSwitchboardOracleV2(sbv2DevnetProgram, accountInfo);
    }
    if (accountInfo.owner.equals(SwitchboardProgram.mainnetPid)) {
        if (!sbv2MainnetProgram) {
            sbv2MainnetProgram = await SwitchboardProgram.loadMainnet(connection);
        }
        return parseSwitchboardOracleV2(sbv2MainnetProgram, accountInfo);
    }
    if (accountInfo.owner.equals(SBV1_DEVNET_PID) ||
        accountInfo.owner.equals(SBV1_MAINNET_PID)) {
        return parseSwitchboardOracleV1(accountInfo);
    }
    throw new Error(`Should not be reached!`);
}
function isSwitchboardOracle(accountInfo) {
    if (accountInfo.owner.equals(SBV1_DEVNET_PID) ||
        accountInfo.owner.equals(SBV1_MAINNET_PID) ||
        accountInfo.owner.equals(SwitchboardProgram.devnetPid) ||
        accountInfo.owner.equals(SwitchboardProgram.mainnetPid)) {
        return true;
    }
    return false;
}
function isPythOracle(accountInfo) {
    return accountInfo.data.readUInt32LE(0) === client.Magic;
}

class PerpMarket {
    publicKey;
    group;
    settleTokenIndex;
    perpMarketIndex;
    groupInsuranceFund;
    baseDecimals;
    bids;
    asks;
    eventQueue;
    oracle;
    stablePriceModel;
    quoteLotSize;
    baseLotSize;
    openInterest;
    seqNum;
    registrationTime;
    impactQuantity;
    fundingLastUpdated;
    feePenalty;
    settleFeeFlat;
    settleFeeAmountThreshold;
    settleFeeFractionLowHealth;
    settlePnlLimitFactor;
    settlePnlLimitWindowSizeTs;
    reduceOnly;
    name;
    oracleConfig;
    maintBaseAssetWeight;
    initBaseAssetWeight;
    maintBaseLiabWeight;
    initBaseLiabWeight;
    baseLiquidationFee;
    makerFee;
    takerFee;
    minFunding;
    maxFunding;
    longFunding;
    shortFunding;
    feesAccrued;
    feesSettled;
    maintOverallAssetWeight;
    initOverallAssetWeight;
    _price;
    _uiPrice;
    _oracleLastUpdatedSlot;
    priceLotsToUiConverter;
    baseLotsToUiConverter;
    quoteLotsToUiConverter;
    _bids;
    _asks;
    static from(publicKey, obj) {
        return new PerpMarket(publicKey, obj.group, obj.settleTokenIndex, obj.perpMarketIndex, obj.groupInsuranceFund == 1, obj.baseDecimals, obj.name, obj.bids, obj.asks, obj.eventQueue, obj.oracle, obj.oracleConfig, obj.stablePriceModel, obj.quoteLotSize, obj.baseLotSize, obj.maintBaseAssetWeight, obj.initBaseAssetWeight, obj.maintBaseLiabWeight, obj.initBaseLiabWeight, obj.openInterest, obj.seqNum, obj.registrationTime, obj.minFunding, obj.maxFunding, obj.impactQuantity, obj.longFunding, obj.shortFunding, obj.fundingLastUpdated, obj.baseLiquidationFee, obj.makerFee, obj.takerFee, obj.feesAccrued, obj.feesSettled, obj.feePenalty, obj.settleFeeFlat, obj.settleFeeAmountThreshold, obj.settleFeeFractionLowHealth, obj.settlePnlLimitFactor, obj.settlePnlLimitWindowSizeTs, obj.reduceOnly == 1, obj.maintOverallAssetWeight, obj.initOverallAssetWeight, obj.positivePnlLiquidationFee);
    }
    constructor(publicKey, group, settleTokenIndex, perpMarketIndex, // TODO rename to marketIndex?
    groupInsuranceFund, baseDecimals, name, bids, asks, eventQueue, oracle, oracleConfig, stablePriceModel, quoteLotSize, baseLotSize, maintBaseAssetWeight, initBaseAssetWeight, maintBaseLiabWeight, initBaseLiabWeight, openInterest, seqNum, registrationTime, minFunding, maxFunding, impactQuantity, longFunding, shortFunding, fundingLastUpdated, baseLiquidationFee, makerFee, takerFee, feesAccrued, feesSettled, feePenalty, settleFeeFlat, settleFeeAmountThreshold, settleFeeFractionLowHealth, settlePnlLimitFactor, settlePnlLimitWindowSizeTs, reduceOnly, maintOverallAssetWeight, initOverallAssetWeight, positivePnlLiquidationFee) {
        this.publicKey = publicKey;
        this.group = group;
        this.settleTokenIndex = settleTokenIndex;
        this.perpMarketIndex = perpMarketIndex;
        this.groupInsuranceFund = groupInsuranceFund;
        this.baseDecimals = baseDecimals;
        this.bids = bids;
        this.asks = asks;
        this.eventQueue = eventQueue;
        this.oracle = oracle;
        this.stablePriceModel = stablePriceModel;
        this.quoteLotSize = quoteLotSize;
        this.baseLotSize = baseLotSize;
        this.openInterest = openInterest;
        this.seqNum = seqNum;
        this.registrationTime = registrationTime;
        this.impactQuantity = impactQuantity;
        this.fundingLastUpdated = fundingLastUpdated;
        this.feePenalty = feePenalty;
        this.settleFeeFlat = settleFeeFlat;
        this.settleFeeAmountThreshold = settleFeeAmountThreshold;
        this.settleFeeFractionLowHealth = settleFeeFractionLowHealth;
        this.settlePnlLimitFactor = settlePnlLimitFactor;
        this.settlePnlLimitWindowSizeTs = settlePnlLimitWindowSizeTs;
        this.reduceOnly = reduceOnly;
        this.name = bytes.utf8.decode(new Uint8Array(name)).split('\x00')[0];
        this.oracleConfig = {
            confFilter: I80F48.from(oracleConfig.confFilter),
            maxStalenessSlots: oracleConfig.maxStalenessSlots,
        };
        this.maintBaseAssetWeight = I80F48.from(maintBaseAssetWeight);
        this.initBaseAssetWeight = I80F48.from(initBaseAssetWeight);
        this.maintBaseLiabWeight = I80F48.from(maintBaseLiabWeight);
        this.initBaseLiabWeight = I80F48.from(initBaseLiabWeight);
        this.baseLiquidationFee = I80F48.from(baseLiquidationFee);
        this.makerFee = I80F48.from(makerFee);
        this.takerFee = I80F48.from(takerFee);
        this.minFunding = I80F48.from(minFunding);
        this.maxFunding = I80F48.from(maxFunding);
        this.longFunding = I80F48.from(longFunding);
        this.shortFunding = I80F48.from(shortFunding);
        this.feesAccrued = I80F48.from(feesAccrued);
        this.feesSettled = I80F48.from(feesSettled);
        this.maintOverallAssetWeight = I80F48.from(maintOverallAssetWeight);
        this.initOverallAssetWeight = I80F48.from(initOverallAssetWeight);
        this.priceLotsToUiConverter = new Big(10)
            .pow(baseDecimals - QUOTE_DECIMALS)
            .mul(new Big(this.quoteLotSize.toString()))
            .div(new Big(this.baseLotSize.toString()))
            .toNumber();
        this.baseLotsToUiConverter = new Big(this.baseLotSize.toString())
            .div(new Big(10).pow(baseDecimals))
            .toNumber();
        this.quoteLotsToUiConverter = new Big(this.quoteLotSize.toString())
            .div(new Big(10).pow(QUOTE_DECIMALS))
            .toNumber();
    }
    get price() {
        if (!this._price) {
            throw new Error(`Undefined price for perpMarket ${this.publicKey} with marketIndex ${this.perpMarketIndex}!`);
        }
        return this._price;
    }
    get uiPrice() {
        if (!this._uiPrice) {
            throw new Error(`Undefined price for perpMarket ${this.publicKey} with marketIndex ${this.perpMarketIndex}!`);
        }
        return this._uiPrice;
    }
    get oracleLastUpdatedSlot() {
        if (!this._oracleLastUpdatedSlot) {
            throw new Error(`Undefined oracleLastUpdatedSlot for perpMarket ${this.publicKey} with marketIndex ${this.perpMarketIndex}!`);
        }
        return this._oracleLastUpdatedSlot;
    }
    get minOrderSize() {
        return this.baseLotsToUiConverter;
    }
    get tickSize() {
        return this.priceLotsToUiConverter;
    }
    insidePriceLimit(side, orderPrice) {
        return ((side === PerpOrderSide.bid &&
            orderPrice <= this.maintBaseLiabWeight.toNumber() * this.uiPrice) ||
            (side === PerpOrderSide.ask &&
                orderPrice >= this.maintBaseAssetWeight.toNumber() * this.uiPrice));
    }
    async loadAsks(client, forceReload = false) {
        if (forceReload || !this._asks) {
            const asks = await client.program.account.bookSide.fetch(this.asks);
            this._asks = BookSide.from(client, this, BookSideType.asks, asks);
        }
        return this._asks;
    }
    async loadBids(client, forceReload = false) {
        if (forceReload || !this._bids) {
            const bids = await client.program.account.bookSide.fetch(this.bids);
            this._bids = BookSide.from(client, this, BookSideType.bids, bids);
        }
        return this._bids;
    }
    async loadEventQueue(client) {
        const eventQueue = await client.program.account.eventQueue.fetch(this.eventQueue);
        return new PerpEventQueue(client, eventQueue.header, eventQueue.buf);
    }
    async loadFills(client, lastSeqNum = new anchor.BN(0)) {
        const eventQueue = await this.loadEventQueue(client);
        return eventQueue
            .eventsSince(lastSeqNum)
            .filter((event) => event.eventType == PerpEventQueue.FILL_EVENT_TYPE)
            .map(this.parseFillEvent.bind(this));
    }
    parseFillEvent(event) {
        const quantity = this.baseLotsToUi(event.quantity);
        const price = this.priceLotsToUi(event.price);
        return {
            ...event,
            quantity,
            size: quantity,
            price,
        };
    }
    async logOb(client) {
        let res = ``;
        res += `  ${this.name} OrderBook`;
        let orders = await this?.loadAsks(client);
        for (const order of orders.items()) {
            res += `\n ${order.uiPrice.toFixed(5).padStart(10)}, ${order.uiSize
                .toString()
                .padStart(10)} ${order.isOraclePegged && order.oraclePeggedProperties
                ? order.oraclePeggedProperties.pegLimit.toNumber() + ' (PegLimit)'
                : ''}`;
        }
        res += `\n  asks ↑ --------- ↓ bids`;
        orders = await this?.loadBids(client);
        for (const order of orders.items()) {
            res += `\n  ${order.uiPrice.toFixed(5).padStart(10)}, ${order.uiSize
                .toString()
                .padStart(10)} ${order.isOraclePegged && order.oraclePeggedProperties
                ? order.oraclePeggedProperties.pegLimit.toNumber() + ' (PegLimit)'
                : ''}`;
        }
        return res;
    }
    /**
     *
     * @param bids
     * @param asks
     * @returns returns funding rate per hour
     */
    getCurrentFundingRate(bids, asks) {
        const MIN_FUNDING = this.minFunding.toNumber();
        const MAX_FUNDING = this.maxFunding.toNumber();
        const bid = bids.getImpactPriceUi(new anchor.BN(this.impactQuantity));
        const ask = asks.getImpactPriceUi(new anchor.BN(this.impactQuantity));
        const indexPrice = this._uiPrice;
        let funding;
        if (bid !== undefined && ask !== undefined) {
            const bookPrice = (bid + ask) / 2;
            funding = Math.min(Math.max(bookPrice / indexPrice - 1, MIN_FUNDING), MAX_FUNDING);
        }
        else if (bid !== undefined) {
            funding = MAX_FUNDING;
        }
        else if (ask !== undefined) {
            funding = MIN_FUNDING;
        }
        else {
            funding = 0;
        }
        return funding / 24 / Math.pow(10, QUOTE_DECIMALS);
    }
    uiPriceToLots(price) {
        return toNative(price, QUOTE_DECIMALS)
            .mul(this.baseLotSize)
            .div(this.quoteLotSize.mul(new anchor.BN(Math.pow(10, this.baseDecimals))));
    }
    uiBaseToLots(quantity) {
        return toNative(quantity, this.baseDecimals).div(this.baseLotSize);
    }
    uiQuoteToLots(uiQuote) {
        return toNative(uiQuote, QUOTE_DECIMALS).div(this.quoteLotSize);
    }
    priceLotsToUi(price) {
        return parseFloat(price.toString()) * this.priceLotsToUiConverter;
    }
    priceNativeToUi(price) {
        return toUiDecimals(price, QUOTE_DECIMALS - this.baseDecimals);
    }
    baseLotsToUi(quantity) {
        return parseFloat(quantity.toString()) * this.baseLotsToUiConverter;
    }
    quoteLotsToUi(quantity) {
        return parseFloat(quantity.toString()) * this.quoteLotsToUiConverter;
    }
    /**
     * Returns a list of (upto count) accounts, and the pnl that is settle'able on this perp market,
     * the list is sorted ascending for 'negative' direction and descending for 'positive' direction.
     *
     * NOTE: keep in sync with perp_pnl.rs:fetch_top
     *
     * TODO: replace with a more performant offchain service call
     * @param client
     * @param group
     * @param direction
     * @returns
     */
    async getSettlePnlCandidates(client, group, direction, count = 2) {
        let accountsWithSettleablePnl = (await client.getAllMangoAccounts(group, true))
            .filter((acc) => acc.perpPositionExistsForMarket(this))
            .map((acc) => {
            const pp = acc
                .perpActive()
                .find((pp) => pp.marketIndex === this.perpMarketIndex);
            return {
                account: acc,
                settleablePnl: pp.getSettleablePnl(group, this, acc),
            };
        });
        accountsWithSettleablePnl = accountsWithSettleablePnl
            .filter((acc) => 
        // need perp positions with -ve pnl to settle +ve pnl and vice versa
        (direction === 'negative' && acc.settleablePnl.lt(ZERO_I80F48())) ||
            (direction === 'positive' && acc.settleablePnl.gt(ZERO_I80F48())))
            .sort((a, b) => direction === 'negative'
            ? // most negative
                a.settleablePnl.cmp(b.settleablePnl)
            : // most positive
                b.settleablePnl.cmp(a.settleablePnl));
        if (direction === 'negative') {
            let stable = 0;
            for (let i = 0; i < accountsWithSettleablePnl.length; i++) {
                const acc = accountsWithSettleablePnl[i];
                const nextPnl = i + 1 < accountsWithSettleablePnl.length
                    ? accountsWithSettleablePnl[i + 1].settleablePnl
                    : ZERO_I80F48();
                const perpSettleHealth = acc.account.getPerpSettleHealth(group);
                acc.settleablePnl =
                    // need positive settle health to settle against +ve pnl
                    perpSettleHealth.gt(ZERO_I80F48())
                        ? // can only settle min
                            acc.settleablePnl.max(perpSettleHealth.neg())
                        : ZERO_I80F48();
                // If the ordering was unchanged `count` times we know we have the top `count` accounts
                if (acc.settleablePnl.lte(nextPnl)) {
                    stable += 1;
                    if (stable >= count) {
                        break;
                    }
                }
            }
        }
        accountsWithSettleablePnl.sort((a, b) => direction === 'negative'
            ? // most negative
                a.settleablePnl.cmp(b.settleablePnl)
            : // most positive
                b.settleablePnl.cmp(a.settleablePnl));
        return accountsWithSettleablePnl.slice(0, count);
    }
    toString() {
        return ('PerpMarket ' +
            '\n perpMarketIndex -' +
            this.perpMarketIndex +
            '\n maintAssetWeight -' +
            this.maintBaseAssetWeight.toString() +
            '\n initAssetWeight -' +
            this.initBaseAssetWeight.toString() +
            '\n maintLiabWeight -' +
            this.maintBaseLiabWeight.toString() +
            '\n initLiabWeight -' +
            this.initBaseLiabWeight.toString() +
            '\n baseLiquidationFee -' +
            this.baseLiquidationFee.toString() +
            '\n makerFee -' +
            this.makerFee.toString() +
            '\n takerFee -' +
            this.takerFee.toString());
    }
}
class BookSide {
    client;
    perpMarket;
    type;
    rootFixed;
    rootOraclePegged;
    orderTreeNodes;
    static INNER_NODE_TAG = 1;
    static LEAF_NODE_TAG = 2;
    now;
    static from(client, perpMarket, bookSideType, obj) {
        return new BookSide(client, perpMarket, bookSideType, obj.roots[0], obj.roots[1], obj.nodes);
    }
    constructor(client, perpMarket, type, rootFixed, rootOraclePegged, orderTreeNodes, maxBookDelay) {
        this.client = client;
        this.perpMarket = perpMarket;
        this.type = type;
        this.rootFixed = rootFixed;
        this.rootOraclePegged = rootOraclePegged;
        this.orderTreeNodes = orderTreeNodes;
        // Determine the maxTimestamp found on the book to use for tif
        // If maxBookDelay is not provided, use 3600 as a very large number
        maxBookDelay = maxBookDelay === undefined ? 3600 : maxBookDelay;
        let maxTimestamp = new anchor.BN(new Date().getTime() / 1000 - maxBookDelay);
        for (const node of this.orderTreeNodes.nodes) {
            if (node.tag !== BookSide.LEAF_NODE_TAG) {
                continue;
            }
            const leafNode = BookSide.toLeafNode(client, node.data);
            if (leafNode.timestamp.gt(maxTimestamp)) {
                maxTimestamp = leafNode.timestamp;
            }
        }
        this.now = maxTimestamp;
    }
    static getPriceFromKey(key) {
        return key.ushrn(64);
    }
    /**
     * iterates over all orders
     */
    *items() {
        function isBetter(type, a, b) {
            return a.priceLots.eq(b.priceLots)
                ? a.seqNum.lt(b.seqNum) // if prices are equal prefer perp orders in the order they are placed
                : type === BookSideType.bids // else compare the actual prices
                    ? a.priceLots.gt(b.priceLots)
                    : b.priceLots.gt(a.priceLots);
        }
        const fGen = this.fixedItems();
        const oPegGen = this.oraclePeggedItems();
        let fOrderRes = fGen.next();
        let oPegOrderRes = oPegGen.next();
        while (true) {
            if (fOrderRes.value && oPegOrderRes.value) {
                if (isBetter(this.type, fOrderRes.value, oPegOrderRes.value)) {
                    yield fOrderRes.value;
                    fOrderRes = fGen.next();
                }
                else {
                    yield oPegOrderRes.value;
                    oPegOrderRes = oPegGen.next();
                }
            }
            else if (fOrderRes.value && !oPegOrderRes.value) {
                yield fOrderRes.value;
                fOrderRes = fGen.next();
            }
            else if (!fOrderRes.value && oPegOrderRes.value) {
                yield oPegOrderRes.value;
                oPegOrderRes = oPegGen.next();
            }
            else if (!fOrderRes.value && !oPegOrderRes.value) {
                break;
            }
        }
    }
    /**
     * iterates over all orders,
     * skips oracle pegged orders which are invalid due to oracle price crossing the peg limit,
     * skips tif orders which are invalid due to tif having elapsed,
     */
    *itemsValid() {
        const itemsGen = this.items();
        let itemsRes = itemsGen.next();
        while (true) {
            if (itemsRes.value) {
                const val = itemsRes.value;
                if (!val.isExpired &&
                    (!val.isOraclePegged ||
                        (val.isOraclePegged && !val.oraclePeggedProperties.isInvalid))) {
                    yield val;
                }
                itemsRes = itemsGen.next();
            }
            else {
                break;
            }
        }
    }
    *fixedItems() {
        if (this.rootFixed.leafCount === 0) {
            return;
        }
        const now = this.now;
        const stack = [this.rootFixed.maybeNode];
        const [left, right] = this.type === BookSideType.bids ? [1, 0] : [0, 1];
        while (stack.length > 0) {
            const index = stack.pop();
            const node = this.orderTreeNodes.nodes[index];
            if (node.tag === BookSide.INNER_NODE_TAG) {
                const innerNode = BookSide.toInnerNode(this.client, node.data);
                stack.push(innerNode.children[right], innerNode.children[left]);
            }
            else if (node.tag === BookSide.LEAF_NODE_TAG) {
                const leafNode = BookSide.toLeafNode(this.client, node.data);
                const expiryTimestamp = leafNode.timeInForce
                    ? leafNode.timestamp.add(new anchor.BN(leafNode.timeInForce))
                    : U64_MAX_BN;
                yield PerpOrder.from(this.perpMarket, leafNode, this.type, now.lt(expiryTimestamp));
            }
        }
    }
    *oraclePeggedItems() {
        if (this.rootOraclePegged.leafCount === 0) {
            return;
        }
        const now = this.now;
        const stack = [this.rootOraclePegged.maybeNode];
        const [left, right] = this.type === BookSideType.bids ? [1, 0] : [0, 1];
        while (stack.length > 0) {
            const index = stack.pop();
            const node = this.orderTreeNodes.nodes[index];
            if (node.tag === BookSide.INNER_NODE_TAG) {
                const innerNode = BookSide.toInnerNode(this.client, node.data);
                stack.push(innerNode.children[right], innerNode.children[left]);
            }
            else if (node.tag === BookSide.LEAF_NODE_TAG) {
                const leafNode = BookSide.toLeafNode(this.client, node.data);
                const expiryTimestamp = leafNode.timeInForce
                    ? leafNode.timestamp.add(new anchor.BN(leafNode.timeInForce))
                    : U64_MAX_BN;
                yield PerpOrder.from(this.perpMarket, leafNode, this.type, now.lt(expiryTimestamp), true);
            }
        }
    }
    best() {
        return this.items().next().value;
    }
    getImpactPriceUi(baseLots) {
        const s = new anchor.BN(0);
        for (const order of this.items()) {
            s.iadd(order.sizeLots);
            if (s.gte(baseLots)) {
                return order.uiPrice;
            }
        }
        return undefined;
    }
    getL2(depth) {
        const levels = [];
        for (const { priceLots, sizeLots } of this.items()) {
            if (levels.length > 0 && levels[levels.length - 1][0].eq(priceLots)) {
                levels[levels.length - 1][1].iadd(sizeLots);
            }
            else if (levels.length === depth) {
                break;
            }
            else {
                levels.push([priceLots, sizeLots]);
            }
        }
        return levels.map(([priceLots, sizeLots]) => [
            this.perpMarket.priceLotsToUi(priceLots),
            this.perpMarket.baseLotsToUi(sizeLots),
            priceLots,
            sizeLots,
        ]);
    }
    getL2Ui(depth) {
        const levels = [];
        for (const { uiPrice: price, uiSize: size } of this.items()) {
            if (levels.length > 0 && levels[levels.length - 1][0] === price) {
                levels[levels.length - 1][1] += size;
            }
            else if (levels.length === depth) {
                break;
            }
            else {
                levels.push([price, size]);
            }
        }
        return levels;
    }
    static toInnerNode(client, data) {
        return client.program._coder.types.typeLayouts
            .get('InnerNode')
            .decode(Buffer.from([BookSide.INNER_NODE_TAG].concat(data)));
    }
    static toLeafNode(client, data) {
        return LeafNode.from(client.program._coder.types.typeLayouts
            .get('LeafNode')
            .decode(Buffer.from([BookSide.LEAF_NODE_TAG].concat(data))));
    }
}
class BookSideType {
    static bids = { bids: {} };
    static asks = { asks: {} };
}
class LeafNode {
    ownerSlot;
    orderType;
    timeInForce;
    key;
    owner;
    quantity;
    timestamp;
    pegLimit;
    static from(obj) {
        return new LeafNode(obj.ownerSlot, obj.orderType, obj.timeInForce, obj.key, obj.owner, obj.quantity, obj.timestamp, obj.pegLimit);
    }
    constructor(ownerSlot, orderType, timeInForce, key, owner, quantity, timestamp, pegLimit) {
        this.ownerSlot = ownerSlot;
        this.orderType = orderType;
        this.timeInForce = timeInForce;
        this.key = key;
        this.owner = owner;
        this.quantity = quantity;
        this.timestamp = timestamp;
        this.pegLimit = pegLimit;
    }
}
class InnerNode {
    children;
    static from(obj) {
        return new InnerNode(obj.children);
    }
    constructor(children) {
        this.children = children;
    }
}
class PerpOrderSide {
    static bid = { bid: {} };
    static ask = { ask: {} };
}
class PerpOrderType {
    static limit = { limit: {} };
    static immediateOrCancel = { immediateOrCancel: {} };
    static postOnly = { postOnly: {} };
    static market = { market: {} };
    static postOnlySlide = { postOnlySlide: {} };
}
class PerpOrder {
    seqNum;
    orderId;
    owner;
    openOrdersSlot;
    feeTier;
    uiPrice;
    priceLots;
    uiSize;
    sizeLots;
    side;
    timestamp;
    expiryTimestamp;
    perpMarketIndex;
    isExpired;
    isOraclePegged;
    oraclePeggedProperties;
    static from(perpMarket, leafNode, type, isExpired = false, isOraclePegged = false) {
        const side = type == BookSideType.bids ? PerpOrderSide.bid : PerpOrderSide.ask;
        let priceLots;
        let oraclePeggedProperties;
        if (isOraclePegged) {
            const priceData = leafNode.key.ushrn(64);
            const priceOffset = priceData.sub(new anchor.BN(1).ushln(63));
            priceLots = perpMarket.uiPriceToLots(perpMarket.uiPrice).add(priceOffset);
            const isInvalid = type === BookSideType.bids
                ? priceLots.gt(leafNode.pegLimit)
                : leafNode.pegLimit.gt(priceLots);
            oraclePeggedProperties = {
                isInvalid,
                priceOffset,
                uiPriceOffset: perpMarket.priceLotsToUi(priceOffset),
                pegLimit: leafNode.pegLimit,
                uiPegLimit: perpMarket.priceLotsToUi(leafNode.pegLimit),
            };
        }
        else {
            priceLots = BookSide.getPriceFromKey(leafNode.key);
        }
        const expiryTimestamp = leafNode.timeInForce
            ? leafNode.timestamp.add(new anchor.BN(leafNode.timeInForce))
            : U64_MAX_BN;
        return new PerpOrder(type === BookSideType.bids
            ? RUST_U64_MAX().sub(leafNode.key.maskn(64))
            : leafNode.key.maskn(64), leafNode.key, leafNode.owner, leafNode.ownerSlot, 0, perpMarket.priceLotsToUi(priceLots), priceLots, perpMarket.baseLotsToUi(leafNode.quantity), leafNode.quantity, side, leafNode.timestamp, expiryTimestamp, perpMarket.perpMarketIndex, isExpired, isOraclePegged, oraclePeggedProperties);
    }
    constructor(seqNum, orderId, owner, openOrdersSlot, feeTier, uiPrice, priceLots, uiSize, sizeLots, side, timestamp, expiryTimestamp, perpMarketIndex, isExpired = false, isOraclePegged = false, oraclePeggedProperties) {
        this.seqNum = seqNum;
        this.orderId = orderId;
        this.owner = owner;
        this.openOrdersSlot = openOrdersSlot;
        this.feeTier = feeTier;
        this.uiPrice = uiPrice;
        this.priceLots = priceLots;
        this.uiSize = uiSize;
        this.sizeLots = sizeLots;
        this.side = side;
        this.timestamp = timestamp;
        this.expiryTimestamp = expiryTimestamp;
        this.perpMarketIndex = perpMarketIndex;
        this.isExpired = isExpired;
        this.isOraclePegged = isOraclePegged;
        this.oraclePeggedProperties = oraclePeggedProperties;
    }
    get price() {
        return this.uiPrice;
    }
    get size() {
        return this.uiSize;
    }
}
class PerpEventQueue {
    static FILL_EVENT_TYPE = 0;
    static OUT_EVENT_TYPE = 1;
    static LIQUIDATE_EVENT_TYPE = 2;
    head;
    count;
    seqNum;
    rawEvents;
    constructor(client, header, buf) {
        this.head = header.head;
        this.count = header.count;
        this.seqNum = header.seqNum;
        this.rawEvents = buf.map((event) => {
            if (event.eventType === PerpEventQueue.FILL_EVENT_TYPE) {
                return client.program._coder.types.typeLayouts
                    .get('FillEvent')
                    .decode(Buffer.from([PerpEventQueue.FILL_EVENT_TYPE].concat(event.padding)));
            }
            else if (event.eventType === PerpEventQueue.OUT_EVENT_TYPE) {
                return client.program._coder.types.typeLayouts
                    .get('OutEvent')
                    .decode(Buffer.from([PerpEventQueue.OUT_EVENT_TYPE].concat(event.padding)));
            }
            else if (event.eventType === PerpEventQueue.LIQUIDATE_EVENT_TYPE) {
                return client.program._coder.types.typeLayouts
                    .get('LiquidateEvent')
                    .decode(Buffer.from([PerpEventQueue.LIQUIDATE_EVENT_TYPE].concat(event.padding)));
            }
            throw new Error(`Unknown event with eventType ${event.eventType}!`);
        });
    }
    getUnconsumedEvents() {
        const events = [];
        const head = this.head;
        for (let i = 0; i < this.count; i++) {
            events.push(this.rawEvents[(head + i) % this.rawEvents.length]);
        }
        return events;
    }
    eventsSince(lastSeqNum) {
        return this.rawEvents
            .filter((e) => e.seqNum.gt(lastSeqNum === undefined ? new anchor.BN(0) : lastSeqNum))
            .sort((a, b) => a.seqNum.cmp(b.seqNum));
    }
}

class Serum3Market {
    publicKey;
    group;
    baseTokenIndex;
    quoteTokenIndex;
    serumProgram;
    serumMarketExternal;
    marketIndex;
    registrationTime;
    reduceOnly;
    name;
    static from(publicKey, obj) {
        return new Serum3Market(publicKey, obj.group, obj.baseTokenIndex, obj.quoteTokenIndex, obj.name, obj.serumProgram, obj.serumMarketExternal, obj.marketIndex, obj.registrationTime, obj.reduceOnly == 1);
    }
    constructor(publicKey, group, baseTokenIndex, quoteTokenIndex, name, serumProgram, serumMarketExternal, marketIndex, registrationTime, reduceOnly) {
        this.publicKey = publicKey;
        this.group = group;
        this.baseTokenIndex = baseTokenIndex;
        this.quoteTokenIndex = quoteTokenIndex;
        this.serumProgram = serumProgram;
        this.serumMarketExternal = serumMarketExternal;
        this.marketIndex = marketIndex;
        this.registrationTime = registrationTime;
        this.reduceOnly = reduceOnly;
        this.name = bytes.utf8.decode(new Uint8Array(name)).split('\x00')[0];
    }
    async findOoPda(programId, mangoAccount) {
        const [openOrderPublicKey] = await web3_js.PublicKey.findProgramAddress([
            Buffer.from('Serum3OO'),
            mangoAccount.toBuffer(),
            this.publicKey.toBuffer(),
        ], programId);
        return openOrderPublicKey;
    }
    getFeeRates(taker = true) {
        // See https://github.com/openbook-dex/program/blob/master/dex/src/fees.rs#L81
        const ratesBps = this.name === 'USDT/USDC'
            ? { maker: -0.5, taker: 1 }
            : { maker: -2, taker: 4 };
        return taker ? ratesBps.maker * 0.0001 : ratesBps.taker * 0.0001;
    }
    /**
     *
     * @param group
     * @returns maximum leverage one can bid on this market, this is only for display purposes,
     *  also see getMaxQuoteForSerum3BidUi and getMaxBaseForSerum3AskUi
     */
    maxBidLeverage(group) {
        const baseBank = group.getFirstBankByTokenIndex(this.baseTokenIndex);
        const quoteBank = group.getFirstBankByTokenIndex(this.quoteTokenIndex);
        if (quoteBank.initLiabWeight.sub(baseBank.initAssetWeight).lte(ZERO_I80F48())) {
            return MAX_I80F48().toNumber();
        }
        return ONE_I80F48()
            .div(quoteBank.initLiabWeight.sub(baseBank.initAssetWeight))
            .toNumber();
    }
    /**
     *
     * @param group
     * @returns maximum leverage one can ask on this market, this is only for display purposes,
     *  also see getMaxQuoteForSerum3BidUi and getMaxBaseForSerum3AskUi
     */
    maxAskLeverage(group) {
        const baseBank = group.getFirstBankByTokenIndex(this.baseTokenIndex);
        const quoteBank = group.getFirstBankByTokenIndex(this.quoteTokenIndex);
        if (baseBank.initLiabWeight.sub(quoteBank.initAssetWeight).lte(ZERO_I80F48())) {
            return MAX_I80F48().toNumber();
        }
        return ONE_I80F48()
            .div(baseBank.initLiabWeight.sub(quoteBank.initAssetWeight))
            .toNumber();
    }
    async loadBids(client, group) {
        const serum3MarketExternal = group.getSerum3ExternalMarket(this.serumMarketExternal);
        return await serum3MarketExternal.loadBids(client.program.provider.connection);
    }
    async loadAsks(client, group) {
        const serum3MarketExternal = group.getSerum3ExternalMarket(this.serumMarketExternal);
        return await serum3MarketExternal.loadAsks(client.program.provider.connection);
    }
    async logOb(client, group) {
        let res = ``;
        res += `  ${this.name} OrderBook`;
        let orders = await this?.loadAsks(client, group);
        for (const order of orders.items(true)) {
            res += `\n  ${order.price.toString().padStart(10)}, ${order.size
                .toString()
                .padStart(10)}`;
        }
        res += `\n  --------------------------`;
        orders = await this?.loadBids(client, group);
        for (const order of orders.items(true)) {
            res += `\n  ${order.price.toString().padStart(10)}, ${order.size
                .toString()
                .padStart(10)}`;
        }
        return res;
    }
}
class Serum3SelfTradeBehavior {
    static decrementTake = { decrementTake: {} };
    static cancelProvide = { cancelProvide: {} };
    static abortTransaction = { abortTransaction: {} };
}
class Serum3OrderType {
    static limit = { limit: {} };
    static immediateOrCancel = { immediateOrCancel: {} };
    static postOnly = { postOnly: {} };
}
class Serum3Side {
    static bid = { bid: {} };
    static ask = { ask: {} };
}
async function generateSerum3MarketExternalVaultSignerAddress(cluster, serum3Market, serum3MarketExternal) {
    return await web3_js.PublicKey.createProgramAddress([
        serum3Market.serumMarketExternal.toBuffer(),
        serum3MarketExternal.decoded.vaultSignerNonce.toArrayLike(Buffer, 'le', 8),
    ], OPENBOOK_PROGRAM_ID[cluster]);
}

class Group {
    publicKey;
    creator;
    groupNum;
    admin;
    fastListingAdmin;
    securityAdmin;
    insuranceMint;
    insuranceVault;
    testing;
    version;
    ixGate;
    addressLookupTables;
    addressLookupTablesList;
    banksMapByName;
    banksMapByMint;
    banksMapByTokenIndex;
    serum3MarketsMapByExternal;
    serum3MarketsMapByMarketIndex;
    serum3ExternalMarketsMap;
    perpMarketsMapByOracle;
    perpMarketsMapByMarketIndex;
    perpMarketsMapByName;
    mintInfosMapByTokenIndex;
    mintInfosMapByMint;
    vaultAmountsMap;
    static from(publicKey, obj) {
        return new Group(publicKey, obj.creator, obj.groupNum, obj.admin, obj.fastListingAdmin, obj.securityAdmin, obj.insuranceMint, obj.insuranceVault, obj.testing, obj.version, obj.ixGate, obj.addressLookupTables, [], // addressLookupTablesList
        new Map(), // banksMapByName
        new Map(), // banksMapByMint
        new Map(), // banksMapByTokenIndex
        new Map(), // serum3MarketsMapByExternal
        new Map(), // serum3MarketsMapByMarketIndex
        new Map(), // serum3MarketExternalsMap
        new Map(), // perpMarketsMapByOracle
        new Map(), // perpMarketsMapByMarketIndex
        new Map(), // perpMarketsMapByName
        new Map(), // mintInfosMapByTokenIndex
        new Map(), // mintInfosMapByMint
        new Map());
    }
    constructor(publicKey, creator, groupNum, admin, fastListingAdmin, securityAdmin, insuranceMint, insuranceVault, testing, version, ixGate, addressLookupTables, addressLookupTablesList, banksMapByName, banksMapByMint, banksMapByTokenIndex, serum3MarketsMapByExternal, serum3MarketsMapByMarketIndex, serum3ExternalMarketsMap, perpMarketsMapByOracle, perpMarketsMapByMarketIndex, perpMarketsMapByName, mintInfosMapByTokenIndex, mintInfosMapByMint, vaultAmountsMap) {
        this.publicKey = publicKey;
        this.creator = creator;
        this.groupNum = groupNum;
        this.admin = admin;
        this.fastListingAdmin = fastListingAdmin;
        this.securityAdmin = securityAdmin;
        this.insuranceMint = insuranceMint;
        this.insuranceVault = insuranceVault;
        this.testing = testing;
        this.version = version;
        this.ixGate = ixGate;
        this.addressLookupTables = addressLookupTables;
        this.addressLookupTablesList = addressLookupTablesList;
        this.banksMapByName = banksMapByName;
        this.banksMapByMint = banksMapByMint;
        this.banksMapByTokenIndex = banksMapByTokenIndex;
        this.serum3MarketsMapByExternal = serum3MarketsMapByExternal;
        this.serum3MarketsMapByMarketIndex = serum3MarketsMapByMarketIndex;
        this.serum3ExternalMarketsMap = serum3ExternalMarketsMap;
        this.perpMarketsMapByOracle = perpMarketsMapByOracle;
        this.perpMarketsMapByMarketIndex = perpMarketsMapByMarketIndex;
        this.perpMarketsMapByName = perpMarketsMapByName;
        this.mintInfosMapByTokenIndex = mintInfosMapByTokenIndex;
        this.mintInfosMapByMint = mintInfosMapByMint;
        this.vaultAmountsMap = vaultAmountsMap;
    }
    async reloadAll(client) {
        const ids = await client.getIds(this.publicKey);
        // console.time('group.reload');
        await Promise.all([
            this.reloadAlts(client),
            this.reloadBanks(client, ids).then(() => Promise.all([
                this.reloadBankOraclePrices(client),
                this.reloadVaults(client),
                this.reloadPerpMarkets(client, ids).then(() => this.reloadPerpMarketOraclePrices(client)),
            ])),
            this.reloadMintInfos(client, ids),
            this.reloadSerum3Markets(client, ids).then(() => this.reloadSerum3ExternalMarkets(client)),
        ]);
        // console.timeEnd('group.reload');
    }
    async reloadAlts(client) {
        const alts = await Promise.all(this.addressLookupTables
            .filter((alt) => !alt.equals(web3_js.PublicKey.default))
            .map((alt) => client.program.provider.connection.getAddressLookupTable(alt)));
        this.addressLookupTablesList = alts.map((res, i) => {
            if (!res || !res.value) {
                throw new Error(`Undefined ALT ${this.addressLookupTables[i]}!`);
            }
            return res.value;
        });
    }
    async reloadBanks(client, ids) {
        let banks;
        if (ids && ids.getBanks().length) {
            banks = (await client.program.account.bank.fetchMultiple(ids.getBanks())).map((account, index) => Bank.from(ids.getBanks()[index], account));
        }
        else {
            banks = await client.getBanksForGroup(this);
        }
        const oldbanksMapByTokenIndex = lodash.cloneDeep(this.banksMapByTokenIndex);
        this.banksMapByName = new Map();
        this.banksMapByMint = new Map();
        this.banksMapByTokenIndex = new Map();
        for (const bank of banks) {
            // ensure that freshly fetched banks have valid price until we fetch oracles again
            const oldBanks = oldbanksMapByTokenIndex.get(bank.tokenIndex);
            if (oldBanks && oldBanks.length > 0) {
                lodash.merge(bank, oldBanks[0]);
            }
            const mintId = bank.mint.toString();
            if (this.banksMapByMint.has(mintId)) {
                this.banksMapByMint.get(mintId)?.push(bank);
                this.banksMapByName.get(bank.name)?.push(bank);
                this.banksMapByTokenIndex.get(bank.tokenIndex)?.push(bank);
            }
            else {
                this.banksMapByMint.set(mintId, [bank]);
                this.banksMapByName.set(bank.name, [bank]);
                this.banksMapByTokenIndex.set(bank.tokenIndex, [bank]);
            }
        }
    }
    async reloadMintInfos(client, ids) {
        let mintInfos;
        if (ids && ids.getMintInfos().length) {
            mintInfos = (await client.program.account.mintInfo.fetchMultiple(ids.getMintInfos())).map((account, index) => MintInfo.from(ids.getMintInfos()[index], account));
        }
        else {
            mintInfos = await client.getMintInfosForGroup(this);
        }
        this.mintInfosMapByTokenIndex = new Map(mintInfos.map((mintInfo) => {
            return [mintInfo.tokenIndex, mintInfo];
        }));
        this.mintInfosMapByMint = new Map(mintInfos.map((mintInfo) => {
            return [mintInfo.mint.toString(), mintInfo];
        }));
    }
    async reloadSerum3Markets(client, ids) {
        let serum3Markets;
        if (ids && ids.getSerum3Markets().length) {
            serum3Markets = (await client.program.account.serum3Market.fetchMultiple(ids.getSerum3Markets())).map((account, index) => Serum3Market.from(ids.getSerum3Markets()[index], account));
        }
        else {
            serum3Markets = await client.serum3GetMarkets(this);
        }
        this.serum3MarketsMapByExternal = new Map(serum3Markets.map((serum3Market) => [
            serum3Market.serumMarketExternal.toBase58(),
            serum3Market,
        ]));
        this.serum3MarketsMapByMarketIndex = new Map(serum3Markets.map((serum3Market) => [
            serum3Market.marketIndex,
            serum3Market,
        ]));
    }
    async reloadSerum3ExternalMarkets(client) {
        const externalMarkets = await Promise.all(Array.from(this.serum3MarketsMapByExternal.values()).map((serum3Market) => serum.Market.load(client.program.provider.connection, serum3Market.serumMarketExternal, { commitment: client.program.provider.connection.commitment }, OPENBOOK_PROGRAM_ID[client.cluster])));
        this.serum3ExternalMarketsMap = new Map(Array.from(this.serum3MarketsMapByExternal.values()).map((serum3Market, index) => [
            serum3Market.serumMarketExternal.toBase58(),
            externalMarkets[index],
        ]));
    }
    async reloadPerpMarkets(client, ids) {
        let perpMarkets;
        if (ids && ids.getPerpMarkets().length) {
            perpMarkets = (await client.program.account.perpMarket.fetchMultiple(ids.getPerpMarkets())).map((account, index) => PerpMarket.from(ids.getPerpMarkets()[index], account));
        }
        else {
            perpMarkets = await client.perpGetMarkets(this);
        }
        // ensure that freshly fetched perp markets have valid price until we fetch oracles again
        const oldPerpMarketByMarketIndex = lodash.cloneDeep(this.perpMarketsMapByMarketIndex);
        for (const perpMarket of perpMarkets) {
            const oldPerpMarket = oldPerpMarketByMarketIndex.get(perpMarket.perpMarketIndex);
            if (oldPerpMarket) {
                lodash.merge(perpMarket, oldPerpMarket);
            }
        }
        this.perpMarketsMapByName = new Map(perpMarkets.map((perpMarket) => [perpMarket.name, perpMarket]));
        this.perpMarketsMapByOracle = new Map(perpMarkets.map((perpMarket) => [
            perpMarket.oracle.toBase58(),
            perpMarket,
        ]));
        this.perpMarketsMapByMarketIndex = new Map(perpMarkets.map((perpMarket) => [perpMarket.perpMarketIndex, perpMarket]));
    }
    async reloadBankOraclePrices(client) {
        const banks = Array.from(this.banksMapByMint, ([, value]) => value);
        const oracles = banks.map((b) => b[0].oracle);
        const ais = await client.program.provider.connection.getMultipleAccountsInfo(oracles);
        const coder = new anchor.BorshAccountsCoder(client.program.idl);
        for (const [index, ai] of ais.entries()) {
            for (const bank of banks[index]) {
                if (bank.name === 'USDC') {
                    bank._price = ONE_I80F48();
                    bank._uiPrice = 1;
                }
                else {
                    if (!ai)
                        throw new Error(`Undefined accountInfo object in reloadBankOraclePrices for ${bank.oracle}!`);
                    const { price, uiPrice, lastUpdatedSlot } = await this.decodePriceFromOracleAi(coder, bank.oracle, ai, this.getMintDecimals(bank.mint), client);
                    bank._price = price;
                    bank._uiPrice = uiPrice;
                    bank._oracleLastUpdatedSlot = lastUpdatedSlot;
                }
            }
        }
    }
    async reloadPerpMarketOraclePrices(client) {
        const perpMarkets = Array.from(this.perpMarketsMapByName.values());
        const oracles = perpMarkets.map((b) => b.oracle);
        const ais = await client.program.provider.connection.getMultipleAccountsInfo(oracles);
        const coder = new anchor.BorshAccountsCoder(client.program.idl);
        await Promise.all(Array.from(ais.entries()).map(async ([i, ai]) => {
            const perpMarket = perpMarkets[i];
            if (!ai)
                throw new Error(`Undefined ai object in reloadPerpMarketOraclePrices for ${perpMarket.oracle}!`);
            const { price, uiPrice, lastUpdatedSlot } = await this.decodePriceFromOracleAi(coder, perpMarket.oracle, ai, perpMarket.baseDecimals, client);
            perpMarket._price = price;
            perpMarket._uiPrice = uiPrice;
            perpMarket._oracleLastUpdatedSlot = lastUpdatedSlot;
        }));
    }
    async decodePriceFromOracleAi(coder, oracle, ai, baseDecimals, client$1) {
        let price, uiPrice, lastUpdatedSlot;
        if (!anchor.BorshAccountsCoder.accountDiscriminator('stubOracle').compare(ai.data.slice(0, 8))) {
            const stubOracle = coder.decode('stubOracle', ai.data);
            price = new I80F48(stubOracle.price.val);
            uiPrice = this.toUiPrice(price, baseDecimals);
            lastUpdatedSlot = stubOracle.lastUpdated.val;
        }
        else if (isPythOracle(ai)) {
            const priceData = client.parsePriceData(ai.data);
            uiPrice = priceData.previousPrice;
            price = this.toNativePrice(uiPrice, baseDecimals);
            lastUpdatedSlot = parseInt(priceData.lastSlot.toString());
        }
        else if (isSwitchboardOracle(ai)) {
            const priceData = await parseSwitchboardOracle(ai, client$1.program.provider.connection);
            uiPrice = priceData.price;
            price = this.toNativePrice(uiPrice, baseDecimals);
            lastUpdatedSlot = priceData.lastUpdatedSlot;
        }
        else {
            throw new Error(`Unknown oracle provider (parsing not implemented) for oracle ${oracle}, with owner ${ai.owner}!`);
        }
        return { price, uiPrice, lastUpdatedSlot };
    }
    async reloadVaults(client) {
        const vaultPks = Array.from(this.banksMapByMint.values())
            .flat()
            .map((bank) => bank.vault);
        const vaultAccounts = await client.program.provider.connection.getMultipleAccountsInfo(vaultPks);
        this.vaultAmountsMap = new Map(vaultAccounts.map((vaultAi, i) => {
            if (!vaultAi) {
                throw new Error(`Undefined vaultAi for ${vaultPks[i]}`);
            }
            const vaultAmount = token.coder().accounts.decode('token', vaultAi.data).amount;
            return [vaultPks[i].toBase58(), vaultAmount];
        }));
    }
    getMintDecimals(mintPk) {
        const bank = this.getFirstBankByMint(mintPk);
        return bank.mintDecimals;
    }
    getMintDecimalsByTokenIndex(tokenIndex) {
        const bank = this.getFirstBankByTokenIndex(tokenIndex);
        return bank.mintDecimals;
    }
    getInsuranceMintDecimals() {
        return this.getMintDecimals(this.insuranceMint);
    }
    getFirstBankByMint(mintPk) {
        const banks = this.banksMapByMint.get(mintPk.toString());
        if (!banks)
            throw new Error(`No bank found for mint ${mintPk}!`);
        return banks[0];
    }
    getFirstBankByTokenIndex(tokenIndex) {
        const banks = this.banksMapByTokenIndex.get(tokenIndex);
        if (!banks)
            throw new Error(`No bank found for tokenIndex ${tokenIndex}!`);
        return banks[0];
    }
    /**
     *
     * @param mintPk
     * @returns sum of ui balances of vaults for all banks for a token
     */
    getTokenVaultBalanceByMintUi(mintPk) {
        const banks = this.banksMapByMint.get(mintPk.toBase58());
        if (!banks) {
            throw new Error(`No bank found for mint ${mintPk}!`);
        }
        const totalAmount = new BN(0);
        for (const bank of banks) {
            const amount = this.vaultAmountsMap.get(bank.vault.toBase58());
            if (!amount) {
                throw new Error(`Vault balance not found for bank ${bank.name} ${bank.bankNum}!`);
            }
            totalAmount.iadd(amount);
        }
        return toUiDecimals(totalAmount, this.getMintDecimals(mintPk));
    }
    getSerum3MarketByMarketIndex(marketIndex) {
        const serum3Market = this.serum3MarketsMapByMarketIndex.get(marketIndex);
        if (!serum3Market) {
            throw new Error(`No serum3Market found for marketIndex ${marketIndex}!`);
        }
        return serum3Market;
    }
    getSerum3MarketByName(name) {
        const serum3Market = Array.from(this.serum3MarketsMapByExternal.values()).find((serum3Market) => serum3Market.name === name);
        if (!serum3Market) {
            throw new Error(`No serum3Market found by name ${name}!`);
        }
        return serum3Market;
    }
    getSerum3MarketByExternalMarket(externalMarketPk) {
        const serum3Market = Array.from(this.serum3MarketsMapByExternal.values()).find((serum3Market) => serum3Market.serumMarketExternal.equals(externalMarketPk));
        if (!serum3Market) {
            throw new Error(`No serum3Market found for external serum3 market ${externalMarketPk.toString()}!`);
        }
        return serum3Market;
    }
    getSerum3ExternalMarket(externalMarketPk) {
        const market = this.serum3ExternalMarketsMap.get(externalMarketPk.toBase58());
        if (!market) {
            throw new Error(`No external market found for pk ${externalMarketPk.toString()}!`);
        }
        return market;
    }
    async loadSerum3BidsForMarket(client, externalMarketPk) {
        const serum3Market = this.getSerum3MarketByExternalMarket(externalMarketPk);
        return await serum3Market.loadBids(client, this);
    }
    async loadSerum3AsksForMarket(client, externalMarketPk) {
        const serum3Market = this.getSerum3MarketByExternalMarket(externalMarketPk);
        return await serum3Market.loadAsks(client, this);
    }
    findPerpMarket(marketIndex) {
        const perpMarket = Array.from(this.perpMarketsMapByName.values()).find((perpMarket) => perpMarket.perpMarketIndex === marketIndex);
        if (!perpMarket) {
            throw new Error(`No perpMarket found for perpMarketIndex ${marketIndex}!`);
        }
        return perpMarket;
    }
    getPerpMarketByOracle(oracle) {
        const perpMarket = this.perpMarketsMapByOracle.get(oracle.toBase58());
        if (!perpMarket) {
            throw new Error(`No PerpMarket found for oracle ${oracle}!`);
        }
        return perpMarket;
    }
    getPerpMarketByMarketIndex(marketIndex) {
        const perpMarket = this.perpMarketsMapByMarketIndex.get(marketIndex);
        if (!perpMarket) {
            throw new Error(`No PerpMarket found with marketIndex ${marketIndex}!`);
        }
        return perpMarket;
    }
    getPerpMarketByName(perpMarketName) {
        const perpMarket = Array.from(this.perpMarketsMapByMarketIndex.values()).find((perpMarket) => perpMarket.name === perpMarketName);
        if (!perpMarket) {
            throw new Error(`No PerpMarket found by name ${perpMarketName}!`);
        }
        return perpMarket;
    }
    async loadPerpBidsForMarket(client, perpMarketIndex) {
        const perpMarket = this.getPerpMarketByMarketIndex(perpMarketIndex);
        return await perpMarket.loadBids(client);
    }
    async loadPerpAsksForMarket(client, group, perpMarketIndex) {
        const perpMarket = this.getPerpMarketByMarketIndex(perpMarketIndex);
        return await perpMarket.loadAsks(client);
    }
    consoleLogBanks() {
        for (const mintBanks of this.banksMapByMint.values()) {
            for (const bank of mintBanks) {
                console.log(bank.toString());
            }
        }
    }
    toUiPrice(price, baseDecimals) {
        return toUiDecimals(price, this.getInsuranceMintDecimals() - baseDecimals);
    }
    toNativePrice(uiPrice, baseDecimals) {
        return toNativeI80F48(uiPrice, 
        // note: our oracles are quoted in USD and our insurance mint is USD
        // please update when these assumptions change
        this.getInsuranceMintDecimals() - baseDecimals);
    }
    toNativeDecimals(uiAmount, mintPk) {
        const decimals = this.getMintDecimals(mintPk);
        return toNative(uiAmount, decimals);
    }
    toString() {
        let res = 'Group\n';
        res = res + ' pk: ' + this.publicKey.toString();
        res =
            res +
                '\n mintInfos:' +
                Array.from(this.mintInfosMapByTokenIndex.entries())
                    .map((mintInfoTuple) => '  \n' + mintInfoTuple[0] + ') ' + mintInfoTuple[1].toString())
                    .join(', ');
        const banks = [];
        for (const tokenBanks of this.banksMapByMint.values()) {
            for (const bank of tokenBanks) {
                banks.push(bank);
            }
        }
        res =
            res +
                '\n banks:' +
                Array.from(banks)
                    .map((bank) => '  \n' + bank.name + ') ' + bank.toString())
                    .join(', ');
        return res;
    }
}

//               ░░░░
//
//                                           ██
//                                         ██░░██
// ░░          ░░                        ██░░░░░░██                            ░░░░
//                                     ██░░░░░░░░░░██
//                                     ██░░░░░░░░░░██
//                                   ██░░░░░░░░░░░░░░██
//                                 ██░░░░░░██████░░░░░░██
//                                 ██░░░░░░██████░░░░░░██
//                               ██░░░░░░░░██████░░░░░░░░██
//                               ██░░░░░░░░██████░░░░░░░░██
//                             ██░░░░░░░░░░██████░░░░░░░░░░██
//                           ██░░░░░░░░░░░░██████░░░░░░░░░░░░██
//                           ██░░░░░░░░░░░░██████░░░░░░░░░░░░██
//                         ██░░░░░░░░░░░░░░██████░░░░░░░░░░░░░░██
//                         ██░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░██
//                       ██░░░░░░░░░░░░░░░░██████░░░░░░░░░░░░░░░░██
//                       ██░░░░░░░░░░░░░░░░██████░░░░░░░░░░░░░░░░██
//                     ██░░░░░░░░░░░░░░░░░░██████░░░░░░░░░░░░░░░░░░██
//       ░░            ██░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░██
//                       ██████████████████████████████████████████
// warning: this code is copy pasta from rust, keep in sync with health.rs
class HealthCache {
    tokenInfos;
    serum3Infos;
    perpInfos;
    constructor(tokenInfos, serum3Infos, perpInfos) {
        this.tokenInfos = tokenInfos;
        this.serum3Infos = serum3Infos;
        this.perpInfos = perpInfos;
    }
    static fromMangoAccount(group, mangoAccount) {
        // token contribution from token accounts
        const tokenInfos = mangoAccount.tokensActive().map((tokenPosition) => {
            const bank = group.getFirstBankByTokenIndex(tokenPosition.tokenIndex);
            return TokenInfo.fromBank(bank, tokenPosition.balance(bank));
        });
        // Fill the TokenInfo balance with free funds in serum3 oo accounts, and fill
        // the serum3MaxReserved with their reserved funds. Also build Serum3Infos.
        const serum3Infos = mangoAccount.serum3Active().map((serum3) => {
            const oo = mangoAccount.getSerum3OoAccount(serum3.marketIndex);
            // find the TokenInfos for the market's base and quote tokens
            const baseIndex = tokenInfos.findIndex((tokenInfo) => tokenInfo.tokenIndex === serum3.baseTokenIndex);
            const baseInfo = tokenInfos[baseIndex];
            if (!baseInfo) {
                throw new Error(`BaseInfo not found for market with marketIndex ${serum3.marketIndex}!`);
            }
            const quoteIndex = tokenInfos.findIndex((tokenInfo) => tokenInfo.tokenIndex === serum3.quoteTokenIndex);
            const quoteInfo = tokenInfos[quoteIndex];
            if (!quoteInfo) {
                throw new Error(`QuoteInfo not found for market with marketIndex ${serum3.marketIndex}!`);
            }
            return Serum3Info.fromOoModifyingTokenInfos(baseIndex, baseInfo, quoteIndex, quoteInfo, serum3.marketIndex, oo);
        });
        // health contribution from perp accounts
        const perpInfos = mangoAccount.perpActive().map((perpPosition) => {
            const perpMarket = group.getPerpMarketByMarketIndex(perpPosition.marketIndex);
            return PerpInfo.fromPerpPosition(perpMarket, perpPosition);
        });
        return new HealthCache(tokenInfos, serum3Infos, perpInfos);
    }
    static fromDto(dto) {
        return new HealthCache(dto.tokenInfos.map((dto) => TokenInfo.fromDto(dto)), dto.serum3Infos.map((dto) => Serum3Info.fromDto(dto)), dto.perpInfos.map((dto) => PerpInfo.fromDto(dto)));
    }
    computeSerum3Reservations(healthType) {
        // For each token, compute the sum of serum-reserved amounts over all markets.
        const tokenMaxReserved = new Array(this.tokenInfos.length)
            .fill(null)
            .map((ignored) => ZERO_I80F48());
        // For each serum market, compute what happened if reserved_base was converted to quote
        // or reserved_quote was converted to base.
        const serum3Reserved = [];
        for (const info of this.serum3Infos) {
            const quote = this.tokenInfos[info.quoteIndex];
            const base = this.tokenInfos[info.baseIndex];
            const reservedBase = info.reservedBase;
            const reservedQuote = info.reservedQuote;
            const quoteAsset = quote.prices.asset(healthType);
            const baseLiab = base.prices.liab(healthType);
            const allReservedAsBase = reservedBase.add(reservedQuote.mul(quoteAsset).div(baseLiab));
            const baseAsset = base.prices.asset(healthType);
            const quoteLiab = quote.prices.liab(healthType);
            const allReservedAsQuote = reservedQuote.add(reservedBase.mul(baseAsset).div(quoteLiab));
            const baseMaxReserved = tokenMaxReserved[info.baseIndex];
            baseMaxReserved.iadd(allReservedAsBase);
            const quoteMaxReserved = tokenMaxReserved[info.quoteIndex];
            quoteMaxReserved.iadd(allReservedAsQuote);
            serum3Reserved.push(new Serum3Reserved(allReservedAsBase, allReservedAsQuote));
        }
        return {
            tokenMaxReserved: tokenMaxReserved,
            serum3Reserved: serum3Reserved,
        };
    }
    health(healthType) {
        const health = ZERO_I80F48();
        for (const tokenInfo of this.tokenInfos) {
            const contrib = tokenInfo.healthContribution(healthType);
            // console.log(` - ti ${contrib}`);
            health.iadd(contrib);
        }
        const res = this.computeSerum3Reservations(healthType);
        for (const [index, serum3Info] of this.serum3Infos.entries()) {
            const contrib = serum3Info.healthContribution(healthType, this.tokenInfos, res.tokenMaxReserved, res.serum3Reserved[index]);
            // console.log(` - si ${contrib}`);
            health.iadd(contrib);
        }
        for (const perpInfo of this.perpInfos) {
            const contrib = perpInfo.healthContribution(healthType);
            // console.log(` - pi ${contrib}`);
            health.iadd(contrib);
        }
        return health;
    }
    // Note: only considers positive perp pnl contributions, see program code for more reasoning
    perpSettleHealth() {
        const health = ZERO_I80F48();
        for (const tokenInfo of this.tokenInfos) {
            const contrib = tokenInfo.healthContribution(HealthType.maint);
            // console.log(` - ti ${contrib}`);
            health.iadd(contrib);
        }
        const res = this.computeSerum3Reservations(HealthType.maint);
        for (const [index, serum3Info] of this.serum3Infos.entries()) {
            const contrib = serum3Info.healthContribution(HealthType.maint, this.tokenInfos, res.tokenMaxReserved, res.serum3Reserved[index]);
            // console.log(` - si ${contrib}`);
            health.iadd(contrib);
        }
        for (const perpInfo of this.perpInfos) {
            const positiveContrib = perpInfo
                .healthContribution(HealthType.maint)
                .max(ZERO_I80F48());
            // console.log(` - pi ${positiveContrib}`);
            health.iadd(positiveContrib);
        }
        return health;
    }
    // An undefined HealthType will use an asset and liab weight of 1
    assets(healthType) {
        const assets = ZERO_I80F48();
        for (const tokenInfo of this.tokenInfos) {
            const contrib = tokenInfo.healthContribution(healthType);
            if (contrib.isPos()) {
                assets.iadd(contrib);
            }
        }
        const res = this.computeSerum3Reservations(HealthType.maint);
        for (const [index, serum3Info] of this.serum3Infos.entries()) {
            const contrib = serum3Info.healthContribution(healthType, this.tokenInfos, res.tokenMaxReserved, res.serum3Reserved[index]);
            if (contrib.isPos()) {
                assets.iadd(contrib);
            }
        }
        for (const perpInfo of this.perpInfos) {
            const contrib = perpInfo.healthContribution(healthType);
            if (contrib.isPos()) {
                assets.iadd(contrib);
            }
        }
        return assets;
    }
    // An undefined HealthType will use an asset and liab weight of 1
    liabs(healthType) {
        const liabs = ZERO_I80F48();
        for (const tokenInfo of this.tokenInfos) {
            const contrib = tokenInfo.healthContribution(healthType);
            if (contrib.isNeg()) {
                liabs.isub(contrib);
            }
        }
        const res = this.computeSerum3Reservations(HealthType.maint);
        for (const [index, serum3Info] of this.serum3Infos.entries()) {
            const contrib = serum3Info.healthContribution(healthType, this.tokenInfos, res.tokenMaxReserved, res.serum3Reserved[index]);
            if (contrib.isNeg()) {
                liabs.isub(contrib);
            }
        }
        for (const perpInfo of this.perpInfos) {
            const contrib = perpInfo.healthContribution(healthType);
            if (contrib.isNeg()) {
                liabs.isub(contrib);
            }
        }
        return liabs;
    }
    healthRatio(healthType) {
        const assets = ZERO_I80F48();
        const liabs = ZERO_I80F48();
        for (const tokenInfo of this.tokenInfos) {
            const contrib = tokenInfo.healthContribution(healthType);
            // console.log(` - ti contrib ${contrib.toLocaleString()}`);
            if (contrib.isPos()) {
                assets.iadd(contrib);
            }
            else {
                liabs.isub(contrib);
            }
        }
        const res = this.computeSerum3Reservations(HealthType.maint);
        for (const [index, serum3Info] of this.serum3Infos.entries()) {
            const contrib = serum3Info.healthContribution(healthType, this.tokenInfos, res.tokenMaxReserved, res.serum3Reserved[index]);
            // console.log(` - si contrib ${contrib.toLocaleString()}`);
            if (contrib.isPos()) {
                assets.iadd(contrib);
            }
            else {
                liabs.isub(contrib);
            }
        }
        for (const perpInfo of this.perpInfos) {
            const contrib = perpInfo.healthContribution(healthType);
            // console.log(` - pi contrib ${contrib.toLocaleString()}`);
            if (contrib.isPos()) {
                assets.iadd(contrib);
            }
            else {
                liabs.isub(contrib);
            }
        }
        // console.log(
        //   ` - assets ${assets.toLocaleString()}, liabs ${liabs.toLocaleString()}`,
        // );
        if (liabs.gt(I80F48.fromNumber(0.001))) {
            return HUNDRED_I80F48().mul(assets.sub(liabs).div(liabs));
        }
        else {
            return MAX_I80F48();
        }
    }
    findTokenInfoIndex(tokenIndex) {
        return this.tokenInfos.findIndex((tokenInfo) => tokenInfo.tokenIndex === tokenIndex);
    }
    getOrCreateTokenInfoIndex(bank) {
        const index = this.findTokenInfoIndex(bank.tokenIndex);
        if (index == -1) {
            this.tokenInfos.push(TokenInfo.fromBank(bank));
        }
        return this.findTokenInfoIndex(bank.tokenIndex);
    }
    simHealthRatioWithTokenPositionChanges(group, nativeTokenChanges, healthType = HealthType.init) {
        const adjustedCache = lodash.cloneDeep(this);
        // HealthCache.logHealthCache('beforeChange', adjustedCache);
        for (const change of nativeTokenChanges) {
            const bank = group.getFirstBankByMint(change.mintPk);
            const changeIndex = adjustedCache.getOrCreateTokenInfoIndex(bank);
            // TODO: this will no longer work as easily because of the health weight changes
            adjustedCache.tokenInfos[changeIndex].balanceNative.iadd(change.nativeTokenAmount);
        }
        // HealthCache.logHealthCache('afterChange', adjustedCache);
        return adjustedCache.healthRatio(healthType);
    }
    findSerum3InfoIndex(marketIndex) {
        return this.serum3Infos.findIndex((serum3Info) => serum3Info.marketIndex === marketIndex);
    }
    getOrCreateSerum3InfoIndex(baseBank, quoteBank, serum3Market) {
        const index = this.findSerum3InfoIndex(serum3Market.marketIndex);
        const baseEntryIndex = this.getOrCreateTokenInfoIndex(baseBank);
        const quoteEntryIndex = this.getOrCreateTokenInfoIndex(quoteBank);
        if (index == -1) {
            this.serum3Infos.push(Serum3Info.emptyFromSerum3Market(serum3Market, baseEntryIndex, quoteEntryIndex));
        }
        return this.findSerum3InfoIndex(serum3Market.marketIndex);
    }
    adjustSerum3Reserved(baseBank, quoteBank, serum3Market, reservedBaseChange, freeBaseChange, reservedQuoteChange, freeQuoteChange) {
        const baseEntryIndex = this.getOrCreateTokenInfoIndex(baseBank);
        const quoteEntryIndex = this.getOrCreateTokenInfoIndex(quoteBank);
        const baseEntry = this.tokenInfos[baseEntryIndex];
        const quoteEntry = this.tokenInfos[quoteEntryIndex];
        // Apply it to the tokens
        baseEntry.balanceNative.iadd(freeBaseChange);
        quoteEntry.balanceNative.iadd(freeQuoteChange);
        // Apply it to the serum3 info
        const index = this.getOrCreateSerum3InfoIndex(baseBank, quoteBank, serum3Market);
        const serum3Info = this.serum3Infos[index];
        serum3Info.reservedBase.iadd(reservedBaseChange);
        serum3Info.reservedQuote.iadd(reservedQuoteChange);
    }
    simHealthRatioWithSerum3BidChanges(baseBank, quoteBank, bidNativeQuoteAmount, serum3Market, healthType = HealthType.init) {
        const adjustedCache = lodash.cloneDeep(this);
        const quoteIndex = adjustedCache.getOrCreateTokenInfoIndex(quoteBank);
        // Move token balance to reserved funds in open orders,
        // essentially simulating a place order
        // Reduce token balance for quote
        adjustedCache.tokenInfos[quoteIndex].balanceNative.isub(bidNativeQuoteAmount);
        // Increase reserved in Serum3Info for quote
        adjustedCache.adjustSerum3Reserved(baseBank, quoteBank, serum3Market, ZERO_I80F48(), ZERO_I80F48(), bidNativeQuoteAmount, ZERO_I80F48());
        return adjustedCache.healthRatio(healthType);
    }
    simHealthRatioWithSerum3AskChanges(baseBank, quoteBank, askNativeBaseAmount, serum3Market, healthType = HealthType.init) {
        const adjustedCache = lodash.cloneDeep(this);
        const baseIndex = adjustedCache.getOrCreateTokenInfoIndex(baseBank);
        // Move token balance to reserved funds in open orders,
        // essentially simulating a place order
        // Reduce token balance for base
        adjustedCache.tokenInfos[baseIndex].balanceNative.isub(askNativeBaseAmount);
        // Increase reserved in Serum3Info for base
        adjustedCache.adjustSerum3Reserved(baseBank, quoteBank, serum3Market, askNativeBaseAmount, ZERO_I80F48(), ZERO_I80F48(), ZERO_I80F48());
        return adjustedCache.healthRatio(healthType);
    }
    findPerpInfoIndex(perpMarketIndex) {
        return this.perpInfos.findIndex((perpInfo) => perpInfo.perpMarketIndex === perpMarketIndex);
    }
    getOrCreatePerpInfoIndex(perpMarket) {
        const index = this.findPerpInfoIndex(perpMarket.perpMarketIndex);
        if (index == -1) {
            this.perpInfos.push(PerpInfo.emptyFromPerpMarket(perpMarket));
        }
        return this.findPerpInfoIndex(perpMarket.perpMarketIndex);
    }
    adjustPerpInfo(perpInfoIndex, price, side, newOrderBaseLots) {
        if (side == PerpOrderSide.bid) {
            this.perpInfos[perpInfoIndex].baseLots.iadd(newOrderBaseLots);
            this.perpInfos[perpInfoIndex].quote.isub(I80F48.fromI64(newOrderBaseLots)
                .mul(I80F48.fromI64(this.perpInfos[perpInfoIndex].baseLotSize))
                .mul(price));
        }
        else {
            this.perpInfos[perpInfoIndex].baseLots.isub(newOrderBaseLots);
            this.perpInfos[perpInfoIndex].quote.iadd(I80F48.fromI64(newOrderBaseLots)
                .mul(I80F48.fromI64(this.perpInfos[perpInfoIndex].baseLotSize))
                .mul(price));
        }
    }
    simHealthRatioWithPerpOrderChanges(perpMarket, existingPerpPosition, side, baseLots, price, healthType = HealthType.init) {
        const clonedHealthCache = lodash.cloneDeep(this);
        const perpInfoIndex = clonedHealthCache.getOrCreatePerpInfoIndex(perpMarket);
        clonedHealthCache.adjustPerpInfo(perpInfoIndex, price, side, baseLots);
        return clonedHealthCache.healthRatio(healthType);
    }
    logHealthCache(debug) {
        if (debug)
            console.log(debug);
        for (const token of this.tokenInfos) {
            console.log(` ${token.toString()}`);
        }
        const res = this.computeSerum3Reservations(HealthType.maint);
        for (const [index, serum3Info] of this.serum3Infos.entries()) {
            console.log(` ${serum3Info.toString(this.tokenInfos, res.tokenMaxReserved, res.serum3Reserved[index])}`);
        }
        console.log(` assets ${this.assets(HealthType.init)}, liabs ${this.liabs(HealthType.init)}, `);
        console.log(` health(HealthType.init) ${this.health(HealthType.init)}`);
        console.log(` healthRatio(HealthType.init) ${this.healthRatio(HealthType.init)}`);
    }
    static scanRightUntilLessThan(start, target, fun) {
        const maxIterations = 20;
        let current = start;
        // console.log(`scanRightUntilLessThan, start ${start.toLocaleString()}`);
        for (const key of Array(maxIterations).fill(0).keys()) {
            const value = fun(current);
            if (value.lt(target)) {
                return current;
            }
            // console.log(
            //   ` - current ${current.toLocaleString()}, value ${value.toLocaleString()}, target ${target.toLocaleString()}`,
            // );
            current = current.max(ONE_I80F48()).mul(I80F48.fromNumber(2));
        }
        throw new Error('Could not find amount that led to health ratio <=0');
    }
    /// This is not a generic function. It assumes there is a unique maximum between left and right.
    static findMaximum(left, right, minStep, fun) {
        const half = I80F48.fromNumber(0.5);
        let mid = half.mul(left.add(right));
        let leftValue = fun(left);
        let rightValue = fun(right);
        let midValue = fun(mid);
        while (right.sub(left).gt(minStep)) {
            if (leftValue.gte(midValue)) {
                // max must be between left and mid
                right = mid;
                rightValue = midValue;
                mid = half.mul(left.add(mid));
                midValue = fun(mid);
            }
            else if (midValue.lte(rightValue)) {
                // max must be between mid and right
                left = mid;
                leftValue = midValue;
                mid = half.mul(mid.add(right));
                midValue = fun(mid);
            }
            else {
                // mid is larger than both left and right, max could be on either side
                const leftmid = half.mul(left.add(mid));
                const leftMidValue = fun(leftmid);
                if (leftMidValue.gte(midValue)) {
                    // max between left and mid
                    right = mid;
                    rightValue = midValue;
                    mid = leftmid;
                    midValue = leftMidValue;
                    continue;
                }
                const rightmid = half.mul(mid.add(right));
                const rightMidValue = fun(rightmid);
                if (rightMidValue.gte(midValue)) {
                    // max between mid and right
                    left = mid;
                    leftValue = midValue;
                    mid = rightmid;
                    midValue = rightMidValue;
                    continue;
                }
                // max between leftmid and rightmid
                left = leftmid;
                leftValue = leftMidValue;
                right = rightmid;
                rightValue = rightMidValue;
            }
        }
        if (leftValue.gte(midValue)) {
            return [left, leftValue];
        }
        else if (midValue.gte(rightValue)) {
            return [mid, midValue];
        }
        else {
            return [right, rightValue];
        }
    }
    static binaryApproximationSearch(left, leftValue, right, targetValue, minStep, fun) {
        const maxIterations = 50;
        const targetError = I80F48.fromNumber(0.1);
        const rightValue = fun(right);
        // console.log(
        //   ` - binaryApproximationSearch left ${left.toLocaleString()}, leftValue ${leftValue.toLocaleString()}, right ${right.toLocaleString()}, rightValue ${rightValue.toLocaleString()}, targetValue ${targetValue.toLocaleString()}`,
        // );
        if ((leftValue.sub(targetValue).isPos() &&
            rightValue.sub(targetValue).isPos()) ||
            (leftValue.sub(targetValue).isNeg() &&
                rightValue.sub(targetValue).isNeg())) {
            throw new Error(`Internal error: left ${leftValue.toNumber()}  and right ${rightValue.toNumber()} don't contain the target value ${targetValue.toNumber()}!`);
        }
        let newAmount, newAmountValue;
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        for (const key of Array(maxIterations).fill(0).keys()) {
            if (right.sub(left).abs().lt(minStep)) {
                return left;
            }
            newAmount = left.add(right).mul(I80F48.fromNumber(0.5));
            newAmountValue = fun(newAmount);
            // console.log(
            //   `   - left ${left.toLocaleString()}, right ${right.toLocaleString()}, newAmount ${newAmount.toLocaleString()}, newAmountValue ${newAmountValue.toLocaleString()}, targetValue ${targetValue.toLocaleString()}`,
            // );
            const error = newAmountValue.sub(targetValue);
            if (error.isPos() && error.lt(targetError)) {
                return newAmount;
            }
            if (newAmountValue.gt(targetValue) != rightValue.gt(targetValue)) {
                left = newAmount;
            }
            else {
                right = newAmount;
            }
        }
        console.error(`Unable to get targetValue within ${maxIterations} iterations, newAmount ${newAmount}, newAmountValue ${newAmountValue}, target ${targetValue}`);
        return newAmount;
    }
    getMaxSwapSource(sourceBank, targetBank, price) {
        const health = this.health(HealthType.init);
        if (health.isNeg()) {
            return this.getMaxSwapSourceForHealth(sourceBank, targetBank, price, toNativeI80F48ForQuote(1));
        }
        return this.getMaxSwapSourceForHealthRatio(sourceBank, targetBank, price, I80F48.fromNumber(2));
    }
    getMaxSwapSourceForHealthRatio(sourceBank, targetBank, price, minRatio) {
        return this.getMaxSwapSourceForHealthFn(sourceBank, targetBank, price, minRatio, function (hc) {
            return hc.healthRatio(HealthType.init);
        });
    }
    getMaxSwapSourceForHealth(sourceBank, targetBank, price, minHealth) {
        return this.getMaxSwapSourceForHealthFn(sourceBank, targetBank, price, minHealth, function (hc) {
            return hc.health(HealthType.init);
        });
    }
    getMaxSwapSourceForHealthFn(sourceBank, targetBank, price, minFnValue, targetFn) {
        if (sourceBank.initLiabWeight
            .sub(targetBank.initAssetWeight)
            .abs()
            .lte(ZERO_I80F48())) {
            return ZERO_I80F48();
        }
        // The health and health_ratio are nonlinear based on swap amount.
        // For large swap amounts the slope is guaranteed to be negative, but small amounts
        // can have positive slope (e.g. using source deposits to pay back target borrows).
        //
        // That means:
        // - even if the initial value is < minRatio it can be useful to swap to *increase* health
        // - even if initial value is < 0, swapping can increase health (maybe above 0)
        // - be careful about finding the minFnValue: the function isn't convex
        this.healthRatio(HealthType.init);
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        const healthCacheClone = lodash.cloneDeep(this);
        const sourceIndex = healthCacheClone.getOrCreateTokenInfoIndex(sourceBank);
        const targetIndex = healthCacheClone.getOrCreateTokenInfoIndex(targetBank);
        const source = healthCacheClone.tokenInfos[sourceIndex];
        const target = healthCacheClone.tokenInfos[targetIndex];
        const res = healthCacheClone.computeSerum3Reservations(HealthType.init);
        const sourceReserved = res.tokenMaxReserved[sourceIndex];
        const targetReserved = res.tokenMaxReserved[targetIndex];
        // If the price is sufficiently good, then health will just increase from swapping:
        // once we've swapped enough, swapping x reduces health by x * source_liab_weight and
        // increases it by x * target_asset_weight * price_factor.
        const finalHealthSlope = source.initLiabWeight
            .neg()
            .mul(source.prices.liab(HealthType.init))
            .add(target.initAssetWeight
            .mul(target.prices.asset(HealthType.init))
            .mul(price));
        if (finalHealthSlope.gte(ZERO_I80F48())) {
            return MAX_I80F48();
        }
        // There are two key slope changes: Assume source.balance > 0 and target.balance < 0. Then
        // initially health ratio goes up. When one of balances flips sign, the health ratio slope
        // may be positive or negative for a bit, until both balances have flipped and the slope is
        // negative.
        // The maximum will be at one of these points (ignoring serum3 effects).
        function cacheAfterSwap(amount) {
            const adjustedCache = lodash.cloneDeep(healthCacheClone);
            // adjustedCache.logHealthCache('beforeSwap', adjustedCache);
            // TODO: make a copy of the bank, apply amount, recompute weights,
            // and set the new weights on the tokenInfos
            adjustedCache.tokenInfos[sourceIndex].balanceNative.isub(amount);
            adjustedCache.tokenInfos[targetIndex].balanceNative.iadd(amount.mul(price));
            // adjustedCache.logHealthCache('afterSwap', adjustedCache);
            return adjustedCache;
        }
        function fnValueAfterSwap(amount) {
            return targetFn(cacheAfterSwap(amount));
        }
        // The function we're looking at has a unique maximum.
        //
        // If we discount serum3 reservations, there are two key slope changes:
        // Assume source.balance > 0 and target.balance < 0.
        // When these values flip sign, the health slope decreases, but could still be positive.
        //
        // The first thing we do is to find this maximum.
        // The largest amount that the maximum could be at
        const rightmost = source.balanceNative
            .abs()
            .add(sourceReserved)
            .max(target.balanceNative.abs().add(targetReserved).div(price));
        const [amountForMaxValue, maxValue] = HealthCache.findMaximum(ZERO_I80F48(), rightmost, I80F48.fromNumber(0.1), fnValueAfterSwap);
        if (maxValue.lte(minFnValue)) {
            // We cannot reach min_ratio, just return the max
            return amountForMaxValue;
        }
        let amount;
        // Now max_value is bigger than minFnValue, the target amount must be >amountForMaxValue.
        // Search to the right of amountForMaxValue: but how far?
        // Use a simple estimation for the amount that would lead to zero health:
        //           health
        //              - source_liab_weight * source_liab_price * a
        //              + target_asset_weight * target_asset_price * price * a = 0.
        // where a is the source token native amount.
        // Note that this is just an estimate. Swapping can increase the amount that serum3
        // reserved contributions offset, moving the actual zero point further to the right.
        const healthAtMaxValue = cacheAfterSwap(amountForMaxValue).health(HealthType.init);
        if (healthAtMaxValue.lte(ZERO_I80F48())) {
            return ZERO_I80F48();
        }
        const zeroHealthEstimate = amountForMaxValue.sub(healthAtMaxValue.div(finalHealthSlope));
        const rightBound = HealthCache.scanRightUntilLessThan(zeroHealthEstimate, minFnValue, fnValueAfterSwap);
        if (rightBound.eq(zeroHealthEstimate)) {
            amount = HealthCache.binaryApproximationSearch(amountForMaxValue, maxValue, rightBound, minFnValue, I80F48.fromNumber(0.1), fnValueAfterSwap);
        }
        else {
            // Must be between 0 and point0_amount
            amount = HealthCache.binaryApproximationSearch(zeroHealthEstimate, fnValueAfterSwap(zeroHealthEstimate), rightBound, minFnValue, I80F48.fromNumber(0.1), fnValueAfterSwap);
        }
        return amount;
    }
    getMaxSerum3OrderForHealthRatio(baseBank, quoteBank, serum3Market, side, minRatio) {
        const healthCacheClone = lodash.cloneDeep(this);
        const baseIndex = healthCacheClone.getOrCreateTokenInfoIndex(baseBank);
        const quoteIndex = healthCacheClone.getOrCreateTokenInfoIndex(quoteBank);
        const base = healthCacheClone.tokenInfos[baseIndex];
        const quote = healthCacheClone.tokenInfos[quoteIndex];
        // Binary search between current health (0 sized new order) and
        // an amount to trade which will bring health to 0.
        // Current health and amount i.e. 0
        const initialAmount = ZERO_I80F48();
        const initialHealth = this.health(HealthType.init);
        const initialRatio = this.healthRatio(HealthType.init);
        if (initialRatio.lte(ZERO_I80F48())) {
            return ZERO_I80F48();
        }
        // console.log(`getMaxSerum3OrderForHealthRatio`);
        // Amount which would bring health to 0
        // where M = max(A_deposits, B_borrows)
        // amount = M + (init_health + M * (B_init_liab - A_init_asset)) / (A_init_liab - B_init_asset);
        // A is what we would be essentially swapping for B
        // So when its an ask, then base->quote,
        // and when its a bid, then quote->bid
        let zeroAmount;
        if (side == Serum3Side.ask) {
            const quoteBorrows = quote.balanceNative.lt(ZERO_I80F48())
                ? quote.balanceNative.abs().mul(quote.prices.liab(HealthType.init))
                : ZERO_I80F48();
            const max = base.balanceNative
                .mul(base.prices.asset(HealthType.init))
                .max(quoteBorrows);
            zeroAmount = max.add(initialHealth
                .add(max.mul(quote.initLiabWeight.sub(base.initAssetWeight)))
                .div(base
                .liabWeight(HealthType.init)
                .sub(quote.assetWeight(HealthType.init))));
            // console.log(` - quoteBorrows ${quoteBorrows.toLocaleString()}`);
            // console.log(` - max ${max.toLocaleString()}`);
        }
        else {
            const baseBorrows = base.balanceNative.lt(ZERO_I80F48())
                ? base.balanceNative.abs().mul(base.prices.liab(HealthType.init))
                : ZERO_I80F48();
            const max = quote.balanceNative
                .mul(quote.prices.asset(HealthType.init))
                .max(baseBorrows);
            zeroAmount = max.add(initialHealth
                .add(max.mul(base.initLiabWeight.sub(quote.initAssetWeight)))
                .div(quote
                .liabWeight(HealthType.init)
                .sub(base.assetWeight(HealthType.init))));
            // console.log(` - baseBorrows ${baseBorrows.toLocaleString()}`);
            // console.log(` - max ${max.toLocaleString()}`);
        }
        const cache = cacheAfterPlacingOrder(zeroAmount);
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        cache.health(HealthType.init);
        cache.healthRatio(HealthType.init);
        // console.log(` - zeroAmount ${zeroAmount.toLocaleString()}`);
        // console.log(` - zeroAmountHealth ${zeroAmountHealth.toLocaleString()}`);
        // console.log(` - zeroAmountRatio ${zeroAmountRatio.toLocaleString()}`);
        function cacheAfterPlacingOrder(amount) {
            const adjustedCache = lodash.cloneDeep(healthCacheClone);
            // adjustedCache.logHealthCache(` before placing order ${amount}`);
            // TODO: there should also be some issue with oracle vs stable price here;
            // probably better to pass in not the quote amount but the base or quote native amount
            side === Serum3Side.ask
                ? adjustedCache.tokenInfos[baseIndex].balanceNative.isub(amount.div(base.prices.oracle))
                : adjustedCache.tokenInfos[quoteIndex].balanceNative.isub(amount.div(quote.prices.oracle));
            adjustedCache.adjustSerum3Reserved(baseBank, quoteBank, serum3Market, side === Serum3Side.ask
                ? amount.div(base.prices.oracle)
                : ZERO_I80F48(), ZERO_I80F48(), side === Serum3Side.bid
                ? amount.div(quote.prices.oracle)
                : ZERO_I80F48(), ZERO_I80F48());
            // adjustedCache.logHealthCache(' after placing order');
            return adjustedCache;
        }
        function healthRatioAfterPlacingOrder(amount) {
            return cacheAfterPlacingOrder(amount).healthRatio(HealthType.init);
        }
        const amount = HealthCache.binaryApproximationSearch(initialAmount, initialRatio, zeroAmount, minRatio, ONE_I80F48(), healthRatioAfterPlacingOrder);
        return amount;
    }
    getMaxPerpForHealthRatio(perpMarket, price, side, minRatio) {
        const healthCacheClone = lodash.cloneDeep(this);
        const initialRatio = this.healthRatio(HealthType.init);
        if (initialRatio.lt(ZERO_I80F48())) {
            return ZERO_I80F48();
        }
        const direction = side == PerpOrderSide.bid ? 1 : -1;
        const perpInfoIndex = healthCacheClone.getOrCreatePerpInfoIndex(perpMarket);
        const perpInfo = healthCacheClone.perpInfos[perpInfoIndex];
        const prices = perpInfo.prices;
        const baseLotSize = I80F48.fromI64(perpMarket.baseLotSize);
        // If the price is sufficiently good then health will just increase from trading
        const finalHealthSlope = direction == 1
            ? perpInfo.initBaseAssetWeight
                .mul(prices.asset(HealthType.init))
                .sub(price)
            : price.sub(perpInfo.initBaseLiabWeight.mul(prices.liab(HealthType.init)));
        if (finalHealthSlope.gte(ZERO_I80F48())) {
            return MAX_I80F48();
        }
        function cacheAfterTrade(baseLots) {
            const adjustedCache = lodash.cloneDeep(healthCacheClone);
            // adjustedCache.logHealthCache(' -- before trade');
            adjustedCache.adjustPerpInfo(perpInfoIndex, price, side, baseLots);
            // adjustedCache.logHealthCache(' -- after trade');
            return adjustedCache;
        }
        function healthRatioAfterTrade(baseLots) {
            return cacheAfterTrade(new anchor.BN(baseLots.toNumber())).healthRatio(HealthType.init);
        }
        function healthRatioAfterTradeTrunc(baseLots) {
            return healthRatioAfterTrade(baseLots.floor());
        }
        const initialBaseLots = I80F48.fromU64(perpInfo.baseLots);
        // There are two cases:
        // 1. We are increasing abs(baseLots)
        // 2. We are bringing the base position to 0, and then going to case 1.
        const hasCase2 = (initialBaseLots.gt(ZERO_I80F48()) && direction == -1) ||
            (initialBaseLots.lt(ZERO_I80F48()) && direction == 1);
        let case1Start, case1StartRatio;
        if (hasCase2) {
            case1Start = initialBaseLots.abs();
            case1StartRatio = healthRatioAfterTrade(case1Start);
        }
        else {
            case1Start = ZERO_I80F48();
            case1StartRatio = initialRatio;
        }
        // If we start out below minRatio and can't go above, pick the best case
        let baseLots;
        if (initialRatio.lte(minRatio) && case1StartRatio.lt(minRatio)) {
            if (case1StartRatio.gte(initialRatio)) {
                baseLots = case1Start;
            }
            else {
                baseLots = ZERO_I80F48();
            }
        }
        else if (case1StartRatio.gte(minRatio)) {
            // Must reach minRatio to the right of case1Start
            // Need to figure out how many lots to trade to reach zero health (zero_health_amount).
            // We do this by looking at the starting health and the health slope per
            // traded base lot (finalHealthSlope).
            const startCache = cacheAfterTrade(new anchor.BN(case1Start.toNumber()));
            const startHealth = startCache.health(HealthType.init);
            if (startHealth.lte(ZERO_I80F48())) {
                return ZERO_I80F48();
            }
            // The perp market's contribution to the health above may be capped. But we need to trade
            // enough to fully reduce any positive-pnl buffer. Thus get the uncapped health:
            const perpInfo = startCache.perpInfos[perpInfoIndex];
            const startHealthUncapped = startHealth
                .sub(perpInfo.healthContribution(HealthType.init))
                .add(perpInfo.unweightedHealthContribution(HealthType.init));
            const zeroHealthAmount = case1Start
                .sub(startHealthUncapped.div(finalHealthSlope).div(baseLotSize))
                .add(ONE_I80F48());
            healthRatioAfterTradeTrunc(zeroHealthAmount);
            baseLots = HealthCache.binaryApproximationSearch(case1Start, case1StartRatio, zeroHealthAmount, minRatio, ONE_I80F48(), healthRatioAfterTradeTrunc);
        }
        else {
            // Between 0 and case1Start
            baseLots = HealthCache.binaryApproximationSearch(ZERO_I80F48(), initialRatio, case1Start, minRatio, ONE_I80F48(), healthRatioAfterTradeTrunc);
        }
        return baseLots.floor();
    }
}
class Prices {
    oracle;
    stable;
    constructor(oracle, stable) {
        this.oracle = oracle;
        this.stable = stable;
    }
    liab(healthType) {
        if (healthType === HealthType.maint ||
            healthType === HealthType.liquidationEnd ||
            healthType === undefined) {
            return this.oracle;
        }
        return this.oracle.max(this.stable);
    }
    asset(healthType) {
        if (healthType === HealthType.maint ||
            healthType === HealthType.liquidationEnd ||
            healthType === undefined) {
            return this.oracle;
        }
        return this.oracle.min(this.stable);
    }
}
class TokenInfo {
    tokenIndex;
    maintAssetWeight;
    initAssetWeight;
    initScaledAssetWeight;
    maintLiabWeight;
    initLiabWeight;
    initScaledLiabWeight;
    prices;
    balanceNative;
    constructor(tokenIndex, maintAssetWeight, initAssetWeight, initScaledAssetWeight, maintLiabWeight, initLiabWeight, initScaledLiabWeight, prices, balanceNative) {
        this.tokenIndex = tokenIndex;
        this.maintAssetWeight = maintAssetWeight;
        this.initAssetWeight = initAssetWeight;
        this.initScaledAssetWeight = initScaledAssetWeight;
        this.maintLiabWeight = maintLiabWeight;
        this.initLiabWeight = initLiabWeight;
        this.initScaledLiabWeight = initScaledLiabWeight;
        this.prices = prices;
        this.balanceNative = balanceNative;
    }
    static fromDto(dto) {
        return new TokenInfo(dto.tokenIndex, I80F48.from(dto.maintAssetWeight), I80F48.from(dto.initAssetWeight), I80F48.from(dto.initScaledAssetWeight), I80F48.from(dto.maintLiabWeight), I80F48.from(dto.initLiabWeight), I80F48.from(dto.initScaledLiabWeight), new Prices(I80F48.from(dto.prices.oracle), I80F48.from(dto.prices.stable)), I80F48.from(dto.balanceNative));
    }
    static fromBank(bank, nativeBalance) {
        const p = new Prices(bank.price, I80F48.fromNumber(bank.stablePriceModel.stablePrice));
        // Use the liab price for computing weight scaling, because it's pessimistic and
        // causes the most unfavorable scaling.
        const liabPrice = p.liab(HealthType.init);
        return new TokenInfo(bank.tokenIndex, bank.maintAssetWeight, bank.initAssetWeight, bank.scaledInitAssetWeight(liabPrice), bank.maintLiabWeight, bank.initLiabWeight, bank.scaledInitLiabWeight(liabPrice), p, nativeBalance ? nativeBalance : ZERO_I80F48());
    }
    assetWeight(healthType) {
        if (healthType == HealthType.init) {
            return this.initScaledAssetWeight;
        }
        else if (healthType == HealthType.liquidationEnd) {
            return this.initAssetWeight;
        }
        // healthType == HealthType.maint
        return this.maintAssetWeight;
    }
    liabWeight(healthType) {
        if (healthType == HealthType.init) {
            return this.initScaledLiabWeight;
        }
        else if (healthType == HealthType.liquidationEnd) {
            return this.initLiabWeight;
        }
        // healthType == HealthType.maint
        return this.maintLiabWeight;
    }
    healthContribution(healthType) {
        let weight, price;
        if (healthType === undefined) {
            return this.balanceNative.mul(this.prices.oracle);
        }
        if (this.balanceNative.isNeg()) {
            weight = this.liabWeight(healthType);
            price = this.prices.liab(healthType);
        }
        else {
            weight = this.assetWeight(healthType);
            price = this.prices.asset(healthType);
        }
        return this.balanceNative.mul(weight).mul(price);
    }
    toString() {
        return `  tokenIndex: ${this.tokenIndex}, balanceNative: ${this.balanceNative}, initHealth ${this.healthContribution(HealthType.init)}`;
    }
}
class Serum3Reserved {
    allReservedAsBase;
    allReservedAsQuote;
    constructor(allReservedAsBase, allReservedAsQuote) {
        this.allReservedAsBase = allReservedAsBase;
        this.allReservedAsQuote = allReservedAsQuote;
    }
}
class Serum3Info {
    reservedBase;
    reservedQuote;
    baseIndex;
    quoteIndex;
    marketIndex;
    constructor(reservedBase, reservedQuote, baseIndex, quoteIndex, marketIndex) {
        this.reservedBase = reservedBase;
        this.reservedQuote = reservedQuote;
        this.baseIndex = baseIndex;
        this.quoteIndex = quoteIndex;
        this.marketIndex = marketIndex;
    }
    static fromDto(dto) {
        return new Serum3Info(I80F48.from(dto.reservedBase), I80F48.from(dto.reservedQuote), dto.baseIndex, dto.quoteIndex, dto.marketIndex);
    }
    static emptyFromSerum3Market(serum3Market, baseEntryIndex, quoteEntryIndex) {
        return new Serum3Info(ZERO_I80F48(), ZERO_I80F48(), baseEntryIndex, quoteEntryIndex, serum3Market.marketIndex);
    }
    static fromOoModifyingTokenInfos(baseIndex, baseInfo, quoteIndex, quoteInfo, marketIndex, oo) {
        // add the amounts that are freely settleable immediately to token balances
        const baseFree = I80F48.fromI64(oo.baseTokenFree);
        // NOTE: referrerRebatesAccrued is not declared on oo class, but the layout
        // is aware of it
        const quoteFree = I80F48.fromI64(oo.quoteTokenFree.add(oo.referrerRebatesAccrued));
        baseInfo.balanceNative.iadd(baseFree);
        quoteInfo.balanceNative.iadd(quoteFree);
        // track the reserved amounts
        const reservedBase = I80F48.fromI64(oo.baseTokenTotal.sub(oo.baseTokenFree));
        const reservedQuote = I80F48.fromI64(oo.quoteTokenTotal.sub(oo.quoteTokenFree));
        return new Serum3Info(reservedBase, reservedQuote, baseIndex, quoteIndex, marketIndex);
    }
    // An undefined HealthType will use an asset and liab weight of 1
    healthContribution(healthType, tokenInfos, tokenMaxReserved, marketReserved) {
        if (marketReserved.allReservedAsBase.isZero() ||
            marketReserved.allReservedAsQuote.isZero()) {
            return ZERO_I80F48();
        }
        const baseInfo = tokenInfos[this.baseIndex];
        const quoteInfo = tokenInfos[this.quoteIndex];
        const baseMaxReserved = tokenMaxReserved[this.baseIndex];
        const quoteMaxReserved = tokenMaxReserved[this.quoteIndex];
        // How much the health would increase if the reserved balance were applied to the passed
        // token info?
        const computeHealthEffect = function (tokenInfo, tokenMaxReserved, marketReserved) {
            // This balance includes all possible reserved funds from markets that relate to the
            // token, including this market itself: `tokenMaxReserved` is already included in `maxBalance`.
            const maxBalance = tokenInfo.balanceNative.add(tokenMaxReserved);
            // Assuming `reserved` was added to `max_balance` last (because that gives the smallest
            // health effects): how much did health change because of it?
            let assetPart, liabPart;
            if (maxBalance.gte(marketReserved)) {
                assetPart = marketReserved;
                liabPart = ZERO_I80F48();
            }
            else if (maxBalance.isNeg()) {
                assetPart = ZERO_I80F48();
                liabPart = marketReserved;
            }
            else {
                assetPart = maxBalance;
                liabPart = marketReserved.sub(maxBalance);
            }
            if (healthType === undefined) {
                return assetPart
                    .mul(tokenInfo.prices.oracle)
                    .add(liabPart.mul(tokenInfo.prices.oracle));
            }
            const assetWeight = tokenInfo.assetWeight(healthType);
            const liabWeight = tokenInfo.liabWeight(healthType);
            const assetPrice = tokenInfo.prices.asset(healthType);
            const liabPrice = tokenInfo.prices.liab(healthType);
            return assetWeight
                .mul(assetPart)
                .mul(assetPrice)
                .add(liabWeight.mul(liabPart).mul(liabPrice));
        };
        const healthBase = computeHealthEffect(baseInfo, baseMaxReserved, marketReserved.allReservedAsBase);
        const healthQuote = computeHealthEffect(quoteInfo, quoteMaxReserved, marketReserved.allReservedAsQuote);
        // console.log(` - healthBase ${healthBase.toLocaleString()}`);
        // console.log(` - healthQuote ${healthQuote.toLocaleString()}`);
        return healthBase.min(healthQuote);
    }
    toString(tokenInfos, tokenMaxReserved, marketReserved) {
        return `  marketIndex: ${this.marketIndex}, baseIndex: ${this.baseIndex}, quoteIndex: ${this.quoteIndex}, reservedBase: ${this.reservedBase}, reservedQuote: ${this.reservedQuote}, initHealth ${this.healthContribution(HealthType.init, tokenInfos, tokenMaxReserved, marketReserved)}`;
    }
}
class PerpInfo {
    perpMarketIndex;
    maintBaseAssetWeight;
    initBaseAssetWeight;
    maintBaseLiabWeight;
    initBaseLiabWeight;
    maintOverallAssetWeight;
    initOverallAssetWeight;
    baseLotSize;
    baseLots;
    bidsBaseLots;
    asksBaseLots;
    quote;
    prices;
    hasOpenOrders;
    constructor(perpMarketIndex, maintBaseAssetWeight, initBaseAssetWeight, maintBaseLiabWeight, initBaseLiabWeight, maintOverallAssetWeight, initOverallAssetWeight, baseLotSize, baseLots, bidsBaseLots, asksBaseLots, quote, prices, hasOpenOrders) {
        this.perpMarketIndex = perpMarketIndex;
        this.maintBaseAssetWeight = maintBaseAssetWeight;
        this.initBaseAssetWeight = initBaseAssetWeight;
        this.maintBaseLiabWeight = maintBaseLiabWeight;
        this.initBaseLiabWeight = initBaseLiabWeight;
        this.maintOverallAssetWeight = maintOverallAssetWeight;
        this.initOverallAssetWeight = initOverallAssetWeight;
        this.baseLotSize = baseLotSize;
        this.baseLots = baseLots;
        this.bidsBaseLots = bidsBaseLots;
        this.asksBaseLots = asksBaseLots;
        this.quote = quote;
        this.prices = prices;
        this.hasOpenOrders = hasOpenOrders;
    }
    static fromDto(dto) {
        return new PerpInfo(dto.perpMarketIndex, I80F48.from(dto.maintBaseAssetWeight), I80F48.from(dto.initBaseAssetWeight), I80F48.from(dto.maintBaseLiabWeight), I80F48.from(dto.initBaseLiabWeight), I80F48.from(dto.maintOverallAssetWeight), I80F48.from(dto.initOverallAssetWeight), dto.baseLotSize, dto.baseLots, dto.bidsBaseLots, dto.asksBaseLots, I80F48.from(dto.quote), new Prices(I80F48.from(dto.prices.oracle), I80F48.from(dto.prices.stable)), dto.hasOpenOrders);
    }
    static fromPerpPosition(perpMarket, perpPosition) {
        const baseLots = perpPosition.basePositionLots.add(perpPosition.takerBaseLots);
        const unsettledFunding = perpPosition.getUnsettledFunding(perpMarket);
        const takerQuote = I80F48.fromI64(new anchor.BN(perpPosition.takerQuoteLots).mul(perpMarket.quoteLotSize));
        const quoteCurrent = perpPosition.quotePositionNative
            .sub(unsettledFunding)
            .add(takerQuote);
        return new PerpInfo(perpMarket.perpMarketIndex, perpMarket.maintBaseAssetWeight, perpMarket.initBaseAssetWeight, perpMarket.maintBaseLiabWeight, perpMarket.initBaseLiabWeight, perpMarket.maintOverallAssetWeight, perpMarket.initOverallAssetWeight, perpMarket.baseLotSize, baseLots, perpPosition.bidsBaseLots, perpPosition.asksBaseLots, quoteCurrent, new Prices(perpMarket.price, I80F48.fromNumber(perpMarket.stablePriceModel.stablePrice)), perpPosition.hasOpenOrders());
    }
    healthContribution(healthType) {
        const contrib = this.unweightedHealthContribution(healthType);
        if (contrib.gt(ZERO_I80F48())) {
            const assetWeight = healthType == HealthType.init || healthType == HealthType.liquidationEnd
                ? this.initOverallAssetWeight
                : this.maintOverallAssetWeight;
            return assetWeight.mul(contrib);
        }
        return contrib;
    }
    unweightedHealthContribution(healthType) {
        function orderExecutionCase(pi, ordersBaseLots, orderPrice) {
            const netBaseNative = I80F48.fromU64(pi.baseLots.add(ordersBaseLots).mul(pi.baseLotSize));
            let weight, basePrice;
            if (healthType == HealthType.init ||
                healthType == HealthType.liquidationEnd) {
                if (netBaseNative.isNeg()) {
                    weight = pi.initBaseLiabWeight;
                }
                else {
                    weight = pi.initBaseAssetWeight;
                }
            }
            // healthType == HealthType.maint
            else {
                if (netBaseNative.isNeg()) {
                    weight = pi.maintBaseLiabWeight;
                }
                else {
                    weight = pi.maintBaseAssetWeight;
                }
            }
            if (netBaseNative.isNeg()) {
                basePrice = pi.prices.liab(healthType);
            }
            else {
                basePrice = pi.prices.asset(healthType);
            }
            // Total value of the order-execution adjusted base position
            const baseHealth = netBaseNative.mul(weight).mul(basePrice);
            const ordersBaseNative = I80F48.fromU64(ordersBaseLots.mul(pi.baseLotSize));
            // The quote change from executing the bids/asks
            const orderQuote = ordersBaseNative.neg().mul(orderPrice);
            return baseHealth.add(orderQuote);
        }
        // What is worse: Executing all bids at oracle_price.liab, or executing all asks at oracle_price.asset?
        const bidsCase = orderExecutionCase(this, this.bidsBaseLots, this.prices.liab(healthType));
        const asksCase = orderExecutionCase(this, this.asksBaseLots.neg(), this.prices.asset(healthType));
        const worstCase = bidsCase.min(asksCase);
        return this.quote.add(worstCase);
    }
    static emptyFromPerpMarket(perpMarket) {
        return new PerpInfo(perpMarket.perpMarketIndex, perpMarket.maintBaseAssetWeight, perpMarket.initBaseAssetWeight, perpMarket.maintBaseLiabWeight, perpMarket.initBaseLiabWeight, perpMarket.maintOverallAssetWeight, perpMarket.initOverallAssetWeight, perpMarket.baseLotSize, new anchor.BN(0), new anchor.BN(0), new anchor.BN(0), ZERO_I80F48(), new Prices(perpMarket.price, I80F48.fromNumber(perpMarket.stablePriceModel.stablePrice)), false);
    }
    toString() {
        return `  perpMarketIndex: ${this.perpMarketIndex}, base: ${this.baseLots}, quote: ${this.quote}, oraclePrice: ${this.prices.oracle}, uncapped health contribution ${this.unweightedHealthContribution(HealthType.init)}`;
    }
}

class MangoAccount {
    publicKey;
    group;
    owner;
    delegate;
    accountNum;
    beingLiquidated;
    inHealthRegion;
    netDeposits;
    perpSpotTransfers;
    healthRegionBeginInitHealth;
    frozenUntil;
    headerVersion;
    serum3OosMapByMarketIndex;
    name;
    tokens;
    serum3;
    perps;
    perpOpenOrders;
    static from(publicKey, obj) {
        return new MangoAccount(publicKey, obj.group, obj.owner, obj.name, obj.delegate, obj.accountNum, obj.beingLiquidated == 1, obj.inHealthRegion == 1, obj.netDeposits, obj.perpSpotTransfers, obj.healthRegionBeginInitHealth, obj.frozenUntil, obj.headerVersion, obj.tokens, obj.serum3, obj.perps, obj.perpOpenOrders, new Map());
    }
    constructor(publicKey, group, owner, name, delegate, accountNum, beingLiquidated, inHealthRegion, netDeposits, perpSpotTransfers, healthRegionBeginInitHealth, frozenUntil, headerVersion, tokens, serum3, perps, perpOpenOrders, serum3OosMapByMarketIndex) {
        this.publicKey = publicKey;
        this.group = group;
        this.owner = owner;
        this.delegate = delegate;
        this.accountNum = accountNum;
        this.beingLiquidated = beingLiquidated;
        this.inHealthRegion = inHealthRegion;
        this.netDeposits = netDeposits;
        this.perpSpotTransfers = perpSpotTransfers;
        this.healthRegionBeginInitHealth = healthRegionBeginInitHealth;
        this.frozenUntil = frozenUntil;
        this.headerVersion = headerVersion;
        this.serum3OosMapByMarketIndex = serum3OosMapByMarketIndex;
        this.name = bytes.utf8.decode(new Uint8Array(name)).split('\x00')[0];
        this.tokens = tokens.map((dto) => TokenPosition.from(dto));
        this.serum3 = serum3.map((dto) => Serum3Orders.from(dto));
        this.perps = perps.map((dto) => PerpPosition.from(dto));
        this.perpOpenOrders = perpOpenOrders.map((dto) => PerpOo.from(dto));
    }
    async reload(client) {
        const mangoAccount = await client.getMangoAccount(this);
        await mangoAccount.reloadSerum3OpenOrders(client);
        Object.assign(this, mangoAccount);
        return mangoAccount;
    }
    async reloadWithSlot(client) {
        const resp = await client.getMangoAccountWithSlot(this.publicKey);
        await resp?.value.reloadSerum3OpenOrders(client);
        Object.assign(this, resp?.value);
        return { value: resp.value, slot: resp.slot };
    }
    async reloadSerum3OpenOrders(client) {
        const serum3Active = this.serum3Active();
        const ais = await client.program.provider.connection.getMultipleAccountsInfo(serum3Active.map((serum3) => serum3.openOrders));
        this.serum3OosMapByMarketIndex = new Map(Array.from(ais.map((ai, i) => {
            if (!ai) {
                throw new Error(`Undefined AI for open orders ${serum3Active[i].openOrders} and market ${serum3Active[i].marketIndex}!`);
            }
            const oo = market.OpenOrders.fromAccountInfo(serum3Active[i].openOrders, ai, OPENBOOK_PROGRAM_ID[client.cluster]);
            return [serum3Active[i].marketIndex, oo];
        })));
        return this;
    }
    isDelegate(client) {
        return this.delegate.equals(client.program.provider.wallet.publicKey);
    }
    isOperational() {
        return this.frozenUntil.lt(new anchor.BN(Date.now() / 1000));
    }
    tokensActive() {
        return this.tokens.filter((token) => token.isActive());
    }
    serum3Active() {
        return this.serum3.filter((serum3) => serum3.isActive());
    }
    perpPositionExistsForMarket(perpMarket) {
        return this.perps.some((pp) => pp.isActive() && pp.marketIndex == perpMarket.perpMarketIndex);
    }
    perpOrderExistsForMarket(perpMarket) {
        return this.perpOpenOrders.some((poo) => poo.isActive() && poo.orderMarket == perpMarket.perpMarketIndex);
    }
    perpActive() {
        return this.perps.filter((perp) => perp.isActive());
    }
    perpOrdersActive() {
        return this.perpOpenOrders.filter((oo) => oo.orderMarket !== PerpOo.OrderMarketUnset);
    }
    getToken(tokenIndex) {
        return this.tokens.find((ta) => ta.tokenIndex == tokenIndex);
    }
    getSerum3Account(marketIndex) {
        return this.serum3.find((sa) => sa.marketIndex == marketIndex);
    }
    getPerpPosition(perpMarketIndex) {
        return this.perps.find((pp) => pp.marketIndex == perpMarketIndex);
    }
    getPerpPositionUi(group, perpMarketIndex, useEventQueue) {
        const pp = this.perps.find((pp) => pp.marketIndex == perpMarketIndex);
        if (!pp) {
            throw new Error(`No position found for PerpMarket ${perpMarketIndex}!`);
        }
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        return pp.getBasePositionUi(perpMarket, useEventQueue);
    }
    getSerum3OoAccount(marketIndex) {
        const oo = this.serum3OosMapByMarketIndex.get(marketIndex);
        if (!oo) {
            throw new Error(`Open orders account not loaded for market with marketIndex ${marketIndex}!`);
        }
        return oo;
    }
    // How to navigate
    // * if a function is returning a I80F48, then usually the return value is in native quote or native token, unless specified
    // * if a function is returning a number, then usually the return value is in ui tokens, unless specified
    // * functions try to be explicit by having native or ui in the name to better reflect the value
    // * some values might appear unexpected large or small, usually the doc contains a "note"
    /**
     *
     * @param bank
     * @returns native balance for a token, is signed
     */
    getTokenBalance(bank) {
        const tp = this.getToken(bank.tokenIndex);
        return tp ? tp.balance(bank) : ZERO_I80F48();
    }
    /**
     *
     * @param bank
     * @returns native deposits for a token, 0 if position has borrows
     */
    getTokenDeposits(bank) {
        const tp = this.getToken(bank.tokenIndex);
        return tp ? tp.deposits(bank) : ZERO_I80F48();
    }
    /**
     *
     * @param bank
     * @returns native borrows for a token, 0 if position has deposits
     */
    getTokenBorrows(bank) {
        const tp = this.getToken(bank.tokenIndex);
        return tp ? tp.borrows(bank) : ZERO_I80F48();
    }
    /**
     *
     * @param bank
     * @returns UI balance for a token, is signed
     */
    getTokenBalanceUi(bank) {
        const tp = this.getToken(bank.tokenIndex);
        return tp ? tp.balanceUi(bank) : 0;
    }
    /**
     *
     * @param bank
     * @returns UI deposits for a token, 0 or more
     */
    getTokenDepositsUi(bank) {
        const ta = this.getToken(bank.tokenIndex);
        return ta ? ta.depositsUi(bank) : 0;
    }
    /**
     *
     * @param bank
     * @returns UI borrows for a token, 0 or less
     */
    getTokenBorrowsUi(bank) {
        const ta = this.getToken(bank.tokenIndex);
        return ta ? ta.borrowsUi(bank) : 0;
    }
    /**
     * Health, see health.rs or https://docs.mango.markets/mango-markets/health-overview
     * @param healthType
     * @returns raw health number, in native quote
     */
    getHealth(group, healthType) {
        const hc = HealthCache.fromMangoAccount(group, this);
        return hc.health(healthType);
    }
    getPerpSettleHealth(group) {
        const hc = HealthCache.fromMangoAccount(group, this);
        return hc.perpSettleHealth();
    }
    /**
     * Health ratio, which is computed so `100 * (assets-liabs)/liabs`
     * Note: health ratio is technically ∞ if liabs are 0
     * @param healthType
     * @returns health ratio, in percentage form
     */
    getHealthRatio(group, healthType) {
        const hc = HealthCache.fromMangoAccount(group, this);
        return hc.healthRatio(healthType);
    }
    /**
     * Health ratio
     * @param healthType
     * @returns health ratio, in percentage form, capped to 100
     */
    getHealthRatioUi(group, healthType) {
        const ratio = this.getHealthRatio(group, healthType).toNumber();
        return ratio > 100 ? 100 : Math.trunc(ratio);
    }
    /**
     * Sum of all the assets i.e. token deposits, borrows, total assets in spot open orders, and perps positions.
     * @returns equity, in native quote
     */
    getEquity(group) {
        const tokensMap = new Map();
        for (const tp of this.tokensActive()) {
            const bank = group.getFirstBankByTokenIndex(tp.tokenIndex);
            tokensMap.set(tp.tokenIndex, tp.balance(bank).mul(bank.price));
        }
        for (const sp of this.serum3Active()) {
            const oo = this.getSerum3OoAccount(sp.marketIndex);
            const baseBank = group.getFirstBankByTokenIndex(sp.baseTokenIndex);
            tokensMap
                .get(baseBank.tokenIndex)
                .iadd(I80F48.fromI64(oo.baseTokenTotal).mul(baseBank.price));
            const quoteBank = group.getFirstBankByTokenIndex(sp.quoteTokenIndex);
            // NOTE: referrerRebatesAccrued is not declared on oo class, but the layout
            // is aware of it
            tokensMap
                .get(baseBank.tokenIndex)
                .iadd(I80F48.fromI64(oo.quoteTokenTotal.add(oo.referrerRebatesAccrued)).mul(quoteBank.price));
        }
        const tokenEquity = Array.from(tokensMap.values()).reduce((a, b) => a.add(b), ZERO_I80F48());
        const perpEquity = this.perpActive().reduce((a, b) => a.add(b.getEquity(group.getPerpMarketByMarketIndex(b.marketIndex))), ZERO_I80F48());
        return tokenEquity.add(perpEquity);
    }
    /**
     * The amount of native quote you could withdraw against your existing assets.
     * @returns collateral value, in native quote
     */
    getCollateralValue(group) {
        return this.getHealth(group, HealthType.init);
    }
    /**
     * Sum of all positive assets.
     * @returns assets, in native quote
     */
    getAssetsValue(group, healthType) {
        const hc = HealthCache.fromMangoAccount(group, this);
        return hc.assets(healthType);
    }
    /**
     * Sum of all negative assets.
     * @returns liabs, in native quote
     */
    getLiabsValue(group, healthType) {
        const hc = HealthCache.fromMangoAccount(group, this);
        return hc.liabs(healthType);
    }
    /**
     * @returns Overall PNL, in native quote
     * PNL is defined here as spot value + serum3 open orders value + perp value - net deposits value (evaluated at native quote price at the time of the deposit/withdraw)
     * spot value + serum3 open orders value + perp value is returned by getEquity (open orders values are added to spot token values implicitly)
     */
    getPnl(group) {
        return this.getEquity(group)?.add(I80F48.fromI64(this.netDeposits).mul(I80F48.fromNumber(-1)));
    }
    /**
     * @returns token cumulative interest, in native token units. Sum of deposit and borrow interest.
     * Caveat: This will only return cumulative interest since the tokenPosition was last opened.
     * If the tokenPosition was closed and reopened multiple times it is necessary to add this result to
     * cumulative interest at each of the prior tokenPosition closings (from mango API) to get the all time
     * cumulative interest.
     */
    getCumulativeInterest(bank) {
        const token = this.getToken(bank.tokenIndex);
        if (token === undefined) {
            // tokenPosition does not exist on mangoAccount so no cumulative interest
            return 0;
        }
        else {
            if (token.indexedPosition.isPos()) {
                const interest = bank.depositIndex
                    .sub(token.previousIndex)
                    .mul(token.indexedPosition)
                    .toNumber();
                return (interest +
                    token.cumulativeDepositInterest +
                    token.cumulativeBorrowInterest);
            }
            else {
                const interest = bank.borrowIndex
                    .sub(token.previousIndex)
                    .mul(token.indexedPosition)
                    .toNumber();
                return (interest +
                    token.cumulativeDepositInterest +
                    token.cumulativeBorrowInterest);
            }
        }
    }
    /**
     * The amount of given native token you can withdraw including borrows, considering all existing assets as collateral.
     * @returns amount of given native token you can borrow, considering all existing assets as collateral, in native token
     *
     * TODO: take into account net_borrow_limit and min_vault_to_deposits_ratio
     */
    getMaxWithdrawWithBorrowForToken(group, mintPk) {
        const tokenBank = group.getFirstBankByMint(mintPk);
        const initHealth = this.getHealth(group, HealthType.init);
        // Case 1:
        // Cannot withdraw if init health is below 0
        if (initHealth.lte(ZERO_I80F48())) {
            return ZERO_I80F48();
        }
        // Deposits need special treatment since they would neither count towards liabilities
        // nor would be charged loanOriginationFeeRate when withdrawn
        const tp = this.getToken(tokenBank.tokenIndex);
        const existingTokenDeposits = tp ? tp.deposits(tokenBank) : ZERO_I80F48();
        let existingPositionHealthContrib = ZERO_I80F48();
        if (existingTokenDeposits.gt(ZERO_I80F48())) {
            existingPositionHealthContrib = existingTokenDeposits
                .mul(tokenBank.price)
                .imul(tokenBank.initAssetWeight);
        }
        // Case 2: token deposits have higher contribution than initHealth,
        // can withdraw without borrowing until initHealth reaches 0
        if (existingPositionHealthContrib.gt(initHealth)) {
            const withdrawAbleExistingPositionHealthContrib = initHealth;
            return withdrawAbleExistingPositionHealthContrib
                .div(tokenBank.initAssetWeight)
                .div(tokenBank.price);
        }
        // Case 3: withdraw = withdraw existing deposits + borrows until initHealth reaches 0
        const initHealthWithoutExistingPosition = initHealth.sub(existingPositionHealthContrib);
        let maxBorrowNative = initHealthWithoutExistingPosition
            .div(tokenBank.initLiabWeight)
            .div(tokenBank.price);
        // Cap maxBorrow to maintain minVaultToDepositsRatio on the bank
        const vaultAmount = group.vaultAmountsMap.get(tokenBank.vault.toBase58());
        if (!vaultAmount) {
            throw new Error(`No vault amount found for ${tokenBank.name} vault ${tokenBank.vault}!`);
        }
        const vaultAmountAfterWithdrawingDeposits = I80F48.fromU64(vaultAmount).sub(existingTokenDeposits);
        const expectedVaultMinAmount = tokenBank
            .nativeDeposits()
            .mul(I80F48.fromNumber(tokenBank.minVaultToDepositsRatio));
        if (vaultAmountAfterWithdrawingDeposits.gt(expectedVaultMinAmount)) {
            maxBorrowNative = maxBorrowNative.min(vaultAmountAfterWithdrawingDeposits.sub(expectedVaultMinAmount));
        }
        const maxBorrowNativeWithoutFees = maxBorrowNative.div(ONE_I80F48().add(tokenBank.loanOriginationFeeRate));
        return maxBorrowNativeWithoutFees.add(existingTokenDeposits);
    }
    getMaxWithdrawWithBorrowForTokenUi(group, mintPk) {
        const maxWithdrawWithBorrow = this.getMaxWithdrawWithBorrowForToken(group, mintPk);
        return toUiDecimals(maxWithdrawWithBorrow, group.getMintDecimals(mintPk));
    }
    /**
     * The max amount of given source ui token you can swap to a target token.
     * @returns max amount of given source ui token you can swap to a target token, in ui token
     */
    getMaxSourceUiForTokenSwap(group, sourceMintPk, targetMintPk, slippageAndFeesFactor = 1) {
        if (sourceMintPk.equals(targetMintPk)) {
            return 0;
        }
        const sourceBank = group.getFirstBankByMint(sourceMintPk);
        const targetBank = group.getFirstBankByMint(targetMintPk);
        const hc = HealthCache.fromMangoAccount(group, this);
        let maxSource = hc.getMaxSwapSource(sourceBank, targetBank, I80F48.fromNumber(slippageAndFeesFactor *
            ((sourceBank.uiPrice / targetBank.uiPrice) *
                Math.pow(10, targetBank.mintDecimals - sourceBank.mintDecimals))));
        const sourceBalance = this.getTokenBalance(sourceBank);
        if (maxSource.gt(sourceBalance)) {
            const sourceBorrow = maxSource.sub(sourceBalance);
            maxSource = sourceBalance.add(sourceBorrow.div(ONE_I80F48().add(sourceBank.loanOriginationFeeRate)));
        }
        return toUiDecimals(maxSource, group.getMintDecimals(sourceMintPk));
    }
    /**
     * Simulates new health ratio after applying tokenChanges to the token positions.
     * Note: token changes are expected in ui amounts
     *
     * e.g. useful to simulate health after a potential swap.
     * Note: health ratio is technically ∞ if liabs are 0
     * @returns health ratio, in percentage form
     */
    simHealthRatioWithTokenPositionUiChanges(group, uiTokenChanges, healthType = HealthType.init) {
        const nativeTokenChanges = uiTokenChanges.map((tokenChange) => {
            return {
                nativeTokenAmount: toNativeI80F48(tokenChange.uiTokenAmount, group.getMintDecimals(tokenChange.mintPk)),
                mintPk: tokenChange.mintPk,
            };
        });
        const hc = HealthCache.fromMangoAccount(group, this);
        return hc
            .simHealthRatioWithTokenPositionChanges(group, nativeTokenChanges, healthType)
            .toNumber();
    }
    async loadSerum3OpenOrdersAccounts(client) {
        const response = await client.program.provider.connection.getMultipleAccountsInfo(this.serum3Active().map((s) => s.openOrders));
        const accounts = response.filter((a) => Boolean(a));
        return accounts.map((acc, index) => {
            return market.OpenOrders.fromAccountInfo(this.serum3[index].openOrders, acc, OPENBOOK_PROGRAM_ID[client.cluster]);
        });
    }
    async loadSerum3OpenOrdersForMarket(client, group, externalMarketPk) {
        const serum3Market = group.getSerum3MarketByExternalMarket(externalMarketPk);
        const serum3OO = this.serum3Active().find((s) => s.marketIndex === serum3Market.marketIndex);
        if (!serum3OO) {
            throw new Error(`No open orders account found for ${externalMarketPk}`);
        }
        const serum3MarketExternal = group.serum3ExternalMarketsMap.get(externalMarketPk.toBase58());
        const [bidsInfo, asksInfo] = await client.program.provider.connection.getMultipleAccountsInfo([
            serum3MarketExternal.bidsAddress,
            serum3MarketExternal.asksAddress,
        ]);
        if (!bidsInfo) {
            throw new Error(`Undefined bidsInfo for serum3Market with externalMarket ${externalMarketPk.toString()}`);
        }
        if (!asksInfo) {
            throw new Error(`Undefined asksInfo for serum3Market with externalMarket ${externalMarketPk.toString()}`);
        }
        const bids = market.Orderbook.decode(serum3MarketExternal, bidsInfo.data);
        const asks = market.Orderbook.decode(serum3MarketExternal, asksInfo.data);
        return [...bids, ...asks].filter((o) => o.openOrdersAddress.equals(serum3OO.openOrders));
    }
    /**
     * TODO REWORK, know to break in binary search, also make work for limit orders
     *
     * @param group
     * @param externalMarketPk
     * @returns maximum ui quote which can be traded at oracle price for base token given current health
     */
    getMaxQuoteForSerum3BidUi(group, externalMarketPk) {
        const serum3Market = group.getSerum3MarketByExternalMarket(externalMarketPk);
        const baseBank = group.getFirstBankByTokenIndex(serum3Market.baseTokenIndex);
        const quoteBank = group.getFirstBankByTokenIndex(serum3Market.quoteTokenIndex);
        const hc = HealthCache.fromMangoAccount(group, this);
        const nativeAmount = hc.getMaxSerum3OrderForHealthRatio(baseBank, quoteBank, serum3Market, Serum3Side.bid, I80F48.fromNumber(2));
        let quoteAmount = nativeAmount.div(quoteBank.price);
        // If its a bid then the reserved fund and potential loan is in base
        // also keep some buffer for fees, use taker fees for worst case simulation.
        const quoteBalance = this.getTokenBalance(quoteBank);
        if (quoteAmount.gt(quoteBalance)) {
            const quoteBorrow = quoteAmount.sub(quoteBalance);
            quoteAmount = quoteBalance.add(quoteBorrow.div(ONE_I80F48().add(quoteBank.loanOriginationFeeRate)));
        }
        quoteAmount = quoteAmount.div(ONE_I80F48().add(I80F48.fromNumber(serum3Market.getFeeRates(true))));
        return toUiDecimals(nativeAmount, quoteBank.mintDecimals);
    }
    /**
     * TODO REWORK, know to break in binary search, also make work for limit orders
     * @param group
     * @param externalMarketPk
     * @returns maximum ui base which can be traded at oracle price for quote token given current health
     */
    getMaxBaseForSerum3AskUi(group, externalMarketPk) {
        const serum3Market = group.getSerum3MarketByExternalMarket(externalMarketPk);
        const baseBank = group.getFirstBankByTokenIndex(serum3Market.baseTokenIndex);
        const quoteBank = group.getFirstBankByTokenIndex(serum3Market.quoteTokenIndex);
        const hc = HealthCache.fromMangoAccount(group, this);
        const nativeAmount = hc.getMaxSerum3OrderForHealthRatio(baseBank, quoteBank, serum3Market, Serum3Side.ask, I80F48.fromNumber(2));
        let baseAmount = nativeAmount.div(baseBank.price);
        // If its a ask then the reserved fund and potential loan is in base
        // also keep some buffer for fees, use taker fees for worst case simulation.
        const baseBalance = this.getTokenBalance(baseBank);
        if (baseAmount.gt(baseBalance)) {
            const baseBorrow = baseAmount.sub(baseBalance);
            baseAmount = baseBalance.add(baseBorrow.div(ONE_I80F48().add(baseBank.loanOriginationFeeRate)));
        }
        baseAmount = baseAmount.div(ONE_I80F48().add(I80F48.fromNumber(serum3Market.getFeeRates(true))));
        return toUiDecimals(baseAmount, baseBank.mintDecimals);
    }
    /**
     *
     * @param group
     * @param uiQuoteAmount
     * @param externalMarketPk
     * @param healthType
     * @returns health ratio after a bid with uiQuoteAmount is placed
     */
    simHealthRatioWithSerum3BidUiChanges(group, uiQuoteAmount, externalMarketPk, healthType = HealthType.init) {
        const serum3Market = group.getSerum3MarketByExternalMarket(externalMarketPk);
        const baseBank = group.getFirstBankByTokenIndex(serum3Market.baseTokenIndex);
        const quoteBank = group.getFirstBankByTokenIndex(serum3Market.quoteTokenIndex);
        const hc = HealthCache.fromMangoAccount(group, this);
        return hc
            .simHealthRatioWithSerum3BidChanges(baseBank, quoteBank, toNativeI80F48(uiQuoteAmount, group.getFirstBankByTokenIndex(serum3Market.quoteTokenIndex)
            .mintDecimals), serum3Market, healthType)
            .toNumber();
    }
    /**
     *
     * @param group
     * @param uiBaseAmount
     * @param externalMarketPk
     * @param healthType
     * @returns health ratio after an ask with uiBaseAmount is placed
     */
    simHealthRatioWithSerum3AskUiChanges(group, uiBaseAmount, externalMarketPk, healthType = HealthType.init) {
        const serum3Market = group.getSerum3MarketByExternalMarket(externalMarketPk);
        const baseBank = group.getFirstBankByTokenIndex(serum3Market.baseTokenIndex);
        const quoteBank = group.getFirstBankByTokenIndex(serum3Market.quoteTokenIndex);
        const hc = HealthCache.fromMangoAccount(group, this);
        return hc
            .simHealthRatioWithSerum3AskChanges(baseBank, quoteBank, toNativeI80F48(uiBaseAmount, group.getFirstBankByTokenIndex(serum3Market.baseTokenIndex)
            .mintDecimals), serum3Market, healthType)
            .toNumber();
    }
    // TODO: don't send a settle instruction if there's nothing to settle
    async serum3SettleFundsForAllMarkets(client, group) {
        // Future: collect ixs, batch them, and send them in fewer txs
        return await Promise.all(this.serum3Active().map((s) => {
            const serum3Market = group.getSerum3MarketByMarketIndex(s.marketIndex);
            return client.serum3SettleFunds(group, this, serum3Market.serumMarketExternal);
        }));
    }
    // TODO: cancel until all are cancelled
    async serum3CancelAllOrdersForAllMarkets(client, group) {
        // Future: collect ixs, batch them, and send them in in fewer txs
        return await Promise.all(this.serum3Active().map((s) => {
            const serum3Market = group.getSerum3MarketByMarketIndex(s.marketIndex);
            return client.serum3CancelAllOrders(group, this, serum3Market.serumMarketExternal);
        }));
    }
    /**
     * TODO: also think about limit orders
     *
     * The max ui quote you can place a market/ioc bid on the market,
     * price is the ui price at which you think the order would materialiase.
     * @param group
     * @param perpMarketName
     * @returns maximum ui quote which can be traded at oracle price for quote token given current health
     */
    getMaxQuoteForPerpBidUi(group, perpMarketIndex) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const hc = HealthCache.fromMangoAccount(group, this);
        const baseLots = hc.getMaxPerpForHealthRatio(perpMarket, I80F48.fromNumber(perpMarket.uiPrice), PerpOrderSide.bid, I80F48.fromNumber(2));
        const nativeBase = baseLots.mul(I80F48.fromI64(perpMarket.baseLotSize));
        const nativeQuote = nativeBase.mul(perpMarket.price);
        return toUiDecimalsForQuote(nativeQuote);
    }
    /**
     * TODO: also think about limit orders
     *
     * The max ui base you can place a market/ioc ask on the market,
     * price is the ui price at which you think the order would materialiase.
     * @param group
     * @param perpMarketName
     * @param uiPrice ui price at which ask would be placed at
     * @returns max ui base ask
     */
    getMaxBaseForPerpAskUi(group, perpMarketIndex) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const hc = HealthCache.fromMangoAccount(group, this);
        const baseLots = hc.getMaxPerpForHealthRatio(perpMarket, I80F48.fromNumber(perpMarket.uiPrice), PerpOrderSide.ask, I80F48.fromNumber(2));
        return perpMarket.baseLotsToUi(new anchor.BN(baseLots.toString()));
    }
    simHealthRatioWithPerpBidUiChanges(group, perpMarketIndex, size) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const pp = this.getPerpPosition(perpMarket.perpMarketIndex);
        const hc = HealthCache.fromMangoAccount(group, this);
        return hc
            .simHealthRatioWithPerpOrderChanges(perpMarket, pp
            ? pp
            : PerpPosition.emptyFromPerpMarketIndex(perpMarket.perpMarketIndex), PerpOrderSide.bid, perpMarket.uiBaseToLots(size), I80F48.fromNumber(perpMarket.uiPrice), HealthType.init)
            .toNumber();
    }
    simHealthRatioWithPerpAskUiChanges(group, perpMarketIndex, size) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const pp = this.getPerpPosition(perpMarket.perpMarketIndex);
        const hc = HealthCache.fromMangoAccount(group, this);
        return hc
            .simHealthRatioWithPerpOrderChanges(perpMarket, pp
            ? pp
            : PerpPosition.emptyFromPerpMarketIndex(perpMarket.perpMarketIndex), PerpOrderSide.ask, perpMarket.uiBaseToLots(size), I80F48.fromNumber(perpMarket.uiPrice), HealthType.init)
            .toNumber();
    }
    async loadPerpOpenOrdersForMarket(client, group, perpMarketIndex) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const [bids, asks] = await Promise.all([
            perpMarket.loadBids(client),
            perpMarket.loadAsks(client),
        ]);
        return [...Array.from(bids.items()), ...Array.from(asks.items())].filter((order) => order.owner.equals(this.publicKey));
    }
    toString(group, onlyTokens = false) {
        let res = 'MangoAccount';
        res = res + '\n pk: ' + this.publicKey.toString();
        res = res + '\n name: ' + this.name;
        res = res + '\n accountNum: ' + this.accountNum;
        res = res + '\n owner: ' + this.owner;
        res = res + '\n delegate: ' + this.delegate;
        res =
            res +
                `\n max token slots ${this.tokens.length}, max serum3 slots ${this.serum3.length}, max perp slots ${this.perps.length}, max perp oo slots ${this.perpOpenOrders.length}`;
        res =
            this.tokensActive().length > 0
                ? res +
                    '\n tokens:' +
                    JSON.stringify(this.tokens
                        .filter((token, i) => token.isActive())
                        .map((token, i) => token.toString(group, i)), null, 4)
                : res + '';
        if (onlyTokens) {
            return res;
        }
        res =
            this.serum3Active().length > 0
                ? res + '\n serum:' + JSON.stringify(this.serum3Active(), null, 4)
                : res + '';
        res =
            this.perpActive().length > 0
                ? res +
                    '\n perps:' +
                    JSON.stringify(this.perpActive().map((p) => p.toString(group?.getPerpMarketByMarketIndex(p.marketIndex))), null, 4)
                : res + '';
        res =
            this.perpOrdersActive().length > 0
                ? res +
                    '\n perps oo:' +
                    JSON.stringify(this.perpOrdersActive(), null, 4)
                : res + '';
        return res;
    }
}
class TokenPosition {
    indexedPosition;
    tokenIndex;
    inUseCount;
    previousIndex;
    cumulativeDepositInterest;
    cumulativeBorrowInterest;
    static TokenIndexUnset = 65535;
    static from(dto) {
        return new TokenPosition(I80F48.from(dto.indexedPosition), dto.tokenIndex, dto.inUseCount, I80F48.from(dto.previousIndex), dto.cumulativeDepositInterest, dto.cumulativeBorrowInterest);
    }
    constructor(indexedPosition, tokenIndex, inUseCount, previousIndex, cumulativeDepositInterest, cumulativeBorrowInterest) {
        this.indexedPosition = indexedPosition;
        this.tokenIndex = tokenIndex;
        this.inUseCount = inUseCount;
        this.previousIndex = previousIndex;
        this.cumulativeDepositInterest = cumulativeDepositInterest;
        this.cumulativeBorrowInterest = cumulativeBorrowInterest;
    }
    isActive() {
        return this.tokenIndex !== TokenPosition.TokenIndexUnset;
    }
    /**
     *
     * @param bank
     * @returns native balance
     */
    balance(bank) {
        if (this.indexedPosition.isPos()) {
            return bank.depositIndex.mul(this.indexedPosition);
        }
        else {
            return bank.borrowIndex.mul(this.indexedPosition);
        }
    }
    /**
     *
     * @param bank
     * @returns native deposits, 0 if position has borrows
     */
    deposits(bank) {
        if (this.indexedPosition && this.indexedPosition.lt(ZERO_I80F48())) {
            return ZERO_I80F48();
        }
        return this.balance(bank);
    }
    /**
     *
     * @param bank
     * @returns native borrows, 0 if position has deposits
     */
    borrows(bank) {
        if (this.indexedPosition && this.indexedPosition.gt(ZERO_I80F48())) {
            return ZERO_I80F48();
        }
        return this.balance(bank).abs();
    }
    /**
     * @param bank
     * @returns UI balance, is signed
     */
    balanceUi(bank) {
        return toUiDecimals(this.balance(bank), bank.mintDecimals);
    }
    /**
     * @param bank
     * @returns UI deposits, 0 if position has borrows
     */
    depositsUi(bank) {
        return toUiDecimals(this.deposits(bank), bank.mintDecimals);
    }
    /**
     * @param bank
     * @returns UI borrows, 0 if position has deposits
     */
    borrowsUi(bank) {
        return toUiDecimals(this.borrows(bank), bank.mintDecimals);
    }
    toString(group, index) {
        let extra = '';
        if (group) {
            const bank = group.getFirstBankByTokenIndex(this.tokenIndex);
            if (bank) {
                const native = this.balance(bank);
                extra += ', native: ' + native.toNumber();
                extra += ', ui: ' + this.balanceUi(bank);
                extra += ', tokenName: ' + bank.name;
            }
        }
        return ((index !== undefined ? 'index: ' + index : '') +
            ', tokenIndex: ' +
            this.tokenIndex +
            ', inUseCount: ' +
            this.inUseCount +
            ', indexedValue: ' +
            this.indexedPosition.toNumber() +
            extra);
    }
}
class TokenPositionDto {
    indexedPosition;
    tokenIndex;
    inUseCount;
    reserved;
    previousIndex;
    cumulativeDepositInterest;
    cumulativeBorrowInterest;
    constructor(indexedPosition, tokenIndex, inUseCount, reserved, previousIndex, cumulativeDepositInterest, cumulativeBorrowInterest) {
        this.indexedPosition = indexedPosition;
        this.tokenIndex = tokenIndex;
        this.inUseCount = inUseCount;
        this.reserved = reserved;
        this.previousIndex = previousIndex;
        this.cumulativeDepositInterest = cumulativeDepositInterest;
        this.cumulativeBorrowInterest = cumulativeBorrowInterest;
    }
}
class Serum3Orders {
    openOrders;
    marketIndex;
    baseTokenIndex;
    quoteTokenIndex;
    static Serum3MarketIndexUnset = 65535;
    static from(dto) {
        return new Serum3Orders(dto.openOrders, dto.marketIndex, dto.baseTokenIndex, dto.quoteTokenIndex);
    }
    constructor(openOrders, marketIndex, baseTokenIndex, quoteTokenIndex) {
        this.openOrders = openOrders;
        this.marketIndex = marketIndex;
        this.baseTokenIndex = baseTokenIndex;
        this.quoteTokenIndex = quoteTokenIndex;
    }
    isActive() {
        return this.marketIndex !== Serum3Orders.Serum3MarketIndexUnset;
    }
}
class Serum3PositionDto {
    openOrders;
    marketIndex;
    baseBorrowsWithoutFee;
    quoteBorrowsWithoutFee;
    baseTokenIndex;
    quoteTokenIndex;
    reserved;
    constructor(openOrders, marketIndex, baseBorrowsWithoutFee, quoteBorrowsWithoutFee, baseTokenIndex, quoteTokenIndex, reserved) {
        this.openOrders = openOrders;
        this.marketIndex = marketIndex;
        this.baseBorrowsWithoutFee = baseBorrowsWithoutFee;
        this.quoteBorrowsWithoutFee = quoteBorrowsWithoutFee;
        this.baseTokenIndex = baseTokenIndex;
        this.quoteTokenIndex = quoteTokenIndex;
        this.reserved = reserved;
    }
}
class PerpPosition {
    marketIndex;
    settlePnlLimitWindow;
    settlePnlLimitSettledInCurrentWindowNative;
    basePositionLots;
    quotePositionNative;
    quoteRunningNative;
    longSettledFunding;
    shortSettledFunding;
    bidsBaseLots;
    asksBaseLots;
    takerBaseLots;
    takerQuoteLots;
    cumulativeLongFunding;
    cumulativeShortFunding;
    makerVolume;
    takerVolume;
    perpSpotTransfers;
    avgEntryPricePerBaseLot;
    realizedTradePnlNative;
    realizedOtherPnlNative;
    settlePnlLimitRealizedTrade;
    realizedPnlForPositionNative;
    static PerpMarketIndexUnset = 65535;
    static from(dto) {
        return new PerpPosition(dto.marketIndex, dto.settlePnlLimitWindow, dto.settlePnlLimitSettledInCurrentWindowNative, dto.basePositionLots, I80F48.from(dto.quotePositionNative), dto.quoteRunningNative, I80F48.from(dto.longSettledFunding), I80F48.from(dto.shortSettledFunding), dto.bidsBaseLots, dto.asksBaseLots, dto.takerBaseLots, dto.takerQuoteLots, dto.cumulativeLongFunding, dto.cumulativeShortFunding, dto.makerVolume, dto.takerVolume, dto.perpSpotTransfers, dto.avgEntryPricePerBaseLot, I80F48.from(dto.realizedTradePnlNative), I80F48.from(dto.realizedOtherPnlNative), dto.settlePnlLimitRealizedTrade, I80F48.from(dto.realizedPnlForPositionNative));
    }
    static emptyFromPerpMarketIndex(perpMarketIndex) {
        return new PerpPosition(perpMarketIndex, 0, new anchor.BN(0), new anchor.BN(0), ZERO_I80F48(), new anchor.BN(0), ZERO_I80F48(), ZERO_I80F48(), new anchor.BN(0), new anchor.BN(0), new anchor.BN(0), new anchor.BN(0), 0, 0, new anchor.BN(0), new anchor.BN(0), new anchor.BN(0), 0, ZERO_I80F48(), ZERO_I80F48(), new anchor.BN(0), ZERO_I80F48());
    }
    constructor(marketIndex, settlePnlLimitWindow, settlePnlLimitSettledInCurrentWindowNative, basePositionLots, quotePositionNative, quoteRunningNative, longSettledFunding, shortSettledFunding, bidsBaseLots, asksBaseLots, takerBaseLots, takerQuoteLots, cumulativeLongFunding, cumulativeShortFunding, makerVolume, takerVolume, perpSpotTransfers, avgEntryPricePerBaseLot, realizedTradePnlNative, realizedOtherPnlNative, settlePnlLimitRealizedTrade, realizedPnlForPositionNative) {
        this.marketIndex = marketIndex;
        this.settlePnlLimitWindow = settlePnlLimitWindow;
        this.settlePnlLimitSettledInCurrentWindowNative = settlePnlLimitSettledInCurrentWindowNative;
        this.basePositionLots = basePositionLots;
        this.quotePositionNative = quotePositionNative;
        this.quoteRunningNative = quoteRunningNative;
        this.longSettledFunding = longSettledFunding;
        this.shortSettledFunding = shortSettledFunding;
        this.bidsBaseLots = bidsBaseLots;
        this.asksBaseLots = asksBaseLots;
        this.takerBaseLots = takerBaseLots;
        this.takerQuoteLots = takerQuoteLots;
        this.cumulativeLongFunding = cumulativeLongFunding;
        this.cumulativeShortFunding = cumulativeShortFunding;
        this.makerVolume = makerVolume;
        this.takerVolume = takerVolume;
        this.perpSpotTransfers = perpSpotTransfers;
        this.avgEntryPricePerBaseLot = avgEntryPricePerBaseLot;
        this.realizedTradePnlNative = realizedTradePnlNative;
        this.realizedOtherPnlNative = realizedOtherPnlNative;
        this.settlePnlLimitRealizedTrade = settlePnlLimitRealizedTrade;
        this.realizedPnlForPositionNative = realizedPnlForPositionNative;
    }
    isActive() {
        return this.marketIndex !== PerpPosition.PerpMarketIndexUnset;
    }
    getBasePositionNative(perpMarket) {
        return I80F48.fromI64(this.basePositionLots.mul(perpMarket.baseLotSize));
    }
    getBasePositionUi(perpMarket, useEventQueue) {
        if (perpMarket.perpMarketIndex !== this.marketIndex) {
            throw new Error("PerpPosition doesn't belong to the given market!");
        }
        return perpMarket.baseLotsToUi(useEventQueue
            ? this.basePositionLots.add(this.takerBaseLots)
            : this.basePositionLots);
    }
    getUnsettledFunding(perpMarket) {
        if (perpMarket.perpMarketIndex !== this.marketIndex) {
            throw new Error("PerpPosition doesn't belong to the given market!");
        }
        if (this.basePositionLots.gt(new anchor.BN(0))) {
            return perpMarket.longFunding
                .sub(this.longSettledFunding)
                .mul(I80F48.fromI64(this.basePositionLots));
        }
        else if (this.basePositionLots.lt(new anchor.BN(0))) {
            return perpMarket.shortFunding
                .sub(this.shortSettledFunding)
                .mul(I80F48.fromI64(this.basePositionLots));
        }
        return ZERO_I80F48();
    }
    getEquityUi(perpMarket) {
        if (perpMarket.perpMarketIndex !== this.marketIndex) {
            throw new Error("PerpPosition doesn't belong to the given market!");
        }
        return toUiDecimalsForQuote(this.getEquity(perpMarket));
    }
    getEquity(perpMarket) {
        if (perpMarket.perpMarketIndex !== this.marketIndex) {
            throw new Error("PerpPosition doesn't belong to the given market!");
        }
        const lotsToQuote = I80F48.fromI64(perpMarket.baseLotSize).mul(perpMarket.price);
        const baseLots = I80F48.fromI64(this.basePositionLots.add(this.takerBaseLots));
        const unsettledFunding = this.getUnsettledFunding(perpMarket);
        const takerQuote = I80F48.fromI64(new anchor.BN(this.takerQuoteLots).mul(perpMarket.quoteLotSize));
        const quoteCurrent = this.quotePositionNative
            .sub(unsettledFunding)
            .add(takerQuote);
        return baseLots.mul(lotsToQuote).add(quoteCurrent);
    }
    hasOpenOrders() {
        const zero = new anchor.BN(0);
        return (!this.asksBaseLots.eq(zero) ||
            !this.bidsBaseLots.eq(zero) ||
            !this.takerBaseLots.eq(zero) ||
            !this.takerQuoteLots.eq(zero));
    }
    getAverageEntryPrice(perpMarket) {
        return I80F48.fromNumber(this.avgEntryPricePerBaseLot).div(I80F48.fromI64(perpMarket.baseLotSize));
    }
    getAverageEntryPriceUi(perpMarket) {
        return perpMarket.priceNativeToUi(this.getAverageEntryPrice(perpMarket).toNumber());
    }
    getBreakEvenPriceUi(perpMarket) {
        if (perpMarket.perpMarketIndex !== this.marketIndex) {
            throw new Error("PerpPosition doesn't belong to the given market!");
        }
        if (this.basePositionLots.eq(new anchor.BN(0))) {
            return 0;
        }
        return perpMarket.priceNativeToUi(-this.quoteRunningNative.toNumber() /
            this.basePositionLots.mul(perpMarket.baseLotSize).toNumber());
    }
    cumulativePnlOverPositionLifetimeUi(perpMarket) {
        if (perpMarket.perpMarketIndex !== this.marketIndex) {
            throw new Error("PerpPosition doesn't belong to the given market!");
        }
        const priceChange = perpMarket.price.sub(this.getAverageEntryPrice(perpMarket));
        return toUiDecimalsForQuote(this.realizedPnlForPositionNative.add(this.getBasePositionNative(perpMarket).mul(priceChange)));
    }
    getUnsettledPnl(perpMarket) {
        if (perpMarket.perpMarketIndex !== this.marketIndex) {
            throw new Error("PerpPosition doesn't belong to the given market!");
        }
        return this.quotePositionNative.add(this.getBasePositionNative(perpMarket).mul(perpMarket.price));
    }
    getUnsettledPnlUi(perpMarket) {
        return toUiDecimalsForQuote(this.getUnsettledPnl(perpMarket));
    }
    updateSettleLimit(perpMarket) {
        if (perpMarket.perpMarketIndex !== this.marketIndex) {
            throw new Error("PerpPosition doesn't belong to the given market!");
        }
        const windowSize = perpMarket.settlePnlLimitWindowSizeTs;
        const windowStart = new anchor.BN(this.settlePnlLimitWindow).mul(windowSize);
        const windowEnd = windowStart.add(windowSize);
        const nowTs = new anchor.BN(Date.now() / 1000);
        const newWindow = nowTs.gte(windowEnd) || nowTs.lt(windowStart);
        if (newWindow) {
            this.settlePnlLimitWindow = nowTs.div(windowSize).toNumber();
            this.settlePnlLimitSettledInCurrentWindowNative = new anchor.BN(0);
        }
    }
    availableSettleLimit(perpMarket) {
        if (perpMarket.perpMarketIndex !== this.marketIndex) {
            throw new Error("PerpPosition doesn't belong to the given market!");
        }
        if (perpMarket.settlePnlLimitFactor < 0) {
            return [RUST_I64_MIN(), RUST_I64_MAX()];
        }
        const baseNative = I80F48.fromI64(this.basePositionLots.mul(perpMarket.baseLotSize));
        const positionValue = I80F48.fromNumber(perpMarket.stablePriceModel.stablePrice)
            .mul(baseNative)
            .toNumber();
        const unrealized = new anchor.BN(perpMarket.settlePnlLimitFactor * positionValue);
        const used = new anchor.BN(this.settlePnlLimitSettledInCurrentWindowNative.toNumber());
        let minPnl = unrealized.neg().sub(used);
        let maxPnl = unrealized.sub(used);
        const realizedTrade = this.settlePnlLimitRealizedTrade;
        if (realizedTrade.gte(new anchor.BN(0))) {
            maxPnl = maxPnl.add(realizedTrade);
        }
        else {
            minPnl = minPnl.add(realizedTrade);
        }
        const realizedOther = new anchor.BN(this.realizedOtherPnlNative.toNumber());
        if (realizedOther.gte(new anchor.BN(0))) {
            maxPnl = maxPnl.add(realizedOther);
        }
        else {
            minPnl = minPnl.add(realizedOther);
        }
        return [anchor.BN.min(minPnl, new anchor.BN(0)), anchor.BN.max(maxPnl, new anchor.BN(0))];
    }
    applyPnlSettleLimit(pnl, perpMarket) {
        if (perpMarket.perpMarketIndex !== this.marketIndex) {
            throw new Error("PerpPosition doesn't belong to the given market!");
        }
        if (perpMarket.settlePnlLimitFactor < 0) {
            return pnl;
        }
        const [minPnl, maxPnl] = this.availableSettleLimit(perpMarket);
        if (pnl.lt(ZERO_I80F48())) {
            return pnl.max(I80F48.fromI64(minPnl));
        }
        else {
            return pnl.min(I80F48.fromI64(maxPnl));
        }
    }
    getSettleablePnl(group, perpMarket, account) {
        if (perpMarket.perpMarketIndex !== this.marketIndex) {
            throw new Error("PerpPosition doesn't belong to the given market!");
        }
        this.updateSettleLimit(perpMarket);
        const perpSettleHealth = account.getPerpSettleHealth(group);
        const limitedUnsettled = this.applyPnlSettleLimit(this.getUnsettledPnl(perpMarket), perpMarket);
        if (limitedUnsettled.lt(ZERO_I80F48())) {
            return limitedUnsettled.max(perpSettleHealth.max(ZERO_I80F48()).neg());
        }
        return limitedUnsettled;
    }
    getSettleablePnlUi(group, perpMarket, account) {
        return toUiDecimalsForQuote(this.getSettleablePnl(group, perpMarket, account));
    }
    canSettlePnl(group, perpMarket, account) {
        return !this.getSettleablePnl(group, perpMarket, account).eq(ZERO_I80F48());
    }
    toString(perpMarket) {
        return perpMarket
            ? 'market - ' +
                perpMarket.name +
                ', basePositionLots - ' +
                perpMarket.baseLotsToUi(this.basePositionLots) +
                ', quotePositive - ' +
                toUiDecimalsForQuote(this.quotePositionNative.toNumber()) +
                ', bidsBaseLots - ' +
                perpMarket.baseLotsToUi(this.bidsBaseLots) +
                ', asksBaseLots - ' +
                perpMarket.baseLotsToUi(this.asksBaseLots) +
                ', takerBaseLots - ' +
                perpMarket.baseLotsToUi(this.takerBaseLots) +
                ', takerQuoteLots - ' +
                perpMarket.quoteLotsToUi(this.takerQuoteLots) +
                ', unsettled pnl - ' +
                this.getUnsettledPnlUi(perpMarket).toString()
            : '';
    }
}
class PerpPositionDto {
    marketIndex;
    settlePnlLimitWindow;
    settlePnlLimitSettledInCurrentWindowNative;
    basePositionLots;
    quotePositionNative;
    quoteRunningNative;
    longSettledFunding;
    shortSettledFunding;
    bidsBaseLots;
    asksBaseLots;
    takerBaseLots;
    takerQuoteLots;
    cumulativeLongFunding;
    cumulativeShortFunding;
    makerVolume;
    takerVolume;
    perpSpotTransfers;
    avgEntryPricePerBaseLot;
    realizedTradePnlNative;
    realizedOtherPnlNative;
    settlePnlLimitRealizedTrade;
    realizedPnlForPositionNative;
    constructor(marketIndex, settlePnlLimitWindow, settlePnlLimitSettledInCurrentWindowNative, basePositionLots, quotePositionNative, quoteRunningNative, longSettledFunding, shortSettledFunding, bidsBaseLots, asksBaseLots, takerBaseLots, takerQuoteLots, cumulativeLongFunding, cumulativeShortFunding, makerVolume, takerVolume, perpSpotTransfers, avgEntryPricePerBaseLot, realizedTradePnlNative, realizedOtherPnlNative, settlePnlLimitRealizedTrade, realizedPnlForPositionNative) {
        this.marketIndex = marketIndex;
        this.settlePnlLimitWindow = settlePnlLimitWindow;
        this.settlePnlLimitSettledInCurrentWindowNative = settlePnlLimitSettledInCurrentWindowNative;
        this.basePositionLots = basePositionLots;
        this.quotePositionNative = quotePositionNative;
        this.quoteRunningNative = quoteRunningNative;
        this.longSettledFunding = longSettledFunding;
        this.shortSettledFunding = shortSettledFunding;
        this.bidsBaseLots = bidsBaseLots;
        this.asksBaseLots = asksBaseLots;
        this.takerBaseLots = takerBaseLots;
        this.takerQuoteLots = takerQuoteLots;
        this.cumulativeLongFunding = cumulativeLongFunding;
        this.cumulativeShortFunding = cumulativeShortFunding;
        this.makerVolume = makerVolume;
        this.takerVolume = takerVolume;
        this.perpSpotTransfers = perpSpotTransfers;
        this.avgEntryPricePerBaseLot = avgEntryPricePerBaseLot;
        this.realizedTradePnlNative = realizedTradePnlNative;
        this.realizedOtherPnlNative = realizedOtherPnlNative;
        this.settlePnlLimitRealizedTrade = settlePnlLimitRealizedTrade;
        this.realizedPnlForPositionNative = realizedPnlForPositionNative;
    }
}
class PerpOo {
    sideAndTree;
    orderMarket;
    clientId;
    id;
    static OrderMarketUnset = 65535;
    static from(dto) {
        return new PerpOo(dto.sideAndTree, dto.market, dto.clientId, dto.id);
    }
    constructor(sideAndTree, orderMarket, clientId, id) {
        this.sideAndTree = sideAndTree;
        this.orderMarket = orderMarket;
        this.clientId = clientId;
        this.id = id;
    }
    isActive() {
        return this.orderMarket !== PerpOo.OrderMarketUnset;
    }
}
class PerpOoDto {
    sideAndTree;
    market;
    clientId;
    id;
    constructor(sideAndTree, market, clientId, id) {
        this.sideAndTree = sideAndTree;
        this.market = market;
        this.clientId = clientId;
        this.id = id;
    }
}
class HealthType {
    static maint = { maint: {} };
    static init = { init: {} };
    static liquidationEnd = { liquidationEnd: {} };
}

const NullTokenEditParams = {
    oracle: null,
    oracleConfig: null,
    groupInsuranceFund: null,
    interestRateParams: null,
    loanFeeRate: null,
    loanOriginationFeeRate: null,
    maintAssetWeight: null,
    initAssetWeight: null,
    maintLiabWeight: null,
    initLiabWeight: null,
    liquidationFee: null,
    stablePriceDelayIntervalSeconds: null,
    stablePriceDelayGrowthLimit: null,
    stablePriceGrowthLimit: null,
    minVaultToDepositsRatio: null,
    netBorrowLimitPerWindowQuote: null,
    netBorrowLimitWindowSizeTs: null,
    borrowWeightScaleStartQuote: null,
    depositWeightScaleStartQuote: null,
    resetStablePrice: null,
    resetNetBorrowLimit: null,
    reduceOnly: null,
};
const NullPerpEditParams = {
    oracle: null,
    oracleConfig: null,
    baseDecimals: null,
    maintBaseAssetWeight: null,
    initBaseAssetWeight: null,
    maintBaseLiabWeight: null,
    initBaseLiabWeight: null,
    maintOverallAssetWeight: null,
    initOverallAssetWeight: null,
    baseLiquidationFee: null,
    makerFee: null,
    takerFee: null,
    feePenalty: null,
    minFunding: null,
    maxFunding: null,
    impactQuantity: null,
    groupInsuranceFund: null,
    settleFeeFlat: null,
    settleFeeAmountThreshold: null,
    settleFeeFractionLowHealth: null,
    stablePriceDelayIntervalSeconds: null,
    stablePriceDelayGrowthLimit: null,
    stablePriceGrowthLimit: null,
    settlePnlLimitFactor: null,
    settlePnlLimitWindowSize: null,
    reduceOnly: null,
    resetStablePrice: null,
    positivePnlLiquidationFee: null,
};
// Default with all ixs enabled, use with buildIxGate
const TrueIxGateParams = {
    AccountClose: true,
    AccountCreate: true,
    AccountEdit: true,
    AccountExpand: true,
    AccountToggleFreeze: true,
    AltExtend: true,
    AltSet: true,
    FlashLoan: true,
    GroupClose: true,
    GroupCreate: true,
    GroupToggleHalt: true,
    HealthRegion: true,
    PerpCancelAllOrders: true,
    PerpCancelAllOrdersBySide: true,
    PerpCancelOrder: true,
    PerpCancelOrderByClientOrderId: true,
    PerpCloseMarket: true,
    PerpConsumeEvents: true,
    PerpCreateMarket: true,
    PerpDeactivatePosition: true,
    PerpEditMarket: true,
    PerpLiqBaseOrPositivePnl: true,
    PerpLiqForceCancelOrders: true,
    PerpLiqNegativePnlOrBankruptcy: true,
    PerpPlaceOrder: true,
    PerpSettleFees: true,
    PerpSettlePnl: true,
    PerpUpdateFunding: true,
    Serum3CancelAllOrders: true,
    Serum3CancelOrder: true,
    Serum3CloseOpenOrders: true,
    Serum3CreateOpenOrders: true,
    Serum3DeregisterMarket: true,
    Serum3EditMarket: true,
    Serum3LiqForceCancelOrders: true,
    Serum3PlaceOrder: true,
    Serum3RegisterMarket: true,
    Serum3SettleFunds: true,
    StubOracleClose: true,
    StubOracleCreate: true,
    StubOracleSet: true,
    TokenAddBank: true,
    TokenDeposit: true,
    TokenDeregister: true,
    TokenEdit: true,
    TokenLiqBankruptcy: true,
    TokenLiqWithToken: true,
    TokenRegister: true,
    TokenRegisterTrustless: true,
    TokenUpdateIndexAndRate: true,
    TokenWithdraw: true,
};
// build ix gate e.g. buildIxGate(Builder(TrueIxGateParams).TokenDeposit(false).build()).toNumber(),
function buildIxGate(p) {
    const ixGate = new anchor.BN(0);
    function toggleIx(ixGate, p, propName, index) {
        if (p[propName] === undefined) {
            throw new Error(`Unknown property ${propName}`);
        }
        ixGate.ior(p[propName] ? new anchor.BN(0) : new anchor.BN(1).ushln(index));
    }
    toggleIx(ixGate, p, 'AccountClose', 0);
    toggleIx(ixGate, p, 'AccountCreate', 1);
    toggleIx(ixGate, p, 'AccountEdit', 2);
    toggleIx(ixGate, p, 'AccountExpand', 3);
    toggleIx(ixGate, p, 'AccountToggleFreeze', 4);
    toggleIx(ixGate, p, 'AltExtend', 5);
    toggleIx(ixGate, p, 'AltSet', 6);
    toggleIx(ixGate, p, 'FlashLoan', 7);
    toggleIx(ixGate, p, 'GroupClose', 8);
    toggleIx(ixGate, p, 'GroupCreate', 9);
    toggleIx(ixGate, p, 'HealthRegion', 10);
    toggleIx(ixGate, p, 'PerpCancelAllOrders', 11);
    toggleIx(ixGate, p, 'PerpCancelAllOrdersBySide', 12);
    toggleIx(ixGate, p, 'PerpCancelOrder', 13);
    toggleIx(ixGate, p, 'PerpCancelOrderByClientOrderId', 14);
    toggleIx(ixGate, p, 'PerpCloseMarket', 15);
    toggleIx(ixGate, p, 'PerpConsumeEvents', 16);
    toggleIx(ixGate, p, 'PerpCreateMarket', 17);
    toggleIx(ixGate, p, 'PerpDeactivatePosition', 18);
    toggleIx(ixGate, p, 'PerpLiqBaseOrPositivePnl', 19);
    toggleIx(ixGate, p, 'PerpLiqForceCancelOrders', 20);
    toggleIx(ixGate, p, 'PerpLiqNegativePnlOrBankruptcy', 21);
    toggleIx(ixGate, p, 'PerpPlaceOrder', 22);
    toggleIx(ixGate, p, 'PerpSettleFees', 23);
    toggleIx(ixGate, p, 'PerpSettlePnl', 24);
    toggleIx(ixGate, p, 'PerpUpdateFunding', 25);
    toggleIx(ixGate, p, 'Serum3CancelAllOrders', 26);
    toggleIx(ixGate, p, 'Serum3CancelOrder', 27);
    toggleIx(ixGate, p, 'Serum3CloseOpenOrders', 28);
    toggleIx(ixGate, p, 'Serum3CreateOpenOrders', 29);
    toggleIx(ixGate, p, 'Serum3DeregisterMarket', 30);
    toggleIx(ixGate, p, 'Serum3EditMarket', 31);
    toggleIx(ixGate, p, 'Serum3LiqForceCancelOrders', 32);
    toggleIx(ixGate, p, 'Serum3PlaceOrder', 33);
    toggleIx(ixGate, p, 'Serum3RegisterMarket', 34);
    toggleIx(ixGate, p, 'Serum3SettleFunds', 35);
    toggleIx(ixGate, p, 'StubOracleClose', 36);
    toggleIx(ixGate, p, 'StubOracleCreate', 37);
    toggleIx(ixGate, p, 'StubOracleSet', 38);
    toggleIx(ixGate, p, 'TokenAddBank', 39);
    toggleIx(ixGate, p, 'TokenDeposit', 40);
    toggleIx(ixGate, p, 'TokenDeregister', 41);
    toggleIx(ixGate, p, 'TokenLiqBankruptcy', 42);
    toggleIx(ixGate, p, 'TokenLiqWithToken', 43);
    toggleIx(ixGate, p, 'TokenRegister', 44);
    toggleIx(ixGate, p, 'TokenRegisterTrustless', 45);
    toggleIx(ixGate, p, 'TokenUpdateIndexAndRate', 46);
    toggleIx(ixGate, p, 'TokenWithdraw', 47);
    return ixGate;
}

var groups = [
	{
		cluster: "mainnet-beta",
		name: "mainnet-beta.clarkeni",
		publicKey: "DLdcpC6AsAJ9xeKMR3WhHrN5sM5o7GVVXQhQ5vwisTtz",
		serum3ProgramId: "9xQeWvG816bUx9EPjHmaT23yvVM2ZWbrrpZb9PusVFin",
		mangoProgramId: "4MangoMjqJ2firMokCjjGgoK8d4MXcrgL7XJaL3w6fVg",
		banks: [
			{
				name: "SOL",
				publicKey: "8Z8ypSDDbC2UWqapP4AqDGbTbxYgh5AKtv1682UYc74o",
				mint: "So11111111111111111111111111111111111111112",
				tokenIndex: 5,
				bankNum: 0,
				active: true
			},
			{
				name: "USDT",
				publicKey: "Bid15jgtr5Ppftj7BZdEJbiWFN3R1U3i2PUELA2ZDU1N",
				mint: "Es9vMFrzaCERmJfrF4H2FYD4KCoNkY11McCe8BenwNYB",
				tokenIndex: 1,
				bankNum: 0,
				active: true
			},
			{
				name: "USDC",
				publicKey: "MZSA6xDJFieoMbydmjWv7FkPCcipmtDC2bxE1qYGvaY",
				mint: "EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v",
				tokenIndex: 0,
				bankNum: 0,
				active: true
			},
			{
				name: "BTC",
				publicKey: "JfS9mG5Sonm6eyhQis2DdFSkfFmtoP8kW5UAaXcEWa4",
				mint: "9n4nbM75f5Ui33ZbPYXn59EwSgE8CGsHtAeTH5YFeJ9E",
				tokenIndex: 2,
				bankNum: 0,
				active: true
			},
			{
				name: "soETH",
				publicKey: "549bGbirHQSQ5mnFQNT5zBUs8XMFyK9hNAWRWcpCJzBn",
				mint: "2FPyTwcZLUg1MDrwsyoP4D6s1tM7hAkHYRjkNb5w6Pxk",
				tokenIndex: 4,
				bankNum: 0,
				active: true
			},
			{
				name: "ETH",
				publicKey: "2emCTQgGDubaZEdstGXbf9cL4sTgCiSBvWjwk925F4Bd",
				mint: "7vfCXTUXx5WJV5JADk17DUJ4ksgau7utNKj4b963voxs",
				tokenIndex: 3,
				bankNum: 0,
				active: true
			},
			{
				name: "MSOL",
				publicKey: "2uL6J4aUqpDerWoXejAZQ585SVYLjhBhgQvpBEHeLybb",
				mint: "mSoLzYCxHdYgdzU16g5QSh3i5K3z3KZK7ytfqcJm7So",
				tokenIndex: 6,
				bankNum: 0,
				active: true
			}
		],
		stubOracles: [
			{
				name: "EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v",
				publicKey: "F1JkVYh6ouopvmgWgaR9p4DyBTeGweHPumE4PzrvK5ti"
			}
		],
		mintInfos: [
			{
				name: "SOL",
				publicKey: "TvWkqqRPT7gRtbAecnYuSgAA11gcQPKJ9eggqt6bM8Q",
				mint: "So11111111111111111111111111111111111111112",
				tokenIndex: 5
			},
			{
				name: "USDT",
				publicKey: "9HGCxDaCfV7L8megn7aSXbQ39DR4BdT2qmvLxVNYgh69",
				mint: "Es9vMFrzaCERmJfrF4H2FYD4KCoNkY11McCe8BenwNYB",
				tokenIndex: 1
			},
			{
				name: "USDC",
				publicKey: "9mbfSj6EWVYShL5ozA7hYFghdjKXtx4GXwUygogqK8ZR",
				mint: "EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v",
				tokenIndex: 0
			},
			{
				name: "BTC",
				publicKey: "2rzUjP3ZHR6ZepqwLQ561HzEe2n5Sh1M67AqmzydzzcD",
				mint: "9n4nbM75f5Ui33ZbPYXn59EwSgE8CGsHtAeTH5YFeJ9E",
				tokenIndex: 2
			},
			{
				name: "soETH",
				publicKey: "BYuxStPHhVfNXXdL1sCAfiYmCzFcpoTtbaBhYqYGaE4D",
				mint: "2FPyTwcZLUg1MDrwsyoP4D6s1tM7hAkHYRjkNb5w6Pxk",
				tokenIndex: 4
			},
			{
				name: "ETH",
				publicKey: "8J9NCLCAwHWk5pzYTSmCZPvmkTJik8M4yXtrgeFHqPha",
				mint: "7vfCXTUXx5WJV5JADk17DUJ4ksgau7utNKj4b963voxs",
				tokenIndex: 3
			},
			{
				name: "MSOL",
				publicKey: "DRnvY2SVHuj8YiyyxVHPJu8ZtaXzsgcqxz3Tzf8az3Zx",
				mint: "mSoLzYCxHdYgdzU16g5QSh3i5K3z3KZK7ytfqcJm7So",
				tokenIndex: 6
			}
		],
		serum3Markets: [
		],
		perpMarkets: [
		]
	}
];
var ids = {
	groups: groups
};

class Id {
    cluster;
    name;
    publicKey;
    serum3ProgramId;
    mangoProgramId;
    banks;
    stubOracles;
    mintInfos;
    serum3Markets;
    perpMarkets;
    constructor(cluster, name, publicKey, serum3ProgramId, mangoProgramId, banks, stubOracles, mintInfos, serum3Markets, perpMarkets) {
        this.cluster = cluster;
        this.name = name;
        this.publicKey = publicKey;
        this.serum3ProgramId = serum3ProgramId;
        this.mangoProgramId = mangoProgramId;
        this.banks = banks;
        this.stubOracles = stubOracles;
        this.mintInfos = mintInfos;
        this.serum3Markets = serum3Markets;
        this.perpMarkets = perpMarkets;
    }
    getBanks() {
        return Array.from(this.banks
            .filter((perpMarket) => perpMarket.active)
            .map((bank) => new web3_js.PublicKey(bank.publicKey)));
    }
    getStubOracles() {
        return Array.from(this.stubOracles.map((stubOracle) => new web3_js.PublicKey(stubOracle.publicKey)));
    }
    getMintInfos() {
        return Array.from(this.mintInfos.map((mintInfo) => new web3_js.PublicKey(mintInfo.publicKey)));
    }
    getSerum3Markets() {
        return Array.from(this.serum3Markets
            .filter((perpMarket) => perpMarket.active)
            .map((serum3Market) => new web3_js.PublicKey(serum3Market.publicKey)));
    }
    getPerpMarkets() {
        return Array.from(this.perpMarkets
            .filter((perpMarket) => perpMarket.active)
            .map((perpMarket) => new web3_js.PublicKey(perpMarket.publicKey)));
    }
    static fromIdsByName(name) {
        const groupConfig = ids.groups.find((id) => id['name'] === name);
        if (!groupConfig)
            throw new Error(`No group config ${name} found in Ids!`);
        return new Id(groupConfig.cluster, groupConfig.name, groupConfig.publicKey, groupConfig.serum3ProgramId, groupConfig.mangoProgramId, groupConfig['banks'], groupConfig['stubOracles'], groupConfig['mintInfos'], groupConfig['serum3Markets'], groupConfig['perpMarkets']);
    }
    static fromIdsByPk(groupPk) {
        const groupConfig = ids.groups.find((id) => id['publicKey'] === groupPk.toString());
        if (!groupConfig)
            throw new Error(`No group config ${groupPk.toString()} found in Ids!`);
        return new Id(groupConfig.cluster, groupConfig.name, groupConfig.publicKey, groupConfig.serum3ProgramId, groupConfig.mangoProgramId, groupConfig['banks'], groupConfig['stubOracles'], groupConfig['mintInfos'], groupConfig['serum3Markets'], groupConfig['perpMarkets']);
    }
    static async fromApi(groupPk) {
        const groupMetadataApiUrl = 'https://mango-transaction-log.herokuapp.com/v4/group-metadata';
        const response = await fetch(groupMetadataApiUrl);
        const jsonData = await response.json();
        const groupConfig = jsonData.groups.find((group) => group.publicKey === groupPk.toString());
        return new Id(groupConfig.cluster, groupConfig.name, groupConfig.publicKey, groupConfig.serum3ProgramId, groupConfig.mangoProgramId, groupConfig.tokens.flatMap((t) => t.banks.map((b) => ({
            name: t.symbol,
            mint: t.mint,
            tokenIndex: t.tokenIndex,
            bankNum: b.bankNum,
            publicKey: b.publicKey,
        }))), groupConfig.stubOracles.map((s) => ({
            mint: s.mint,
            publicKey: s.publicKey,
        })), groupConfig.tokens.map((t) => ({
            name: t.symbol,
            mint: t.mint,
            tokenIndex: t.tokenIndex,
            publicKey: t.mintInfo,
        })), groupConfig.serum3Markets.map((s) => ({
            name: s.name,
            publicKey: s.publicKey,
            marketExternal: s.marketExternal,
        })), groupConfig.perpMarkets.map((p) => ({
            name: p.name,
            publicKey: p.publicKey,
        })));
    }
}

const IDL = {
    "version": "0.6.0",
    "name": "mango_v4",
    "instructions": [
        {
            "name": "groupCreate",
            "accounts": [
                {
                    "name": "group",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "Group"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "creator"
                            },
                            {
                                "kind": "arg",
                                "type": "u32",
                                "path": "group_num"
                            }
                        ]
                    }
                },
                {
                    "name": "creator",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "insuranceMint",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "insuranceVault",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "InsuranceVault"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            }
                        ]
                    }
                },
                {
                    "name": "payer",
                    "isMut": true,
                    "isSigner": true
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "systemProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "rent",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "groupNum",
                    "type": "u32"
                },
                {
                    "name": "testing",
                    "type": "u8"
                },
                {
                    "name": "version",
                    "type": "u8"
                }
            ]
        },
        {
            "name": "groupEdit",
            "accounts": [
                {
                    "name": "group",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                }
            ],
            "args": [
                {
                    "name": "adminOpt",
                    "type": {
                        "option": "publicKey"
                    }
                },
                {
                    "name": "fastListingAdminOpt",
                    "type": {
                        "option": "publicKey"
                    }
                },
                {
                    "name": "securityAdminOpt",
                    "type": {
                        "option": "publicKey"
                    }
                },
                {
                    "name": "testingOpt",
                    "type": {
                        "option": "u8"
                    }
                },
                {
                    "name": "versionOpt",
                    "type": {
                        "option": "u8"
                    }
                },
                {
                    "name": "depositLimitQuoteOpt",
                    "type": {
                        "option": "u64"
                    }
                }
            ]
        },
        {
            "name": "ixGateSet",
            "accounts": [
                {
                    "name": "group",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                }
            ],
            "args": [
                {
                    "name": "ixGate",
                    "type": "u128"
                }
            ]
        },
        {
            "name": "groupClose",
            "accounts": [
                {
                    "name": "group",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "insuranceVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "solDestination",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "tokenRegister",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "mint",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "bank",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "Bank"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "arg",
                                "type": "u16",
                                "path": "token_index"
                            },
                            {
                                "kind": "const",
                                "type": "u32",
                                "value": 0
                            }
                        ]
                    }
                },
                {
                    "name": "vault",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "Vault"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "arg",
                                "type": "u16",
                                "path": "token_index"
                            },
                            {
                                "kind": "const",
                                "type": "u32",
                                "value": 0
                            }
                        ]
                    }
                },
                {
                    "name": "mintInfo",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "MintInfo"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "account": "Mint",
                                "path": "mint"
                            }
                        ]
                    }
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "payer",
                    "isMut": true,
                    "isSigner": true
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "systemProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "rent",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "tokenIndex",
                    "type": "u16"
                },
                {
                    "name": "name",
                    "type": "string"
                },
                {
                    "name": "oracleConfig",
                    "type": {
                        "defined": "OracleConfigParams"
                    }
                },
                {
                    "name": "interestRateParams",
                    "type": {
                        "defined": "InterestRateParams"
                    }
                },
                {
                    "name": "loanFeeRate",
                    "type": "f32"
                },
                {
                    "name": "loanOriginationFeeRate",
                    "type": "f32"
                },
                {
                    "name": "maintAssetWeight",
                    "type": "f32"
                },
                {
                    "name": "initAssetWeight",
                    "type": "f32"
                },
                {
                    "name": "maintLiabWeight",
                    "type": "f32"
                },
                {
                    "name": "initLiabWeight",
                    "type": "f32"
                },
                {
                    "name": "liquidationFee",
                    "type": "f32"
                },
                {
                    "name": "minVaultToDepositsRatio",
                    "type": "f64"
                },
                {
                    "name": "netBorrowLimitWindowSizeTs",
                    "type": "u64"
                },
                {
                    "name": "netBorrowLimitPerWindowQuote",
                    "type": "i64"
                }
            ]
        },
        {
            "name": "tokenRegisterTrustless",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "fastListingAdmin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "mint",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "bank",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "Bank"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "arg",
                                "type": "u16",
                                "path": "token_index"
                            },
                            {
                                "kind": "const",
                                "type": "u32",
                                "value": 0
                            }
                        ]
                    }
                },
                {
                    "name": "vault",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "Vault"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "arg",
                                "type": "u16",
                                "path": "token_index"
                            },
                            {
                                "kind": "const",
                                "type": "u32",
                                "value": 0
                            }
                        ]
                    }
                },
                {
                    "name": "mintInfo",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "MintInfo"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "account": "Mint",
                                "path": "mint"
                            }
                        ]
                    }
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "payer",
                    "isMut": true,
                    "isSigner": true
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "systemProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "rent",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "tokenIndex",
                    "type": "u16"
                },
                {
                    "name": "name",
                    "type": "string"
                }
            ]
        },
        {
            "name": "tokenEdit",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "mintInfo",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false,
                    "docs": [
                        "The oracle account is optional and only used when reset_stable_price is set.",
                        ""
                    ]
                }
            ],
            "args": [
                {
                    "name": "oracleOpt",
                    "type": {
                        "option": "publicKey"
                    }
                },
                {
                    "name": "oracleConfigOpt",
                    "type": {
                        "option": {
                            "defined": "OracleConfigParams"
                        }
                    }
                },
                {
                    "name": "groupInsuranceFundOpt",
                    "type": {
                        "option": "bool"
                    }
                },
                {
                    "name": "interestRateParamsOpt",
                    "type": {
                        "option": {
                            "defined": "InterestRateParams"
                        }
                    }
                },
                {
                    "name": "loanFeeRateOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "loanOriginationFeeRateOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "maintAssetWeightOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "initAssetWeightOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "maintLiabWeightOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "initLiabWeightOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "liquidationFeeOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "stablePriceDelayIntervalSecondsOpt",
                    "type": {
                        "option": "u32"
                    }
                },
                {
                    "name": "stablePriceDelayGrowthLimitOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "stablePriceGrowthLimitOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "minVaultToDepositsRatioOpt",
                    "type": {
                        "option": "f64"
                    }
                },
                {
                    "name": "netBorrowLimitPerWindowQuoteOpt",
                    "type": {
                        "option": "i64"
                    }
                },
                {
                    "name": "netBorrowLimitWindowSizeTsOpt",
                    "type": {
                        "option": "u64"
                    }
                },
                {
                    "name": "borrowWeightScaleStartQuoteOpt",
                    "type": {
                        "option": "f64"
                    }
                },
                {
                    "name": "depositWeightScaleStartQuoteOpt",
                    "type": {
                        "option": "f64"
                    }
                },
                {
                    "name": "resetStablePrice",
                    "type": "bool"
                },
                {
                    "name": "resetNetBorrowLimit",
                    "type": "bool"
                },
                {
                    "name": "reduceOnlyOpt",
                    "type": {
                        "option": "bool"
                    }
                }
            ]
        },
        {
            "name": "tokenAddBank",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "mint",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "existingBank",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "bank",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "Bank"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "arg",
                                "type": "u16",
                                "path": "token_index"
                            },
                            {
                                "kind": "arg",
                                "type": "u32",
                                "path": "bank_num"
                            }
                        ]
                    }
                },
                {
                    "name": "vault",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "Vault"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "arg",
                                "type": "u16",
                                "path": "token_index"
                            },
                            {
                                "kind": "arg",
                                "type": "u32",
                                "path": "bank_num"
                            }
                        ]
                    }
                },
                {
                    "name": "mintInfo",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "payer",
                    "isMut": true,
                    "isSigner": true
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "systemProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "rent",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "tokenIndex",
                    "type": "u16"
                },
                {
                    "name": "bankNum",
                    "type": "u32"
                }
            ]
        },
        {
            "name": "tokenDeregister",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "mintInfo",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "dustVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "solDestination",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "tokenUpdateIndexAndRate",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "mintInfo",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "instructions",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "accountCreate",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "MangoAccount"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "owner"
                            },
                            {
                                "kind": "arg",
                                "type": "u32",
                                "path": "account_num"
                            }
                        ]
                    }
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "payer",
                    "isMut": true,
                    "isSigner": true
                },
                {
                    "name": "systemProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "accountNum",
                    "type": "u32"
                },
                {
                    "name": "tokenCount",
                    "type": "u8"
                },
                {
                    "name": "serum3Count",
                    "type": "u8"
                },
                {
                    "name": "perpCount",
                    "type": "u8"
                },
                {
                    "name": "perpOoCount",
                    "type": "u8"
                },
                {
                    "name": "name",
                    "type": "string"
                }
            ]
        },
        {
            "name": "accountExpand",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "payer",
                    "isMut": true,
                    "isSigner": true
                },
                {
                    "name": "systemProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "tokenCount",
                    "type": "u8"
                },
                {
                    "name": "serum3Count",
                    "type": "u8"
                },
                {
                    "name": "perpCount",
                    "type": "u8"
                },
                {
                    "name": "perpOoCount",
                    "type": "u8"
                }
            ]
        },
        {
            "name": "accountEdit",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                }
            ],
            "args": [
                {
                    "name": "nameOpt",
                    "type": {
                        "option": "string"
                    }
                },
                {
                    "name": "delegateOpt",
                    "type": {
                        "option": "publicKey"
                    }
                }
            ]
        },
        {
            "name": "accountToggleFreeze",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                }
            ],
            "args": [
                {
                    "name": "freeze",
                    "type": "bool"
                }
            ]
        },
        {
            "name": "accountClose",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "solDestination",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "forceClose",
                    "type": "bool"
                }
            ]
        },
        {
            "name": "stubOracleCreate",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "StubOracle"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "account": "Mint",
                                "path": "mint"
                            }
                        ]
                    }
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "mint",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "payer",
                    "isMut": true,
                    "isSigner": true
                },
                {
                    "name": "systemProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "price",
                    "type": {
                        "defined": "I80F48"
                    }
                }
            ]
        },
        {
            "name": "stubOracleClose",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "oracle",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "solDestination",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "stubOracleSet",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "oracle",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "price",
                    "type": {
                        "defined": "I80F48"
                    }
                }
            ]
        },
        {
            "name": "tokenDeposit",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "bank",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "vault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "tokenAccount",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenAuthority",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "amount",
                    "type": "u64"
                },
                {
                    "name": "reduceOnly",
                    "type": "bool"
                }
            ]
        },
        {
            "name": "tokenDepositIntoExisting",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "bank",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "vault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "tokenAccount",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenAuthority",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "amount",
                    "type": "u64"
                },
                {
                    "name": "reduceOnly",
                    "type": "bool"
                }
            ]
        },
        {
            "name": "tokenWithdraw",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "bank",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "vault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "tokenAccount",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "amount",
                    "type": "u64"
                },
                {
                    "name": "allowBorrow",
                    "type": "bool"
                }
            ]
        },
        {
            "name": "flashLoanBegin",
            "accounts": [
                {
                    "name": "account",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "instructions",
                    "isMut": false,
                    "isSigner": false,
                    "docs": [
                        "Instructions Sysvar for instruction introspection"
                    ]
                }
            ],
            "args": [
                {
                    "name": "loanAmounts",
                    "type": {
                        "vec": "u64"
                    }
                }
            ]
        },
        {
            "name": "flashLoanEnd",
            "accounts": [
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "flashLoanType",
                    "type": {
                        "defined": "FlashLoanType"
                    }
                }
            ]
        },
        {
            "name": "healthRegionBegin",
            "accounts": [
                {
                    "name": "instructions",
                    "isMut": false,
                    "isSigner": false,
                    "docs": [
                        "Instructions Sysvar for instruction introspection"
                    ]
                },
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "healthRegionEnd",
            "accounts": [
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "serum3RegisterMarket",
            "docs": [
                "",
                "Serum",
                ""
            ],
            "accounts": [
                {
                    "name": "group",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "serumProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumMarketExternal",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumMarket",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "Serum3Market"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "serum_market_external"
                            }
                        ]
                    }
                },
                {
                    "name": "indexReservation",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "Serum3Index"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "arg",
                                "type": "u16",
                                "path": "market_index"
                            }
                        ]
                    }
                },
                {
                    "name": "quoteBank",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "baseBank",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "payer",
                    "isMut": true,
                    "isSigner": true
                },
                {
                    "name": "systemProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "marketIndex",
                    "type": "u16"
                },
                {
                    "name": "name",
                    "type": "string"
                }
            ]
        },
        {
            "name": "serum3EditMarket",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "market",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "reduceOnlyOpt",
                    "type": {
                        "option": "bool"
                    }
                }
            ]
        },
        {
            "name": "serum3DeregisterMarket",
            "accounts": [
                {
                    "name": "group",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "serumMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "indexReservation",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "solDestination",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "serum3CreateOpenOrders",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "serumMarket",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumMarketExternal",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "openOrders",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "Serum3OO"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "account"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "serum_market"
                            }
                        ]
                    }
                },
                {
                    "name": "payer",
                    "isMut": true,
                    "isSigner": true
                },
                {
                    "name": "systemProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "rent",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "serum3CloseOpenOrders",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "serumMarket",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumMarketExternal",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "openOrders",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "solDestination",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "serum3PlaceOrder",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "openOrders",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "serumMarket",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumMarketExternal",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketBids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketAsks",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketEventQueue",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketRequestQueue",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketBaseVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketQuoteVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketVaultSigner",
                    "isMut": false,
                    "isSigner": false,
                    "docs": [
                        "needed for the automatic settle_funds call"
                    ]
                },
                {
                    "name": "payerBank",
                    "isMut": true,
                    "isSigner": false,
                    "docs": [
                        "The bank that pays for the order, if necessary"
                    ]
                },
                {
                    "name": "payerVault",
                    "isMut": true,
                    "isSigner": false,
                    "docs": [
                        "The bank vault that pays for the order, if necessary"
                    ]
                },
                {
                    "name": "payerOracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "side",
                    "type": {
                        "defined": "Serum3Side"
                    }
                },
                {
                    "name": "limitPrice",
                    "type": "u64"
                },
                {
                    "name": "maxBaseQty",
                    "type": "u64"
                },
                {
                    "name": "maxNativeQuoteQtyIncludingFees",
                    "type": "u64"
                },
                {
                    "name": "selfTradeBehavior",
                    "type": {
                        "defined": "Serum3SelfTradeBehavior"
                    }
                },
                {
                    "name": "orderType",
                    "type": {
                        "defined": "Serum3OrderType"
                    }
                },
                {
                    "name": "clientOrderId",
                    "type": "u64"
                },
                {
                    "name": "limit",
                    "type": "u16"
                }
            ]
        },
        {
            "name": "serum3CancelOrder",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "openOrders",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "serumMarket",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumMarketExternal",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketBids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketAsks",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketEventQueue",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "side",
                    "type": {
                        "defined": "Serum3Side"
                    }
                },
                {
                    "name": "orderId",
                    "type": "u128"
                }
            ]
        },
        {
            "name": "serum3CancelAllOrders",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "openOrders",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "serumMarket",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumMarketExternal",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketBids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketAsks",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketEventQueue",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "limit",
                    "type": "u8"
                }
            ]
        },
        {
            "name": "serum3SettleFunds",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "openOrders",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "serumMarket",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumMarketExternal",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketBaseVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketQuoteVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketVaultSigner",
                    "isMut": false,
                    "isSigner": false,
                    "docs": [
                        "needed for the automatic settle_funds call"
                    ]
                },
                {
                    "name": "quoteBank",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "quoteVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "baseBank",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "baseVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "serum3LiqForceCancelOrders",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "openOrders",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "serumMarket",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumProgram",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "serumMarketExternal",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketBids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketAsks",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketEventQueue",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketBaseVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketQuoteVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "marketVaultSigner",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "quoteBank",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "quoteVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "baseBank",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "baseVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "limit",
                    "type": "u8"
                }
            ]
        },
        {
            "name": "liqTokenWithToken",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "liqor",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "liqorOwner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "liqee",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "assetTokenIndex",
                    "type": "u16"
                },
                {
                    "name": "liabTokenIndex",
                    "type": "u16"
                },
                {
                    "name": "maxLiabTransfer",
                    "type": {
                        "defined": "I80F48"
                    }
                }
            ]
        },
        {
            "name": "liqTokenBankruptcy",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "liqor",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "liqorOwner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "liqee",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "liabMintInfo",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "quoteVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "insuranceVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "maxLiabTransfer",
                    "type": {
                        "defined": "I80F48"
                    }
                }
            ]
        },
        {
            "name": "tokenLiqWithToken",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "liqor",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "liqorOwner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "liqee",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "assetTokenIndex",
                    "type": "u16"
                },
                {
                    "name": "liabTokenIndex",
                    "type": "u16"
                },
                {
                    "name": "maxLiabTransfer",
                    "type": {
                        "defined": "I80F48"
                    }
                }
            ]
        },
        {
            "name": "tokenLiqBankruptcy",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "liqor",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "liqorOwner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "liqee",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "liabMintInfo",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "quoteVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "insuranceVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "maxLiabTransfer",
                    "type": {
                        "defined": "I80F48"
                    }
                }
            ]
        },
        {
            "name": "perpCreateMarket",
            "docs": [
                "",
                "Perps",
                ""
            ],
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false,
                    "pda": {
                        "seeds": [
                            {
                                "kind": "const",
                                "type": "string",
                                "value": "PerpMarket"
                            },
                            {
                                "kind": "account",
                                "type": "publicKey",
                                "path": "group"
                            },
                            {
                                "kind": "arg",
                                "type": "u16",
                                "path": "perp_market_index"
                            }
                        ]
                    }
                },
                {
                    "name": "bids",
                    "isMut": true,
                    "isSigner": false,
                    "docs": [
                        "Accounts are initialised by client,",
                        "anchor discriminator is set first when ix exits,"
                    ]
                },
                {
                    "name": "asks",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "eventQueue",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "payer",
                    "isMut": true,
                    "isSigner": true
                },
                {
                    "name": "systemProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "perpMarketIndex",
                    "type": "u16"
                },
                {
                    "name": "name",
                    "type": "string"
                },
                {
                    "name": "oracleConfig",
                    "type": {
                        "defined": "OracleConfigParams"
                    }
                },
                {
                    "name": "baseDecimals",
                    "type": "u8"
                },
                {
                    "name": "quoteLotSize",
                    "type": "i64"
                },
                {
                    "name": "baseLotSize",
                    "type": "i64"
                },
                {
                    "name": "maintBaseAssetWeight",
                    "type": "f32"
                },
                {
                    "name": "initBaseAssetWeight",
                    "type": "f32"
                },
                {
                    "name": "maintBaseLiabWeight",
                    "type": "f32"
                },
                {
                    "name": "initBaseLiabWeight",
                    "type": "f32"
                },
                {
                    "name": "maintOverallAssetWeight",
                    "type": "f32"
                },
                {
                    "name": "initOverallAssetWeight",
                    "type": "f32"
                },
                {
                    "name": "baseLiquidationFee",
                    "type": "f32"
                },
                {
                    "name": "makerFee",
                    "type": "f32"
                },
                {
                    "name": "takerFee",
                    "type": "f32"
                },
                {
                    "name": "minFunding",
                    "type": "f32"
                },
                {
                    "name": "maxFunding",
                    "type": "f32"
                },
                {
                    "name": "impactQuantity",
                    "type": "i64"
                },
                {
                    "name": "groupInsuranceFund",
                    "type": "bool"
                },
                {
                    "name": "feePenalty",
                    "type": "f32"
                },
                {
                    "name": "settleFeeFlat",
                    "type": "f32"
                },
                {
                    "name": "settleFeeAmountThreshold",
                    "type": "f32"
                },
                {
                    "name": "settleFeeFractionLowHealth",
                    "type": "f32"
                },
                {
                    "name": "settleTokenIndex",
                    "type": "u16"
                },
                {
                    "name": "settlePnlLimitFactor",
                    "type": "f32"
                },
                {
                    "name": "settlePnlLimitWindowSizeTs",
                    "type": "u64"
                },
                {
                    "name": "positivePnlLiquidationFee",
                    "type": "f32"
                }
            ]
        },
        {
            "name": "perpEditMarket",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false,
                    "docs": [
                        "The oracle account is optional and only used when reset_stable_price is set.",
                        ""
                    ]
                }
            ],
            "args": [
                {
                    "name": "oracleOpt",
                    "type": {
                        "option": "publicKey"
                    }
                },
                {
                    "name": "oracleConfigOpt",
                    "type": {
                        "option": {
                            "defined": "OracleConfigParams"
                        }
                    }
                },
                {
                    "name": "baseDecimalsOpt",
                    "type": {
                        "option": "u8"
                    }
                },
                {
                    "name": "maintBaseAssetWeightOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "initBaseAssetWeightOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "maintBaseLiabWeightOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "initBaseLiabWeightOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "maintOverallAssetWeightOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "initOverallAssetWeightOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "baseLiquidationFeeOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "makerFeeOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "takerFeeOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "minFundingOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "maxFundingOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "impactQuantityOpt",
                    "type": {
                        "option": "i64"
                    }
                },
                {
                    "name": "groupInsuranceFundOpt",
                    "type": {
                        "option": "bool"
                    }
                },
                {
                    "name": "feePenaltyOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "settleFeeFlatOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "settleFeeAmountThresholdOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "settleFeeFractionLowHealthOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "stablePriceDelayIntervalSecondsOpt",
                    "type": {
                        "option": "u32"
                    }
                },
                {
                    "name": "stablePriceDelayGrowthLimitOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "stablePriceGrowthLimitOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "settlePnlLimitFactorOpt",
                    "type": {
                        "option": "f32"
                    }
                },
                {
                    "name": "settlePnlLimitWindowSizeTsOpt",
                    "type": {
                        "option": "u64"
                    }
                },
                {
                    "name": "reduceOnlyOpt",
                    "type": {
                        "option": "bool"
                    }
                },
                {
                    "name": "resetStablePrice",
                    "type": "bool"
                },
                {
                    "name": "positivePnlLiquidationFeeOpt",
                    "type": {
                        "option": "f32"
                    }
                }
            ]
        },
        {
            "name": "perpCloseMarket",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "bids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "asks",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "eventQueue",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "solDestination",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "perpDeactivatePosition",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "perpMarket",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "perpPlaceOrder",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "bids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "asks",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "eventQueue",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "side",
                    "type": {
                        "defined": "Side"
                    }
                },
                {
                    "name": "priceLots",
                    "type": "i64"
                },
                {
                    "name": "maxBaseLots",
                    "type": "i64"
                },
                {
                    "name": "maxQuoteLots",
                    "type": "i64"
                },
                {
                    "name": "clientOrderId",
                    "type": "u64"
                },
                {
                    "name": "orderType",
                    "type": {
                        "defined": "PlaceOrderType"
                    }
                },
                {
                    "name": "reduceOnly",
                    "type": "bool"
                },
                {
                    "name": "expiryTimestamp",
                    "type": "u64"
                },
                {
                    "name": "limit",
                    "type": "u8"
                }
            ],
            "returns": {
                "option": "u128"
            }
        },
        {
            "name": "perpPlaceOrderPegged",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "bids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "asks",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "eventQueue",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "side",
                    "type": {
                        "defined": "Side"
                    }
                },
                {
                    "name": "priceOffsetLots",
                    "type": "i64"
                },
                {
                    "name": "pegLimit",
                    "type": "i64"
                },
                {
                    "name": "maxBaseLots",
                    "type": "i64"
                },
                {
                    "name": "maxQuoteLots",
                    "type": "i64"
                },
                {
                    "name": "clientOrderId",
                    "type": "u64"
                },
                {
                    "name": "orderType",
                    "type": {
                        "defined": "PlaceOrderType"
                    }
                },
                {
                    "name": "reduceOnly",
                    "type": "bool"
                },
                {
                    "name": "expiryTimestamp",
                    "type": "u64"
                },
                {
                    "name": "limit",
                    "type": "u8"
                },
                {
                    "name": "maxOracleStalenessSlots",
                    "type": "i32"
                }
            ],
            "returns": {
                "option": "u128"
            }
        },
        {
            "name": "perpCancelOrder",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "bids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "asks",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "orderId",
                    "type": "u128"
                }
            ]
        },
        {
            "name": "perpCancelOrderByClientOrderId",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "bids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "asks",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "clientOrderId",
                    "type": "u64"
                }
            ]
        },
        {
            "name": "perpCancelAllOrders",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "bids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "asks",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "limit",
                    "type": "u8"
                }
            ]
        },
        {
            "name": "perpCancelAllOrdersBySide",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "owner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "bids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "asks",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "sideOption",
                    "type": {
                        "option": {
                            "defined": "Side"
                        }
                    }
                },
                {
                    "name": "limit",
                    "type": "u8"
                }
            ]
        },
        {
            "name": "perpConsumeEvents",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "eventQueue",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "limit",
                    "type": "u64"
                }
            ]
        },
        {
            "name": "perpUpdateFunding",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "bids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "asks",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "perpSettlePnl",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "settler",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "settlerOwner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "perpMarket",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "accountA",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "accountB",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "settleBank",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "settleOracle",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "perpSettleFees",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "settleBank",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "settleOracle",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "maxSettleAmount",
                    "type": "u64"
                }
            ]
        },
        {
            "name": "perpLiqBaseOrPositivePnl",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "liqor",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "liqorOwner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "liqee",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "settleBank",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "settleVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "settleOracle",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "maxBaseTransfer",
                    "type": "i64"
                },
                {
                    "name": "maxPnlTransfer",
                    "type": "u64"
                }
            ]
        },
        {
            "name": "perpLiqForceCancelOrders",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "bids",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "asks",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "limit",
                    "type": "u8"
                }
            ]
        },
        {
            "name": "perpLiqNegativePnlOrBankruptcy",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "liqor",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "liqorOwner",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "liqee",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "perpMarket",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "oracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "settleBank",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "settleVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "settleOracle",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "insuranceVault",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "tokenProgram",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "maxLiabTransfer",
                    "type": "u64"
                }
            ]
        },
        {
            "name": "altSet",
            "accounts": [
                {
                    "name": "group",
                    "isMut": true,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "addressLookupTable",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "index",
                    "type": "u8"
                }
            ]
        },
        {
            "name": "altExtend",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "admin",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "payer",
                    "isMut": false,
                    "isSigner": true
                },
                {
                    "name": "addressLookupTable",
                    "isMut": true,
                    "isSigner": false
                }
            ],
            "args": [
                {
                    "name": "index",
                    "type": "u8"
                },
                {
                    "name": "newAddresses",
                    "type": {
                        "vec": "publicKey"
                    }
                }
            ]
        },
        {
            "name": "computeAccountData",
            "accounts": [
                {
                    "name": "group",
                    "isMut": false,
                    "isSigner": false
                },
                {
                    "name": "account",
                    "isMut": false,
                    "isSigner": false
                }
            ],
            "args": []
        },
        {
            "name": "benchmark",
            "docs": [
                "",
                "benchmark",
                ""
            ],
            "accounts": [],
            "args": []
        }
    ],
    "accounts": [
        {
            "name": "bank",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "group",
                        "type": "publicKey"
                    },
                    {
                        "name": "name",
                        "type": {
                            "array": [
                                "u8",
                                16
                            ]
                        }
                    },
                    {
                        "name": "mint",
                        "type": "publicKey"
                    },
                    {
                        "name": "vault",
                        "type": "publicKey"
                    },
                    {
                        "name": "oracle",
                        "type": "publicKey"
                    },
                    {
                        "name": "oracleConfig",
                        "type": {
                            "defined": "OracleConfig"
                        }
                    },
                    {
                        "name": "stablePriceModel",
                        "type": {
                            "defined": "StablePriceModel"
                        }
                    },
                    {
                        "name": "depositIndex",
                        "docs": [
                            "the index used to scale the value of an IndexedPosition",
                            "TODO: should always be >= 0, add checks?"
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "borrowIndex",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "indexedDeposits",
                        "docs": [
                            "deposits/borrows for this bank",
                            "",
                            "Note that these may become negative. It's perfectly fine for users to borrow one one bank",
                            "(increasing indexed_borrows there) and paying back on another (possibly decreasing indexed_borrows",
                            "below zero).",
                            "",
                            "The vault amount is not deducable from these values.",
                            "",
                            "These become meaningful when summed over all banks (like in update_index_and_rate)."
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "indexedBorrows",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "indexLastUpdated",
                        "type": "u64"
                    },
                    {
                        "name": "bankRateLastUpdated",
                        "type": "u64"
                    },
                    {
                        "name": "avgUtilization",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "adjustmentFactor",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "util0",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "rate0",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "util1",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "rate1",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "maxRate",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "collectedFeesNative",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "loanOriginationFeeRate",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "loanFeeRate",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "maintAssetWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "initAssetWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "maintLiabWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "initLiabWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "liquidationFee",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "dust",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "flashLoanTokenAccountInitial",
                        "type": "u64"
                    },
                    {
                        "name": "flashLoanApprovedAmount",
                        "type": "u64"
                    },
                    {
                        "name": "tokenIndex",
                        "type": "u16"
                    },
                    {
                        "name": "bump",
                        "type": "u8"
                    },
                    {
                        "name": "mintDecimals",
                        "type": "u8"
                    },
                    {
                        "name": "bankNum",
                        "type": "u32"
                    },
                    {
                        "name": "minVaultToDepositsRatio",
                        "docs": [
                            "Min fraction of deposits that must remain in the vault when borrowing."
                        ],
                        "type": "f64"
                    },
                    {
                        "name": "netBorrowLimitWindowSizeTs",
                        "docs": [
                            "Size in seconds of a net borrows window"
                        ],
                        "type": "u64"
                    },
                    {
                        "name": "lastNetBorrowsWindowStartTs",
                        "docs": [
                            "Timestamp at which the last net borrows window started"
                        ],
                        "type": "u64"
                    },
                    {
                        "name": "netBorrowLimitPerWindowQuote",
                        "docs": [
                            "Net borrow limit per window in quote native; set to -1 to disable."
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "netBorrowsInWindow",
                        "docs": [
                            "Sum of all deposits and borrows in the last window, in native units."
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "borrowWeightScaleStartQuote",
                        "docs": [
                            "Soft borrow limit in native quote",
                            "",
                            "Once the borrows on the bank exceed this quote value, init_liab_weight is scaled up.",
                            "Set to f64::MAX to disable.",
                            "",
                            "See scaled_init_liab_weight()."
                        ],
                        "type": "f64"
                    },
                    {
                        "name": "depositWeightScaleStartQuote",
                        "docs": [
                            "Limit for collateral of deposits in native quote",
                            "",
                            "Once the deposits in the bank exceed this quote value, init_asset_weight is scaled",
                            "down to keep the total collateral value constant.",
                            "Set to f64::MAX to disable.",
                            "",
                            "See scaled_init_asset_weight()."
                        ],
                        "type": "f64"
                    },
                    {
                        "name": "reduceOnly",
                        "type": "u8"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                2119
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "group",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "creator",
                        "type": "publicKey"
                    },
                    {
                        "name": "groupNum",
                        "type": "u32"
                    },
                    {
                        "name": "admin",
                        "type": "publicKey"
                    },
                    {
                        "name": "fastListingAdmin",
                        "type": "publicKey"
                    },
                    {
                        "name": "padding",
                        "type": {
                            "array": [
                                "u8",
                                4
                            ]
                        }
                    },
                    {
                        "name": "insuranceVault",
                        "type": "publicKey"
                    },
                    {
                        "name": "insuranceMint",
                        "type": "publicKey"
                    },
                    {
                        "name": "bump",
                        "type": "u8"
                    },
                    {
                        "name": "testing",
                        "type": "u8"
                    },
                    {
                        "name": "version",
                        "type": "u8"
                    },
                    {
                        "name": "padding2",
                        "type": {
                            "array": [
                                "u8",
                                5
                            ]
                        }
                    },
                    {
                        "name": "addressLookupTables",
                        "type": {
                            "array": [
                                "publicKey",
                                20
                            ]
                        }
                    },
                    {
                        "name": "securityAdmin",
                        "type": "publicKey"
                    },
                    {
                        "name": "depositLimitQuote",
                        "type": "u64"
                    },
                    {
                        "name": "ixGate",
                        "type": "u128"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                1864
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "mangoAccount",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "group",
                        "type": "publicKey"
                    },
                    {
                        "name": "owner",
                        "type": "publicKey"
                    },
                    {
                        "name": "name",
                        "type": {
                            "array": [
                                "u8",
                                32
                            ]
                        }
                    },
                    {
                        "name": "delegate",
                        "type": "publicKey"
                    },
                    {
                        "name": "accountNum",
                        "type": "u32"
                    },
                    {
                        "name": "beingLiquidated",
                        "docs": [
                            "Tracks that this account should be liquidated until init_health >= 0.",
                            "",
                            "Normally accounts can not be liquidated while maint_health >= 0. But when an account",
                            "reaches maint_health < 0, liquidators will call a liquidation instruction and thereby",
                            "set this flag. Now the account may be liquidated until init_health >= 0.",
                            "",
                            "Many actions should be disabled while the account is being liquidated, even if",
                            "its maint health has recovered to positive. Creating new open orders would, for example,",
                            "confuse liquidators."
                        ],
                        "type": "u8"
                    },
                    {
                        "name": "inHealthRegion",
                        "docs": [
                            "The account is currently inside a health region marked by HealthRegionBegin...HealthRegionEnd.",
                            "",
                            "Must never be set after a transaction ends."
                        ],
                        "type": "u8"
                    },
                    {
                        "name": "bump",
                        "type": "u8"
                    },
                    {
                        "name": "padding",
                        "type": {
                            "array": [
                                "u8",
                                1
                            ]
                        }
                    },
                    {
                        "name": "netDeposits",
                        "type": "i64"
                    },
                    {
                        "name": "perpSpotTransfers",
                        "type": "i64"
                    },
                    {
                        "name": "healthRegionBeginInitHealth",
                        "docs": [
                            "Init health as calculated during HealthReginBegin, rounded up."
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "frozenUntil",
                        "type": "u64"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                232
                            ]
                        }
                    },
                    {
                        "name": "headerVersion",
                        "type": "u8"
                    },
                    {
                        "name": "padding3",
                        "type": {
                            "array": [
                                "u8",
                                7
                            ]
                        }
                    },
                    {
                        "name": "padding4",
                        "type": "u32"
                    },
                    {
                        "name": "tokens",
                        "type": {
                            "vec": {
                                "defined": "TokenPosition"
                            }
                        }
                    },
                    {
                        "name": "padding5",
                        "type": "u32"
                    },
                    {
                        "name": "serum3",
                        "type": {
                            "vec": {
                                "defined": "Serum3Orders"
                            }
                        }
                    },
                    {
                        "name": "padding6",
                        "type": "u32"
                    },
                    {
                        "name": "perps",
                        "type": {
                            "vec": {
                                "defined": "PerpPosition"
                            }
                        }
                    },
                    {
                        "name": "padding7",
                        "type": "u32"
                    },
                    {
                        "name": "perpOpenOrders",
                        "type": {
                            "vec": {
                                "defined": "PerpOpenOrder"
                            }
                        }
                    }
                ]
            }
        },
        {
            "name": "mintInfo",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "group",
                        "type": "publicKey"
                    },
                    {
                        "name": "tokenIndex",
                        "type": "u16"
                    },
                    {
                        "name": "groupInsuranceFund",
                        "type": "u8"
                    },
                    {
                        "name": "padding1",
                        "type": {
                            "array": [
                                "u8",
                                5
                            ]
                        }
                    },
                    {
                        "name": "mint",
                        "type": "publicKey"
                    },
                    {
                        "name": "banks",
                        "type": {
                            "array": [
                                "publicKey",
                                6
                            ]
                        }
                    },
                    {
                        "name": "vaults",
                        "type": {
                            "array": [
                                "publicKey",
                                6
                            ]
                        }
                    },
                    {
                        "name": "oracle",
                        "type": "publicKey"
                    },
                    {
                        "name": "registrationTime",
                        "type": "u64"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                2560
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "stubOracle",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "group",
                        "type": "publicKey"
                    },
                    {
                        "name": "mint",
                        "type": "publicKey"
                    },
                    {
                        "name": "price",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "lastUpdated",
                        "type": "i64"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                128
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "bookSide",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "roots",
                        "type": {
                            "array": [
                                {
                                    "defined": "OrderTreeRoot"
                                },
                                2
                            ]
                        }
                    },
                    {
                        "name": "reservedRoots",
                        "type": {
                            "array": [
                                {
                                    "defined": "OrderTreeRoot"
                                },
                                4
                            ]
                        }
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                256
                            ]
                        }
                    },
                    {
                        "name": "nodes",
                        "type": {
                            "defined": "OrderTreeNodes"
                        }
                    }
                ]
            }
        },
        {
            "name": "eventQueue",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "header",
                        "type": {
                            "defined": "EventQueueHeader"
                        }
                    },
                    {
                        "name": "buf",
                        "type": {
                            "array": [
                                {
                                    "defined": "AnyEvent"
                                },
                                488
                            ]
                        }
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                64
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "perpMarket",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "group",
                        "type": "publicKey"
                    },
                    {
                        "name": "settleTokenIndex",
                        "docs": [
                            "Token index that settlements happen in.",
                            "",
                            "Currently required to be 0, USDC. In the future settlement",
                            "may be allowed to happen in other tokens."
                        ],
                        "type": "u16"
                    },
                    {
                        "name": "perpMarketIndex",
                        "docs": [
                            "Index of this perp market. Other data, like the MangoAccount's PerpPosition",
                            "reference this market via this index. Unique for this group's perp markets."
                        ],
                        "type": "u16"
                    },
                    {
                        "name": "blocked1",
                        "docs": [
                            "Field used to contain the trusted_market flag and is now unused."
                        ],
                        "type": "u8"
                    },
                    {
                        "name": "groupInsuranceFund",
                        "docs": [
                            "Is this market covered by the group insurance fund?"
                        ],
                        "type": "u8"
                    },
                    {
                        "name": "bump",
                        "docs": [
                            "PDA bump"
                        ],
                        "type": "u8"
                    },
                    {
                        "name": "baseDecimals",
                        "docs": [
                            "Number of decimals used for the base token.",
                            "",
                            "Used to convert the oracle's price into a native/native price."
                        ],
                        "type": "u8"
                    },
                    {
                        "name": "name",
                        "docs": [
                            "Name. Trailing zero bytes are ignored."
                        ],
                        "type": {
                            "array": [
                                "u8",
                                16
                            ]
                        }
                    },
                    {
                        "name": "bids",
                        "docs": [
                            "Address of the BookSide account for bids"
                        ],
                        "type": "publicKey"
                    },
                    {
                        "name": "asks",
                        "docs": [
                            "Address of the BookSide account for asks"
                        ],
                        "type": "publicKey"
                    },
                    {
                        "name": "eventQueue",
                        "docs": [
                            "Address of the EventQueue account"
                        ],
                        "type": "publicKey"
                    },
                    {
                        "name": "oracle",
                        "docs": [
                            "Oracle account address"
                        ],
                        "type": "publicKey"
                    },
                    {
                        "name": "oracleConfig",
                        "docs": [
                            "Oracle configuration"
                        ],
                        "type": {
                            "defined": "OracleConfig"
                        }
                    },
                    {
                        "name": "stablePriceModel",
                        "docs": [
                            "Maintains a stable price based on the oracle price that is less volatile."
                        ],
                        "type": {
                            "defined": "StablePriceModel"
                        }
                    },
                    {
                        "name": "quoteLotSize",
                        "docs": [
                            "Number of quote native in a quote lot. Must be a power of 10.",
                            "",
                            "Primarily useful for increasing the tick size on the market: A lot price",
                            "of 1 becomes a native price of quote_lot_size/base_lot_size becomes a",
                            "ui price of quote_lot_size*base_decimals/base_lot_size/quote_decimals."
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "baseLotSize",
                        "docs": [
                            "Number of base native in a base lot. Must be a power of 10.",
                            "",
                            "Example: If base decimals for the underlying asset is 6, base lot size",
                            "is 100 and and base position lots is 10_000 then base position native is",
                            "1_000_000 and base position ui is 1."
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "maintBaseAssetWeight",
                        "docs": [
                            "These weights apply to the base position. The quote position has",
                            "no explicit weight (but may be covered by the overall pnl asset weight)."
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "initBaseAssetWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "maintBaseLiabWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "initBaseLiabWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "openInterest",
                        "docs": [
                            "Number of base lot pairs currently active in the market. Always >= 0."
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "seqNum",
                        "docs": [
                            "Total number of orders seen"
                        ],
                        "type": "u64"
                    },
                    {
                        "name": "registrationTime",
                        "docs": [
                            "Timestamp in seconds that the market was registered at."
                        ],
                        "type": "u64"
                    },
                    {
                        "name": "minFunding",
                        "docs": [
                            "Minimal funding rate per day, must be <= 0."
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "maxFunding",
                        "docs": [
                            "Maximal funding rate per day, must be >= 0."
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "impactQuantity",
                        "docs": [
                            "For funding, get the impact price this many base lots deep into the book."
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "longFunding",
                        "docs": [
                            "Current long funding value. Increasing it means that every long base lot",
                            "needs to pay that amount in funding.",
                            "",
                            "PerpPosition uses and tracks it settle funding. Updated by the perp",
                            "keeper instruction."
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "shortFunding",
                        "docs": [
                            "See long_funding."
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "fundingLastUpdated",
                        "docs": [
                            "timestamp that funding was last updated in"
                        ],
                        "type": "u64"
                    },
                    {
                        "name": "baseLiquidationFee",
                        "docs": [
                            "Fees",
                            "Fee for base position liquidation"
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "makerFee",
                        "docs": [
                            "Fee when matching maker orders. May be negative."
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "takerFee",
                        "docs": [
                            "Fee for taker orders, may not be negative."
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "feesAccrued",
                        "docs": [
                            "Fees accrued in native quote currency"
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "feesSettled",
                        "docs": [
                            "Fees settled in native quote currency"
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "feePenalty",
                        "docs": [
                            "Fee (in quote native) to charge for ioc orders"
                        ],
                        "type": "f32"
                    },
                    {
                        "name": "settleFeeFlat",
                        "docs": [
                            "In native units of settlement token, given to each settle call above the",
                            "settle_fee_amount_threshold."
                        ],
                        "type": "f32"
                    },
                    {
                        "name": "settleFeeAmountThreshold",
                        "docs": [
                            "Pnl settlement amount needed to be eligible for fees."
                        ],
                        "type": "f32"
                    },
                    {
                        "name": "settleFeeFractionLowHealth",
                        "docs": [
                            "Fraction of pnl to pay out as fee if +pnl account has low health."
                        ],
                        "type": "f32"
                    },
                    {
                        "name": "settlePnlLimitFactor",
                        "docs": [
                            "Controls the strictness of the settle limit.",
                            "Set to a negative value to disable the limit.",
                            "",
                            "This factor applies to the settle limit in two ways",
                            "- for the unrealized pnl settle limit, the factor is multiplied with the stable perp base value",
                            "(i.e. limit_factor * base_native * stable_price)",
                            "- when increasing the realized pnl settle limit (stored per PerpPosition), the factor is",
                            "multiplied with the stable value of the perp pnl being realized",
                            "(i.e. limit_factor * reduced_native * stable_price)",
                            "",
                            "See also PerpPosition::settle_pnl_limit_realized_trade"
                        ],
                        "type": "f32"
                    },
                    {
                        "name": "padding3",
                        "type": {
                            "array": [
                                "u8",
                                4
                            ]
                        }
                    },
                    {
                        "name": "settlePnlLimitWindowSizeTs",
                        "docs": [
                            "Window size in seconds for the perp settlement limit"
                        ],
                        "type": "u64"
                    },
                    {
                        "name": "reduceOnly",
                        "docs": [
                            "If true, users may no longer increase their market exposure. Only actions",
                            "that reduce their position are still allowed."
                        ],
                        "type": "u8"
                    },
                    {
                        "name": "padding4",
                        "type": {
                            "array": [
                                "u8",
                                7
                            ]
                        }
                    },
                    {
                        "name": "maintOverallAssetWeight",
                        "docs": [
                            "Weights for full perp market health, if positive"
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "initOverallAssetWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "positivePnlLiquidationFee",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                1888
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "serum3Market",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "group",
                        "type": "publicKey"
                    },
                    {
                        "name": "baseTokenIndex",
                        "type": "u16"
                    },
                    {
                        "name": "quoteTokenIndex",
                        "type": "u16"
                    },
                    {
                        "name": "reduceOnly",
                        "type": "u8"
                    },
                    {
                        "name": "padding1",
                        "type": {
                            "array": [
                                "u8",
                                3
                            ]
                        }
                    },
                    {
                        "name": "name",
                        "type": {
                            "array": [
                                "u8",
                                16
                            ]
                        }
                    },
                    {
                        "name": "serumProgram",
                        "type": "publicKey"
                    },
                    {
                        "name": "serumMarketExternal",
                        "type": "publicKey"
                    },
                    {
                        "name": "marketIndex",
                        "type": "u16"
                    },
                    {
                        "name": "bump",
                        "type": "u8"
                    },
                    {
                        "name": "padding2",
                        "type": {
                            "array": [
                                "u8",
                                5
                            ]
                        }
                    },
                    {
                        "name": "registrationTime",
                        "type": "u64"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                128
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "serum3MarketIndexReservation",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "group",
                        "type": "publicKey"
                    },
                    {
                        "name": "marketIndex",
                        "type": "u16"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                38
                            ]
                        }
                    }
                ]
            }
        }
    ],
    "types": [
        {
            "name": "InterestRateParams",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "util0",
                        "type": "f32"
                    },
                    {
                        "name": "rate0",
                        "type": "f32"
                    },
                    {
                        "name": "util1",
                        "type": "f32"
                    },
                    {
                        "name": "rate1",
                        "type": "f32"
                    },
                    {
                        "name": "maxRate",
                        "type": "f32"
                    },
                    {
                        "name": "adjustmentFactor",
                        "type": "f32"
                    }
                ]
            }
        },
        {
            "name": "Equity",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "tokens",
                        "type": {
                            "vec": {
                                "defined": "TokenEquity"
                            }
                        }
                    },
                    {
                        "name": "perps",
                        "type": {
                            "vec": {
                                "defined": "PerpEquity"
                            }
                        }
                    }
                ]
            }
        },
        {
            "name": "TokenEquity",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "tokenIndex",
                        "type": "u16"
                    },
                    {
                        "name": "value",
                        "type": {
                            "defined": "I80F48"
                        }
                    }
                ]
            }
        },
        {
            "name": "PerpEquity",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "perpMarketIndex",
                        "type": "u16"
                    },
                    {
                        "name": "value",
                        "type": {
                            "defined": "I80F48"
                        }
                    }
                ]
            }
        },
        {
            "name": "Prices",
            "docs": [
                "Information about prices for a bank or perp market."
            ],
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "oracle",
                        "docs": [
                            "The current oracle price"
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "stable",
                        "docs": [
                            "A \"stable\" price, provided by StablePriceModel"
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    }
                ]
            }
        },
        {
            "name": "TokenInfo",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "tokenIndex",
                        "type": "u16"
                    },
                    {
                        "name": "maintAssetWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "initAssetWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "initScaledAssetWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "maintLiabWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "initLiabWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "initScaledLiabWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "prices",
                        "type": {
                            "defined": "Prices"
                        }
                    },
                    {
                        "name": "balanceNative",
                        "type": {
                            "defined": "I80F48"
                        }
                    }
                ]
            }
        },
        {
            "name": "Serum3Info",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "reservedBase",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "reservedQuote",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "baseIndex",
                        "type": "u64"
                    },
                    {
                        "name": "quoteIndex",
                        "type": "u64"
                    },
                    {
                        "name": "marketIndex",
                        "type": "u16"
                    },
                    {
                        "name": "hasZeroFunds",
                        "docs": [
                            "The open orders account has no free or reserved funds"
                        ],
                        "type": "bool"
                    }
                ]
            }
        },
        {
            "name": "PerpInfo",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "perpMarketIndex",
                        "type": "u16"
                    },
                    {
                        "name": "maintBaseAssetWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "initBaseAssetWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "maintBaseLiabWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "initBaseLiabWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "maintOverallAssetWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "initOverallAssetWeight",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "baseLotSize",
                        "type": "i64"
                    },
                    {
                        "name": "baseLots",
                        "type": "i64"
                    },
                    {
                        "name": "bidsBaseLots",
                        "type": "i64"
                    },
                    {
                        "name": "asksBaseLots",
                        "type": "i64"
                    },
                    {
                        "name": "quote",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "prices",
                        "type": {
                            "defined": "Prices"
                        }
                    },
                    {
                        "name": "hasOpenOrders",
                        "type": "bool"
                    },
                    {
                        "name": "hasOpenFills",
                        "type": "bool"
                    }
                ]
            }
        },
        {
            "name": "HealthCache",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "tokenInfos",
                        "type": {
                            "vec": {
                                "defined": "TokenInfo"
                            }
                        }
                    },
                    {
                        "name": "serum3Infos",
                        "type": {
                            "vec": {
                                "defined": "Serum3Info"
                            }
                        }
                    },
                    {
                        "name": "perpInfos",
                        "type": {
                            "vec": {
                                "defined": "PerpInfo"
                            }
                        }
                    },
                    {
                        "name": "beingLiquidated",
                        "type": "bool"
                    }
                ]
            }
        },
        {
            "name": "FlashLoanTokenDetail",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "tokenIndex",
                        "type": "u16"
                    },
                    {
                        "name": "changeAmount",
                        "type": "i128"
                    },
                    {
                        "name": "loan",
                        "type": "i128"
                    },
                    {
                        "name": "loanOriginationFee",
                        "type": "i128"
                    },
                    {
                        "name": "depositIndex",
                        "type": "i128"
                    },
                    {
                        "name": "borrowIndex",
                        "type": "i128"
                    },
                    {
                        "name": "price",
                        "type": "i128"
                    }
                ]
            }
        },
        {
            "name": "TokenPosition",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "indexedPosition",
                        "docs": [
                            "The deposit_index (if positive) or borrow_index (if negative) scaled position"
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "tokenIndex",
                        "docs": [
                            "index into Group.tokens"
                        ],
                        "type": "u16"
                    },
                    {
                        "name": "inUseCount",
                        "docs": [
                            "incremented when a market requires this position to stay alive"
                        ],
                        "type": "u8"
                    },
                    {
                        "name": "padding",
                        "type": {
                            "array": [
                                "u8",
                                5
                            ]
                        }
                    },
                    {
                        "name": "previousIndex",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "cumulativeDepositInterest",
                        "type": "f64"
                    },
                    {
                        "name": "cumulativeBorrowInterest",
                        "type": "f64"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                128
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "Serum3Orders",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "openOrders",
                        "type": "publicKey"
                    },
                    {
                        "name": "baseBorrowsWithoutFee",
                        "docs": [
                            "Tracks the amount of borrows that have flowed into the serum open orders account.",
                            "These borrows did not have the loan origination fee applied, and that may happen",
                            "later (in serum3_settle_funds) if we can guarantee that the funds were used.",
                            "In particular a place-on-book, cancel, settle should not cost fees."
                        ],
                        "type": "u64"
                    },
                    {
                        "name": "quoteBorrowsWithoutFee",
                        "type": "u64"
                    },
                    {
                        "name": "marketIndex",
                        "type": "u16"
                    },
                    {
                        "name": "baseTokenIndex",
                        "docs": [
                            "Store the base/quote token index, so health computations don't need",
                            "to get passed the static SerumMarket to find which tokens a market",
                            "uses and look up the correct oracles."
                        ],
                        "type": "u16"
                    },
                    {
                        "name": "quoteTokenIndex",
                        "type": "u16"
                    },
                    {
                        "name": "padding",
                        "type": {
                            "array": [
                                "u8",
                                2
                            ]
                        }
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                64
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "PerpPosition",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "marketIndex",
                        "type": "u16"
                    },
                    {
                        "name": "padding",
                        "type": {
                            "array": [
                                "u8",
                                2
                            ]
                        }
                    },
                    {
                        "name": "settlePnlLimitWindow",
                        "docs": [
                            "Index of the current settle pnl limit window"
                        ],
                        "type": "u32"
                    },
                    {
                        "name": "settlePnlLimitSettledInCurrentWindowNative",
                        "docs": [
                            "Amount of realized trade pnl and unrealized pnl that was already settled this window.",
                            "",
                            "Will be negative when negative pnl was settled.",
                            "",
                            "Note that this will be adjusted for bookkeeping reasons when the realized_trade settle",
                            "limitchanges and is not useable for actually tracking how much pnl was settled",
                            "on balance."
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "basePositionLots",
                        "docs": [
                            "Active position size, measured in base lots"
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "quotePositionNative",
                        "docs": [
                            "Active position in quote (conversation rate is that of the time the order was settled)",
                            "measured in native quote"
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "quoteRunningNative",
                        "docs": [
                            "Tracks what the position is to calculate average entry & break even price"
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "longSettledFunding",
                        "docs": [
                            "Already settled long funding"
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "shortSettledFunding",
                        "docs": [
                            "Already settled short funding"
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "bidsBaseLots",
                        "docs": [
                            "Base lots in open bids"
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "asksBaseLots",
                        "docs": [
                            "Base lots in open asks"
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "takerBaseLots",
                        "docs": [
                            "Amount of base lots on the EventQueue waiting to be processed"
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "takerQuoteLots",
                        "docs": [
                            "Amount of quote lots on the EventQueue waiting to be processed"
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "cumulativeLongFunding",
                        "type": "f64"
                    },
                    {
                        "name": "cumulativeShortFunding",
                        "type": "f64"
                    },
                    {
                        "name": "makerVolume",
                        "type": "u64"
                    },
                    {
                        "name": "takerVolume",
                        "type": "u64"
                    },
                    {
                        "name": "perpSpotTransfers",
                        "type": "i64"
                    },
                    {
                        "name": "avgEntryPricePerBaseLot",
                        "docs": [
                            "The native average entry price for the base lots of the current position.",
                            "Reset to 0 when the base position reaches or crosses 0."
                        ],
                        "type": "f64"
                    },
                    {
                        "name": "realizedTradePnlNative",
                        "docs": [
                            "Amount of pnl that was realized by bringing the base position closer to 0.",
                            "",
                            "The settlement of this type of pnl is limited by settle_pnl_limit_realized_trade.",
                            "Settling pnl reduces this value once other_pnl below is exhausted."
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "realizedOtherPnlNative",
                        "docs": [
                            "Amount of pnl realized from fees, funding and liquidation.",
                            "",
                            "This type of realized pnl is always settleable.",
                            "Settling pnl reduces this value first."
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "settlePnlLimitRealizedTrade",
                        "docs": [
                            "Settle limit contribution from realized pnl.",
                            "",
                            "Every time pnl is realized, this is increased by a fraction of the stable",
                            "value of the realization. It magnitude decreases when realized pnl drops below its value."
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "realizedPnlForPositionNative",
                        "docs": [
                            "Trade pnl, fees, funding that were added over the current position's lifetime.",
                            "",
                            "Reset when the position changes sign or goes to zero.",
                            "Not decreased by settling.",
                            "",
                            "This is tracked for display purposes: this value plus the difference between entry",
                            "price and current price of the base position is the overall pnl."
                        ],
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                88
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "PerpOpenOrder",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "sideAndTree",
                        "type": "u8"
                    },
                    {
                        "name": "padding1",
                        "type": {
                            "array": [
                                "u8",
                                1
                            ]
                        }
                    },
                    {
                        "name": "market",
                        "type": "u16"
                    },
                    {
                        "name": "padding2",
                        "type": {
                            "array": [
                                "u8",
                                4
                            ]
                        }
                    },
                    {
                        "name": "clientId",
                        "type": "u64"
                    },
                    {
                        "name": "id",
                        "type": "u128"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                64
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "MangoAccountFixed",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "group",
                        "type": "publicKey"
                    },
                    {
                        "name": "owner",
                        "type": "publicKey"
                    },
                    {
                        "name": "name",
                        "type": {
                            "array": [
                                "u8",
                                32
                            ]
                        }
                    },
                    {
                        "name": "delegate",
                        "type": "publicKey"
                    },
                    {
                        "name": "accountNum",
                        "type": "u32"
                    },
                    {
                        "name": "beingLiquidated",
                        "type": "u8"
                    },
                    {
                        "name": "inHealthRegion",
                        "type": "u8"
                    },
                    {
                        "name": "bump",
                        "type": "u8"
                    },
                    {
                        "name": "padding",
                        "type": {
                            "array": [
                                "u8",
                                1
                            ]
                        }
                    },
                    {
                        "name": "netDeposits",
                        "type": "i64"
                    },
                    {
                        "name": "perpSpotTransfers",
                        "type": "i64"
                    },
                    {
                        "name": "healthRegionBeginInitHealth",
                        "type": "i64"
                    },
                    {
                        "name": "frozenUntil",
                        "type": "u64"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                232
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "OracleConfig",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "confFilter",
                        "type": {
                            "defined": "I80F48"
                        }
                    },
                    {
                        "name": "maxStalenessSlots",
                        "type": "i64"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                72
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "OracleConfigParams",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "confFilter",
                        "type": "f32"
                    },
                    {
                        "name": "maxStalenessSlots",
                        "type": {
                            "option": "u32"
                        }
                    }
                ]
            }
        },
        {
            "name": "InnerNode",
            "docs": [
                "InnerNodes and LeafNodes compose the binary tree of orders.",
                "",
                "Each InnerNode has exactly two children, which are either InnerNodes themselves,",
                "or LeafNodes. The children share the top `prefix_len` bits of `key`. The left",
                "child has a 0 in the next bit, and the right a 1."
            ],
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "tag",
                        "type": "u8"
                    },
                    {
                        "name": "padding",
                        "type": {
                            "array": [
                                "u8",
                                3
                            ]
                        }
                    },
                    {
                        "name": "prefixLen",
                        "docs": [
                            "number of highest `key` bits that all children share",
                            "e.g. if it's 2, the two highest bits of `key` will be the same on all children"
                        ],
                        "type": "u32"
                    },
                    {
                        "name": "key",
                        "docs": [
                            "only the top `prefix_len` bits of `key` are relevant"
                        ],
                        "type": "u128"
                    },
                    {
                        "name": "children",
                        "docs": [
                            "indexes into `BookSide::nodes`"
                        ],
                        "type": {
                            "array": [
                                "u32",
                                2
                            ]
                        }
                    },
                    {
                        "name": "childEarliestExpiry",
                        "docs": [
                            "The earliest expiry timestamp for the left and right subtrees.",
                            "",
                            "Needed to be able to find and remove expired orders without having to",
                            "iterate through the whole bookside."
                        ],
                        "type": {
                            "array": [
                                "u64",
                                2
                            ]
                        }
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                72
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "LeafNode",
            "docs": [
                "LeafNodes represent an order in the binary tree"
            ],
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "tag",
                        "docs": [
                            "NodeTag"
                        ],
                        "type": "u8"
                    },
                    {
                        "name": "ownerSlot",
                        "docs": [
                            "Index into the owning MangoAccount's PerpOpenOrders"
                        ],
                        "type": "u8"
                    },
                    {
                        "name": "orderType",
                        "docs": [
                            "PostOrderType, this was added for TradingView move order"
                        ],
                        "type": "u8"
                    },
                    {
                        "name": "padding",
                        "type": {
                            "array": [
                                "u8",
                                1
                            ]
                        }
                    },
                    {
                        "name": "timeInForce",
                        "docs": [
                            "Time in seconds after `timestamp` at which the order expires.",
                            "A value of 0 means no expiry."
                        ],
                        "type": "u16"
                    },
                    {
                        "name": "padding2",
                        "type": {
                            "array": [
                                "u8",
                                2
                            ]
                        }
                    },
                    {
                        "name": "key",
                        "docs": [
                            "The binary tree key, see new_node_key()"
                        ],
                        "type": "u128"
                    },
                    {
                        "name": "owner",
                        "docs": [
                            "Address of the owning MangoAccount"
                        ],
                        "type": "publicKey"
                    },
                    {
                        "name": "quantity",
                        "docs": [
                            "Number of base lots to buy or sell, always >=1"
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "timestamp",
                        "docs": [
                            "The time the order was placed"
                        ],
                        "type": "u64"
                    },
                    {
                        "name": "pegLimit",
                        "docs": [
                            "If the effective price of an oracle pegged order exceeds this limit,",
                            "it will be considered invalid and may be removed.",
                            "",
                            "Only applicable in the oracle_pegged OrderTree"
                        ],
                        "type": "i64"
                    },
                    {
                        "name": "clientOrderId",
                        "docs": [
                            "User defined id for this order, used in FillEvents"
                        ],
                        "type": "u64"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                32
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "AnyNode",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "tag",
                        "type": "u8"
                    },
                    {
                        "name": "data",
                        "type": {
                            "array": [
                                "u8",
                                119
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "OrderTreeRoot",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "maybeNode",
                        "type": "u32"
                    },
                    {
                        "name": "leafCount",
                        "type": "u32"
                    }
                ]
            }
        },
        {
            "name": "OrderTreeNodes",
            "docs": [
                "A binary tree on AnyNode::key()",
                "",
                "The key encodes the price in the top 64 bits."
            ],
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "orderTreeType",
                        "type": "u8"
                    },
                    {
                        "name": "padding",
                        "type": {
                            "array": [
                                "u8",
                                3
                            ]
                        }
                    },
                    {
                        "name": "bumpIndex",
                        "type": "u32"
                    },
                    {
                        "name": "freeListLen",
                        "type": "u32"
                    },
                    {
                        "name": "freeListHead",
                        "type": "u32"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                512
                            ]
                        }
                    },
                    {
                        "name": "nodes",
                        "type": {
                            "array": [
                                {
                                    "defined": "AnyNode"
                                },
                                1024
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "EventQueueHeader",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "head",
                        "type": "u32"
                    },
                    {
                        "name": "count",
                        "type": "u32"
                    },
                    {
                        "name": "seqNum",
                        "type": "u64"
                    }
                ]
            }
        },
        {
            "name": "AnyEvent",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "eventType",
                        "type": "u8"
                    },
                    {
                        "name": "padding",
                        "type": {
                            "array": [
                                "u8",
                                207
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "FillEvent",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "eventType",
                        "type": "u8"
                    },
                    {
                        "name": "takerSide",
                        "type": "u8"
                    },
                    {
                        "name": "makerOut",
                        "type": "u8"
                    },
                    {
                        "name": "makerSlot",
                        "type": "u8"
                    },
                    {
                        "name": "padding",
                        "type": {
                            "array": [
                                "u8",
                                4
                            ]
                        }
                    },
                    {
                        "name": "timestamp",
                        "type": "u64"
                    },
                    {
                        "name": "seqNum",
                        "type": "u64"
                    },
                    {
                        "name": "maker",
                        "type": "publicKey"
                    },
                    {
                        "name": "padding2",
                        "type": {
                            "array": [
                                "u8",
                                32
                            ]
                        }
                    },
                    {
                        "name": "makerTimestamp",
                        "type": "u64"
                    },
                    {
                        "name": "taker",
                        "type": "publicKey"
                    },
                    {
                        "name": "padding3",
                        "type": {
                            "array": [
                                "u8",
                                16
                            ]
                        }
                    },
                    {
                        "name": "takerClientOrderId",
                        "type": "u64"
                    },
                    {
                        "name": "padding4",
                        "type": {
                            "array": [
                                "u8",
                                16
                            ]
                        }
                    },
                    {
                        "name": "price",
                        "type": "i64"
                    },
                    {
                        "name": "quantity",
                        "type": "i64"
                    },
                    {
                        "name": "makerClientOrderId",
                        "type": "u64"
                    },
                    {
                        "name": "makerFee",
                        "type": "f32"
                    },
                    {
                        "name": "takerFee",
                        "type": "f32"
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                8
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "OutEvent",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "eventType",
                        "type": "u8"
                    },
                    {
                        "name": "side",
                        "type": "u8"
                    },
                    {
                        "name": "ownerSlot",
                        "type": "u8"
                    },
                    {
                        "name": "padding0",
                        "type": {
                            "array": [
                                "u8",
                                5
                            ]
                        }
                    },
                    {
                        "name": "timestamp",
                        "type": "u64"
                    },
                    {
                        "name": "seqNum",
                        "type": "u64"
                    },
                    {
                        "name": "owner",
                        "type": "publicKey"
                    },
                    {
                        "name": "quantity",
                        "type": "i64"
                    },
                    {
                        "name": "padding1",
                        "type": {
                            "array": [
                                "u8",
                                144
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "StablePriceModel",
            "docs": [
                "Maintains a \"stable_price\" based on the oracle price.",
                "",
                "The stable price follows the oracle price, but its relative rate of",
                "change is limited (to `stable_growth_limit`) and futher reduced if",
                "the oracle price is far from the `delay_price`.",
                "",
                "Conceptually the `delay_price` is itself a time delayed",
                "(`24 * delay_interval_seconds`, assume 24h) and relative rate of change limited",
                "function of the oracle price. It is implemented as averaging the oracle",
                "price over every `delay_interval_seconds` (assume 1h) and then applying the",
                "`delay_growth_limit` between intervals."
            ],
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "stablePrice",
                        "docs": [
                            "Current stable price to use in health"
                        ],
                        "type": "f64"
                    },
                    {
                        "name": "lastUpdateTimestamp",
                        "type": "u64"
                    },
                    {
                        "name": "delayPrices",
                        "docs": [
                            "Stored delay_price for each delay_interval.",
                            "If we want the delay_price to be 24h delayed, we would store one for each hour.",
                            "This is used in a cyclical way: We use the maximally-delayed value at delay_interval_index",
                            "and once enough time passes to move to the next delay interval, that gets overwritten and",
                            "we use the next one."
                        ],
                        "type": {
                            "array": [
                                "f64",
                                24
                            ]
                        }
                    },
                    {
                        "name": "delayAccumulatorPrice",
                        "docs": [
                            "The delay price is based on an average over each delay_interval. The contributions",
                            "to the average are summed up here."
                        ],
                        "type": "f64"
                    },
                    {
                        "name": "delayAccumulatorTime",
                        "docs": [
                            "Accumulating the total time for the above average."
                        ],
                        "type": "u32"
                    },
                    {
                        "name": "delayIntervalSeconds",
                        "docs": [
                            "Length of a delay_interval"
                        ],
                        "type": "u32"
                    },
                    {
                        "name": "delayGrowthLimit",
                        "docs": [
                            "Maximal relative difference between two delay_price in consecutive intervals."
                        ],
                        "type": "f32"
                    },
                    {
                        "name": "stableGrowthLimit",
                        "docs": [
                            "Maximal per-second relative difference of the stable price.",
                            "It gets further reduced if stable and delay price disagree."
                        ],
                        "type": "f32"
                    },
                    {
                        "name": "lastDelayIntervalIndex",
                        "docs": [
                            "The delay_interval_index that update() was last called on."
                        ],
                        "type": "u8"
                    },
                    {
                        "name": "padding",
                        "type": {
                            "array": [
                                "u8",
                                7
                            ]
                        }
                    },
                    {
                        "name": "reserved",
                        "type": {
                            "array": [
                                "u8",
                                48
                            ]
                        }
                    }
                ]
            }
        },
        {
            "name": "TokenIndex",
            "docs": [
                "Nothing in Rust shall use these types. They only exist so that the Anchor IDL",
                "knows about them and typescript can deserialize it."
            ],
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "val",
                        "type": "u16"
                    }
                ]
            }
        },
        {
            "name": "Serum3MarketIndex",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "val",
                        "type": "u16"
                    }
                ]
            }
        },
        {
            "name": "PerpMarketIndex",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "val",
                        "type": "u16"
                    }
                ]
            }
        },
        {
            "name": "I80F48",
            "type": {
                "kind": "struct",
                "fields": [
                    {
                        "name": "val",
                        "type": "i128"
                    }
                ]
            }
        },
        {
            "name": "FlashLoanType",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Unknown"
                    },
                    {
                        "name": "Swap"
                    }
                ]
            }
        },
        {
            "name": "Serum3SelfTradeBehavior",
            "docs": [
                "Copy paste a bunch of enums so that we could AnchorSerialize & AnchorDeserialize them"
            ],
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "DecrementTake"
                    },
                    {
                        "name": "CancelProvide"
                    },
                    {
                        "name": "AbortTransaction"
                    }
                ]
            }
        },
        {
            "name": "Serum3OrderType",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Limit"
                    },
                    {
                        "name": "ImmediateOrCancel"
                    },
                    {
                        "name": "PostOnly"
                    }
                ]
            }
        },
        {
            "name": "Serum3Side",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Bid"
                    },
                    {
                        "name": "Ask"
                    }
                ]
            }
        },
        {
            "name": "HealthType",
            "docs": [
                "There are three types of health:",
                "- initial health (\"init\"): users can only open new positions if it's >= 0",
                "- maintenance health (\"maint\"): users get liquidated if it's < 0",
                "- liquidation end health: once liquidation started (see being_liquidated), it",
                "only stops once this is >= 0",
                "",
                "The ordering is",
                "init health <= liquidation end health <= maint health",
                "",
                "The different health types are realized by using different weights and prices:",
                "- init health: init weights with scaling, stable-price adjusted prices",
                "- liq end health: init weights without scaling, oracle prices",
                "- maint health: maint weights, oracle prices",
                ""
            ],
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Init"
                    },
                    {
                        "name": "Maint"
                    },
                    {
                        "name": "LiquidationEnd"
                    }
                ]
            }
        },
        {
            "name": "LoanOriginationFeeInstruction",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Unknown"
                    },
                    {
                        "name": "LiqTokenBankruptcy"
                    },
                    {
                        "name": "LiqTokenWithToken"
                    },
                    {
                        "name": "Serum3LiqForceCancelOrders"
                    },
                    {
                        "name": "Serum3PlaceOrder"
                    },
                    {
                        "name": "Serum3SettleFunds"
                    },
                    {
                        "name": "TokenWithdraw"
                    }
                ]
            }
        },
        {
            "name": "IxGate",
            "docs": [
                "Enum for lookup into ix gate",
                "note:",
                "total ix files 56,",
                "ix files included 48,",
                "ix files not included 8,",
                "- Benchmark,",
                "- ComputeAccountData,",
                "- GroupCreate",
                "- GroupEdit",
                "- IxGateSet,",
                "- PerpZeroOut,",
                "- PerpEditMarket,",
                "- TokenEdit,"
            ],
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "AccountClose"
                    },
                    {
                        "name": "AccountCreate"
                    },
                    {
                        "name": "AccountEdit"
                    },
                    {
                        "name": "AccountExpand"
                    },
                    {
                        "name": "AccountToggleFreeze"
                    },
                    {
                        "name": "AltExtend"
                    },
                    {
                        "name": "AltSet"
                    },
                    {
                        "name": "FlashLoan"
                    },
                    {
                        "name": "GroupClose"
                    },
                    {
                        "name": "GroupCreate"
                    },
                    {
                        "name": "HealthRegion"
                    },
                    {
                        "name": "PerpCancelAllOrders"
                    },
                    {
                        "name": "PerpCancelAllOrdersBySide"
                    },
                    {
                        "name": "PerpCancelOrder"
                    },
                    {
                        "name": "PerpCancelOrderByClientOrderId"
                    },
                    {
                        "name": "PerpCloseMarket"
                    },
                    {
                        "name": "PerpConsumeEvents"
                    },
                    {
                        "name": "PerpCreateMarket"
                    },
                    {
                        "name": "PerpDeactivatePosition"
                    },
                    {
                        "name": "PerpLiqBaseOrPositivePnl"
                    },
                    {
                        "name": "PerpLiqForceCancelOrders"
                    },
                    {
                        "name": "PerpLiqNegativePnlOrBankruptcy"
                    },
                    {
                        "name": "PerpPlaceOrder"
                    },
                    {
                        "name": "PerpSettleFees"
                    },
                    {
                        "name": "PerpSettlePnl"
                    },
                    {
                        "name": "PerpUpdateFunding"
                    },
                    {
                        "name": "Serum3CancelAllOrders"
                    },
                    {
                        "name": "Serum3CancelOrder"
                    },
                    {
                        "name": "Serum3CloseOpenOrders"
                    },
                    {
                        "name": "Serum3CreateOpenOrders"
                    },
                    {
                        "name": "Serum3DeregisterMarket"
                    },
                    {
                        "name": "Serum3EditMarket"
                    },
                    {
                        "name": "Serum3LiqForceCancelOrders"
                    },
                    {
                        "name": "Serum3PlaceOrder"
                    },
                    {
                        "name": "Serum3RegisterMarket"
                    },
                    {
                        "name": "Serum3SettleFunds"
                    },
                    {
                        "name": "StubOracleClose"
                    },
                    {
                        "name": "StubOracleCreate"
                    },
                    {
                        "name": "StubOracleSet"
                    },
                    {
                        "name": "TokenAddBank"
                    },
                    {
                        "name": "TokenDeposit"
                    },
                    {
                        "name": "TokenDeregister"
                    },
                    {
                        "name": "TokenLiqBankruptcy"
                    },
                    {
                        "name": "TokenLiqWithToken"
                    },
                    {
                        "name": "TokenRegister"
                    },
                    {
                        "name": "TokenRegisterTrustless"
                    },
                    {
                        "name": "TokenUpdateIndexAndRate"
                    },
                    {
                        "name": "TokenWithdraw"
                    }
                ]
            }
        },
        {
            "name": "OracleType",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Pyth"
                    },
                    {
                        "name": "Stub"
                    },
                    {
                        "name": "SwitchboardV1"
                    },
                    {
                        "name": "SwitchboardV2"
                    }
                ]
            }
        },
        {
            "name": "OrderState",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Valid"
                    },
                    {
                        "name": "Invalid"
                    },
                    {
                        "name": "Skipped"
                    }
                ]
            }
        },
        {
            "name": "BookSideOrderTree",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Fixed"
                    },
                    {
                        "name": "OraclePegged"
                    }
                ]
            }
        },
        {
            "name": "NodeTag",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Uninitialized"
                    },
                    {
                        "name": "InnerNode"
                    },
                    {
                        "name": "LeafNode"
                    },
                    {
                        "name": "FreeNode"
                    },
                    {
                        "name": "LastFreeNode"
                    }
                ]
            }
        },
        {
            "name": "PlaceOrderType",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Limit"
                    },
                    {
                        "name": "ImmediateOrCancel"
                    },
                    {
                        "name": "PostOnly"
                    },
                    {
                        "name": "Market"
                    },
                    {
                        "name": "PostOnlySlide"
                    }
                ]
            }
        },
        {
            "name": "PostOrderType",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Limit"
                    },
                    {
                        "name": "PostOnly"
                    },
                    {
                        "name": "PostOnlySlide"
                    }
                ]
            }
        },
        {
            "name": "Side",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Bid"
                    },
                    {
                        "name": "Ask"
                    }
                ]
            }
        },
        {
            "name": "SideAndOrderTree",
            "docs": [
                "SideAndOrderTree is a storage optimization, so we don't need two bytes for the data"
            ],
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "BidFixed"
                    },
                    {
                        "name": "AskFixed"
                    },
                    {
                        "name": "BidOraclePegged"
                    },
                    {
                        "name": "AskOraclePegged"
                    }
                ]
            }
        },
        {
            "name": "OrderParams",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Market"
                    },
                    {
                        "name": "ImmediateOrCancel",
                        "fields": [
                            {
                                "name": "price_lots",
                                "type": "i64"
                            }
                        ]
                    },
                    {
                        "name": "Fixed",
                        "fields": [
                            {
                                "name": "price_lots",
                                "type": "i64"
                            },
                            {
                                "name": "order_type",
                                "type": {
                                    "defined": "PostOrderType"
                                }
                            }
                        ]
                    },
                    {
                        "name": "OraclePegged",
                        "fields": [
                            {
                                "name": "price_offset_lots",
                                "type": "i64"
                            },
                            {
                                "name": "order_type",
                                "type": {
                                    "defined": "PostOrderType"
                                }
                            },
                            {
                                "name": "peg_limit",
                                "type": "i64"
                            },
                            {
                                "name": "max_oracle_staleness_slots",
                                "type": "i32"
                            }
                        ]
                    }
                ]
            }
        },
        {
            "name": "OrderTreeType",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Bids"
                    },
                    {
                        "name": "Asks"
                    }
                ]
            }
        },
        {
            "name": "EventType",
            "type": {
                "kind": "enum",
                "variants": [
                    {
                        "name": "Fill"
                    },
                    {
                        "name": "Out"
                    },
                    {
                        "name": "Liquidate"
                    }
                ]
            }
        }
    ],
    "events": [
        {
            "name": "MangoAccountData",
            "fields": [
                {
                    "name": "healthCache",
                    "type": {
                        "defined": "HealthCache"
                    },
                    "index": false
                },
                {
                    "name": "initHealth",
                    "type": {
                        "defined": "I80F48"
                    },
                    "index": false
                },
                {
                    "name": "maintHealth",
                    "type": {
                        "defined": "I80F48"
                    },
                    "index": false
                },
                {
                    "name": "equity",
                    "type": {
                        "defined": "Equity"
                    },
                    "index": false
                }
            ]
        },
        {
            "name": "PerpBalanceLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccount",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "marketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "basePosition",
                    "type": "i64",
                    "index": false
                },
                {
                    "name": "quotePosition",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "longSettledFunding",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "shortSettledFunding",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "longFunding",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "shortFunding",
                    "type": "i128",
                    "index": false
                }
            ]
        },
        {
            "name": "TokenBalanceLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccount",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "tokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "indexedPosition",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "depositIndex",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "borrowIndex",
                    "type": "i128",
                    "index": false
                }
            ]
        },
        {
            "name": "FlashLoanLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccount",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "tokenLoanDetails",
                    "type": {
                        "vec": {
                            "defined": "FlashLoanTokenDetail"
                        }
                    },
                    "index": false
                },
                {
                    "name": "flashLoanType",
                    "type": {
                        "defined": "FlashLoanType"
                    },
                    "index": false
                }
            ]
        },
        {
            "name": "WithdrawLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccount",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "signer",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "tokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "quantity",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "price",
                    "type": "i128",
                    "index": false
                }
            ]
        },
        {
            "name": "DepositLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccount",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "signer",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "tokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "quantity",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "price",
                    "type": "i128",
                    "index": false
                }
            ]
        },
        {
            "name": "FillLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "marketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "takerSide",
                    "type": "u8",
                    "index": false
                },
                {
                    "name": "makerSlot",
                    "type": "u8",
                    "index": false
                },
                {
                    "name": "makerOut",
                    "type": "bool",
                    "index": false
                },
                {
                    "name": "timestamp",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "seqNum",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "maker",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "makerOrderId",
                    "type": "u128",
                    "index": false
                },
                {
                    "name": "makerFee",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "makerTimestamp",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "taker",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "takerOrderId",
                    "type": "u128",
                    "index": false
                },
                {
                    "name": "takerClientOrderId",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "takerFee",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "price",
                    "type": "i64",
                    "index": false
                },
                {
                    "name": "quantity",
                    "type": "i64",
                    "index": false
                }
            ]
        },
        {
            "name": "FillLogV2",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "marketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "takerSide",
                    "type": "u8",
                    "index": false
                },
                {
                    "name": "makerSlot",
                    "type": "u8",
                    "index": false
                },
                {
                    "name": "makerOut",
                    "type": "bool",
                    "index": false
                },
                {
                    "name": "timestamp",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "seqNum",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "maker",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "makerClientOrderId",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "makerFee",
                    "type": "f32",
                    "index": false
                },
                {
                    "name": "makerTimestamp",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "taker",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "takerClientOrderId",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "takerFee",
                    "type": "f32",
                    "index": false
                },
                {
                    "name": "price",
                    "type": "i64",
                    "index": false
                },
                {
                    "name": "quantity",
                    "type": "i64",
                    "index": false
                }
            ]
        },
        {
            "name": "PerpUpdateFundingLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "marketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "longFunding",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "shortFunding",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "price",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "stablePrice",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "feesAccrued",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "openInterest",
                    "type": "i64",
                    "index": false
                },
                {
                    "name": "instantaneousFundingRate",
                    "type": "i128",
                    "index": false
                }
            ]
        },
        {
            "name": "UpdateIndexLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "tokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "depositIndex",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "borrowIndex",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "avgUtilization",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "price",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "stablePrice",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "collectedFees",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "loanFeeRate",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "totalBorrows",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "totalDeposits",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "borrowRate",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "depositRate",
                    "type": "i128",
                    "index": false
                }
            ]
        },
        {
            "name": "UpdateRateLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "tokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "rate0",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "rate1",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "maxRate",
                    "type": "i128",
                    "index": false
                }
            ]
        },
        {
            "name": "TokenLiqWithTokenLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "liqee",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "liqor",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "assetTokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "liabTokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "assetTransfer",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "liabTransfer",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "assetPrice",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "liabPrice",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "bankruptcy",
                    "type": "bool",
                    "index": false
                }
            ]
        },
        {
            "name": "Serum3OpenOrdersBalanceLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccount",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "baseTokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "quoteTokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "baseTotal",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "baseFree",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "quoteTotal",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "quoteFree",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "referrerRebatesAccrued",
                    "type": "u64",
                    "index": false
                }
            ]
        },
        {
            "name": "Serum3OpenOrdersBalanceLogV2",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccount",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "marketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "baseTokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "quoteTokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "baseTotal",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "baseFree",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "quoteTotal",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "quoteFree",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "referrerRebatesAccrued",
                    "type": "u64",
                    "index": false
                }
            ]
        },
        {
            "name": "WithdrawLoanOriginationFeeLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccount",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "tokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "loanOriginationFee",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "instruction",
                    "type": {
                        "defined": "LoanOriginationFeeInstruction"
                    },
                    "index": false
                }
            ]
        },
        {
            "name": "TokenLiqBankruptcyLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "liqee",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "liqor",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "liabTokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "initialLiabNative",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "liabPrice",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "insuranceTokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "insuranceTransfer",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "socializedLoss",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "startingLiabDepositIndex",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "endingLiabDepositIndex",
                    "type": "i128",
                    "index": false
                }
            ]
        },
        {
            "name": "DeactivateTokenPositionLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccount",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "tokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "cumulativeDepositInterest",
                    "type": "f64",
                    "index": false
                },
                {
                    "name": "cumulativeBorrowInterest",
                    "type": "f64",
                    "index": false
                }
            ]
        },
        {
            "name": "DeactivatePerpPositionLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccount",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "marketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "cumulativeLongFunding",
                    "type": "f64",
                    "index": false
                },
                {
                    "name": "cumulativeShortFunding",
                    "type": "f64",
                    "index": false
                },
                {
                    "name": "makerVolume",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "takerVolume",
                    "type": "u64",
                    "index": false
                },
                {
                    "name": "perpSpotTransfers",
                    "type": "i64",
                    "index": false
                }
            ]
        },
        {
            "name": "TokenMetaDataLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mint",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "tokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "mintDecimals",
                    "type": "u8",
                    "index": false
                },
                {
                    "name": "oracle",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mintInfo",
                    "type": "publicKey",
                    "index": false
                }
            ]
        },
        {
            "name": "PerpMarketMetaDataLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "perpMarket",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "perpMarketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "baseDecimals",
                    "type": "u8",
                    "index": false
                },
                {
                    "name": "baseLotSize",
                    "type": "i64",
                    "index": false
                },
                {
                    "name": "quoteLotSize",
                    "type": "i64",
                    "index": false
                },
                {
                    "name": "oracle",
                    "type": "publicKey",
                    "index": false
                }
            ]
        },
        {
            "name": "Serum3RegisterMarketLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "serumMarket",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "marketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "baseTokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "quoteTokenIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "serumProgram",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "serumProgramExternal",
                    "type": "publicKey",
                    "index": false
                }
            ]
        },
        {
            "name": "PerpLiqBaseOrPositivePnlLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "perpMarketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "liqor",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "liqee",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "baseTransfer",
                    "type": "i64",
                    "index": false
                },
                {
                    "name": "quoteTransfer",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "pnlTransfer",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "pnlSettleLimitTransfer",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "price",
                    "type": "i128",
                    "index": false
                }
            ]
        },
        {
            "name": "PerpLiqBankruptcyLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "liqee",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "liqor",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "perpMarketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "insuranceTransfer",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "socializedLoss",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "startingLongFunding",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "startingShortFunding",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "endingLongFunding",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "endingShortFunding",
                    "type": "i128",
                    "index": false
                }
            ]
        },
        {
            "name": "PerpLiqNegativePnlOrBankruptcyLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "liqee",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "liqor",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "perpMarketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "settlement",
                    "type": "i128",
                    "index": false
                }
            ]
        },
        {
            "name": "PerpSettlePnlLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccountA",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccountB",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "perpMarketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "settlement",
                    "type": "i128",
                    "index": false
                },
                {
                    "name": "settler",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "fee",
                    "type": "i128",
                    "index": false
                }
            ]
        },
        {
            "name": "PerpSettleFeesLog",
            "fields": [
                {
                    "name": "mangoGroup",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "mangoAccount",
                    "type": "publicKey",
                    "index": false
                },
                {
                    "name": "perpMarketIndex",
                    "type": "u16",
                    "index": false
                },
                {
                    "name": "settlement",
                    "type": "i128",
                    "index": false
                }
            ]
        }
    ],
    "errors": [
        {
            "code": 6000,
            "name": "SomeError",
            "msg": ""
        },
        {
            "code": 6001,
            "name": "NotImplementedError",
            "msg": ""
        },
        {
            "code": 6002,
            "name": "MathError",
            "msg": "checked math error"
        },
        {
            "code": 6003,
            "name": "UnexpectedOracle",
            "msg": ""
        },
        {
            "code": 6004,
            "name": "UnknownOracleType",
            "msg": "oracle type cannot be determined"
        },
        {
            "code": 6005,
            "name": "InvalidFlashLoanTargetCpiProgram",
            "msg": ""
        },
        {
            "code": 6006,
            "name": "HealthMustBePositive",
            "msg": "health must be positive"
        },
        {
            "code": 6007,
            "name": "HealthMustBePositiveOrIncrease",
            "msg": "health must be positive or increase"
        },
        {
            "code": 6008,
            "name": "HealthMustBeNegative",
            "msg": "health must be negative"
        },
        {
            "code": 6009,
            "name": "IsBankrupt",
            "msg": "the account is bankrupt"
        },
        {
            "code": 6010,
            "name": "IsNotBankrupt",
            "msg": "the account is not bankrupt"
        },
        {
            "code": 6011,
            "name": "NoFreeTokenPositionIndex",
            "msg": "no free token position index"
        },
        {
            "code": 6012,
            "name": "NoFreeSerum3OpenOrdersIndex",
            "msg": "no free serum3 open orders index"
        },
        {
            "code": 6013,
            "name": "NoFreePerpPositionIndex",
            "msg": "no free perp position index"
        },
        {
            "code": 6014,
            "name": "Serum3OpenOrdersExistAlready",
            "msg": "serum3 open orders exist already"
        },
        {
            "code": 6015,
            "name": "InsufficentBankVaultFunds",
            "msg": "bank vault has insufficent funds"
        },
        {
            "code": 6016,
            "name": "BeingLiquidated",
            "msg": "account is currently being liquidated"
        },
        {
            "code": 6017,
            "name": "InvalidBank",
            "msg": "invalid bank"
        },
        {
            "code": 6018,
            "name": "ProfitabilityMismatch",
            "msg": "account profitability is mismatched"
        },
        {
            "code": 6019,
            "name": "CannotSettleWithSelf",
            "msg": "cannot settle with self"
        },
        {
            "code": 6020,
            "name": "PerpPositionDoesNotExist",
            "msg": "perp position does not exist"
        },
        {
            "code": 6021,
            "name": "MaxSettleAmountMustBeGreaterThanZero",
            "msg": "max settle amount must be greater than zero"
        },
        {
            "code": 6022,
            "name": "HasOpenPerpOrders",
            "msg": "the perp position has open orders or unprocessed fill events"
        },
        {
            "code": 6023,
            "name": "OracleConfidence",
            "msg": "an oracle does not reach the confidence threshold"
        },
        {
            "code": 6024,
            "name": "OracleStale",
            "msg": "an oracle is stale"
        },
        {
            "code": 6025,
            "name": "SettlementAmountMustBePositive",
            "msg": "settlement amount must always be positive"
        },
        {
            "code": 6026,
            "name": "BankBorrowLimitReached",
            "msg": "bank utilization has reached limit"
        },
        {
            "code": 6027,
            "name": "BankNetBorrowsLimitReached",
            "msg": "bank net borrows has reached limit - this is an intermittent error - the limit will reset regularly"
        },
        {
            "code": 6028,
            "name": "TokenPositionDoesNotExist",
            "msg": "token position does not exist"
        },
        {
            "code": 6029,
            "name": "DepositsIntoLiquidatingMustRecover",
            "msg": "token deposits into accounts that are being liquidated must bring their health above the init threshold"
        },
        {
            "code": 6030,
            "name": "TokenInReduceOnlyMode",
            "msg": "token is in reduce only mode"
        },
        {
            "code": 6031,
            "name": "MarketInReduceOnlyMode",
            "msg": "market is in reduce only mode"
        },
        {
            "code": 6032,
            "name": "GroupIsHalted",
            "msg": "group is halted"
        },
        {
            "code": 6033,
            "name": "PerpHasBaseLots",
            "msg": "the perp position has non-zero base lots"
        },
        {
            "code": 6034,
            "name": "HasOpenOrUnsettledSerum3Orders",
            "msg": "there are open or unsettled serum3 orders"
        },
        {
            "code": 6035,
            "name": "HasLiquidatableTokenPosition",
            "msg": "has liquidatable token position"
        },
        {
            "code": 6036,
            "name": "HasLiquidatablePerpBasePosition",
            "msg": "has liquidatable perp base position"
        },
        {
            "code": 6037,
            "name": "HasLiquidatablePositivePerpPnl",
            "msg": "has liquidatable positive perp pnl"
        },
        {
            "code": 6038,
            "name": "AccountIsFrozen",
            "msg": "account is frozen"
        },
        {
            "code": 6039,
            "name": "InitAssetWeightCantBeNegative",
            "msg": "Init Asset Weight can't be negative"
        },
        {
            "code": 6040,
            "name": "HasOpenPerpTakerFills",
            "msg": "has open perp taker fills"
        },
        {
            "code": 6041,
            "name": "DepositLimit",
            "msg": "deposit crosses the current group deposit limit"
        },
        {
            "code": 6042,
            "name": "IxIsDisabled",
            "msg": "instruction is disabled"
        },
        {
            "code": 6043,
            "name": "NoLiquidatablePerpBasePosition",
            "msg": "no liquidatable perp base position"
        },
        {
            "code": 6044,
            "name": "PerpOrderIdNotFound",
            "msg": "perp order id not found on the orderbook"
        }
    ]
};

async function sendTransaction(provider, ixs, alts, opts = {}) {
    const connection = provider.connection;
    const latestBlockhash = await connection.getLatestBlockhash(opts.preflightCommitment ??
        provider.opts.preflightCommitment ??
        'finalized');
    const payer = provider.wallet;
    if (opts.prioritizationFee) {
        ixs = [createComputeBudgetIx(opts.prioritizationFee), ...ixs];
    }
    const message = web3_js.MessageV0.compile({
        payerKey: provider.wallet.publicKey,
        instructions: ixs,
        recentBlockhash: latestBlockhash.blockhash,
        addressLookupTableAccounts: alts,
    });
    let vtx = new web3_js.VersionedTransaction(message);
    if (opts?.additionalSigners?.length) {
        vtx.sign([...opts?.additionalSigners]);
    }
    if (typeof payer.signTransaction === 'function' &&
        !(payer instanceof NodeWallet)) {
        vtx = (await payer.signTransaction(vtx));
    }
    else {
        // Maybe this path is only correct for NodeWallet?
        vtx.sign([payer.payer]);
    }
    const signature = await connection.sendRawTransaction(vtx.serialize(), {
        skipPreflight: true, // mergedOpts.skipPreflight,
    });
    // const signature = await connection.sendTransactionss(
    //   vtx as any as VersionedTransaction,
    //   {
    //     skipPreflight: true,
    //   },
    // );
    if (opts.postSendTxCallback) {
        try {
            opts.postSendTxCallback({ txid: signature });
        }
        catch (e) {
            console.warn(`postSendTxCallback error ${e}`);
        }
    }
    const txConfirmationCommitment = opts.txConfirmationCommitment ?? 'processed';
    let status;
    if (latestBlockhash.blockhash != null &&
        latestBlockhash.lastValidBlockHeight != null) {
        status = (await connection.confirmTransaction({
            signature: signature,
            blockhash: latestBlockhash.blockhash,
            lastValidBlockHeight: latestBlockhash.lastValidBlockHeight,
        }, txConfirmationCommitment)).value;
    }
    else {
        status = (await connection.confirmTransaction(signature, txConfirmationCommitment)).value;
    }
    if (status.err) {
        console.warn('Tx status: ', status);
        throw new MangoError({
            txid: signature,
            message: `${JSON.stringify(status)}`,
        });
    }
    return signature;
}
const createComputeBudgetIx = (microLamports) => {
    const computeBudgetIx = web3_js.ComputeBudgetProgram.setComputeUnitPrice({
        microLamports,
    });
    return computeBudgetIx;
};
class MangoError extends Error {
    message;
    txid;
    constructor({ txid, message }) {
        super();
        this.message = message;
        this.txid = txid;
    }
}

var AccountRetriever;
(function (AccountRetriever) {
    AccountRetriever[AccountRetriever["Scanning"] = 0] = "Scanning";
    AccountRetriever[AccountRetriever["Fixed"] = 1] = "Fixed";
})(AccountRetriever || (AccountRetriever = {}));
class MangoClient {
    program;
    programId;
    cluster;
    opts;
    idsSource;
    postSendTxCallback;
    prioritizationFee;
    txConfirmationCommitment;
    constructor(program, programId, cluster, opts = {}) {
        this.program = program;
        this.programId = programId;
        this.cluster = cluster;
        this.opts = opts;
        this.idsSource = opts?.idsSource || 'get-program-accounts';
        this.prioritizationFee = opts?.prioritizationFee || 0;
        this.postSendTxCallback = opts?.postSendTxCallback;
        this.txConfirmationCommitment =
            opts?.txConfirmationCommitment ??
                program.provider.opts.commitment ??
                'processed';
        // TODO: evil side effect, but limited backtraces are a nightmare
        Error.stackTraceLimit = 1000;
    }
    /// Transactions
    async sendAndConfirmTransaction(ixs, opts = {}) {
        return await sendTransaction(this.program.provider, ixs, opts.alts ?? [], {
            postSendTxCallback: this.postSendTxCallback,
            prioritizationFee: this.prioritizationFee,
            txConfirmationCommitment: this.txConfirmationCommitment,
            ...opts,
        });
    }
    async sendAndConfirmTransactionForGroup(group, ixs, opts = {}) {
        return await this.sendAndConfirmTransaction(ixs, {
            alts: group.addressLookupTablesList,
            ...opts,
        });
    }
    // Group
    async groupCreate(groupNum, testing, version, insuranceMintPk) {
        const adminPk = this.program.provider.wallet.publicKey;
        const ix = await this.program.methods
            .groupCreate(groupNum, testing ? 1 : 0, version)
            .accounts({
            creator: adminPk,
            payer: adminPk,
            insuranceMint: insuranceMintPk,
        })
            .instruction();
        return await this.sendAndConfirmTransaction([ix]);
    }
    async groupEdit(group, admin, fastListingAdmin, securityAdmin, testing, version, depositLimitQuote) {
        const ix = await this.program.methods
            .groupEdit(admin ?? null, fastListingAdmin ?? null, securityAdmin ?? null, testing ?? null, version ?? null, depositLimitQuote !== undefined ? depositLimitQuote : null)
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async ixGateSet(group, ixGateParams) {
        const ix = await this.program.methods
            .ixGateSet(buildIxGate(ixGateParams))
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async groupClose(group) {
        const adminPk = this.program.provider.wallet.publicKey;
        const ix = await this.program.methods
            .groupClose()
            .accounts({
            group: group.publicKey,
            insuranceVault: group.insuranceVault,
            admin: adminPk,
            solDestination: this.program.provider.wallet
                .publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async getGroup(groupPk) {
        const groupAccount = await this.program.account.group.fetch(groupPk);
        const group = Group.from(groupPk, groupAccount);
        await group.reloadAll(this);
        return group;
    }
    async getGroupsForCreator(creatorPk) {
        const filters = [
            {
                memcmp: {
                    bytes: creatorPk.toBase58(),
                    offset: 8,
                },
            },
        ];
        const groups = (await this.program.account.group.all(filters)).map((tuple) => Group.from(tuple.publicKey, tuple.account));
        groups.forEach((group) => group.reloadAll(this));
        return groups;
    }
    async getGroupForCreator(creatorPk, groupNum) {
        const bbuf = Buffer.alloc(4);
        bbuf.writeUInt32LE(groupNum);
        const filters = [
            {
                memcmp: {
                    bytes: creatorPk.toBase58(),
                    offset: 8,
                },
            },
            {
                memcmp: {
                    bytes: bs58.encode(bbuf),
                    offset: 40,
                },
            },
        ];
        const groups = (await this.program.account.group.all(filters)).map((tuple) => Group.from(tuple.publicKey, tuple.account));
        await groups[0].reloadAll(this);
        return groups[0];
    }
    async getIds(groupPk) {
        switch (this.idsSource) {
            case 'api':
                return await Id.fromApi(groupPk);
            case 'get-program-accounts':
                return undefined;
            case 'static':
                return Id.fromIdsByPk(groupPk);
        }
    }
    // Tokens/Banks
    async tokenRegister(group, mintPk, oraclePk, oracleConfig, tokenIndex, name, interestRateParams, loanFeeRate, loanOriginationFeeRate, maintAssetWeight, initAssetWeight, maintLiabWeight, initLiabWeight, liquidationFee, minVaultToDepositsRatio, netBorrowLimitWindowSizeTs, netBorrowLimitPerWindowQuote) {
        const ix = await this.program.methods
            .tokenRegister(tokenIndex, name, oracleConfig, interestRateParams, loanFeeRate, loanOriginationFeeRate, maintAssetWeight, initAssetWeight, maintLiabWeight, initLiabWeight, liquidationFee, minVaultToDepositsRatio, new anchor.BN(netBorrowLimitWindowSizeTs), new anchor.BN(netBorrowLimitPerWindowQuote))
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            mint: mintPk,
            oracle: oraclePk,
            payer: this.program.provider.wallet.publicKey,
            rent: web3_js.SYSVAR_RENT_PUBKEY,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async tokenRegisterTrustless(group, mintPk, oraclePk, tokenIndex, name) {
        const ix = await this.program.methods
            .tokenRegisterTrustless(tokenIndex, name)
            .accounts({
            group: group.publicKey,
            fastListingAdmin: this.program.provider.wallet
                .publicKey,
            mint: mintPk,
            oracle: oraclePk,
            payer: this.program.provider.wallet.publicKey,
            rent: web3_js.SYSVAR_RENT_PUBKEY,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async tokenEdit(group, mintPk, params) {
        const bank = group.getFirstBankByMint(mintPk);
        const mintInfo = group.mintInfosMapByTokenIndex.get(bank.tokenIndex);
        const ix = await this.program.methods
            .tokenEdit(params.oracle, params.oracleConfig, params.groupInsuranceFund, params.interestRateParams, params.loanFeeRate, params.loanOriginationFeeRate, params.maintAssetWeight, params.initAssetWeight, params.maintLiabWeight, params.initLiabWeight, params.liquidationFee, params.stablePriceDelayIntervalSeconds, params.stablePriceDelayGrowthLimit, params.stablePriceGrowthLimit, params.minVaultToDepositsRatio, params.netBorrowLimitPerWindowQuote !== null
            ? new anchor.BN(params.netBorrowLimitPerWindowQuote)
            : null, params.netBorrowLimitWindowSizeTs !== null
            ? new anchor.BN(params.netBorrowLimitWindowSizeTs)
            : null, params.borrowWeightScaleStartQuote, params.depositWeightScaleStartQuote, params.resetStablePrice ?? false, params.resetNetBorrowLimit ?? false, params.reduceOnly)
            .accounts({
            group: group.publicKey,
            oracle: params.oracle ?? bank.oracle,
            admin: this.program.provider.wallet.publicKey,
            mintInfo: mintInfo.publicKey,
        })
            .remainingAccounts([
            {
                pubkey: bank.publicKey,
                isWritable: true,
                isSigner: false,
            },
        ])
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async tokenDeregister(group, mintPk) {
        const bank = group.getFirstBankByMint(mintPk);
        const adminPk = this.program.provider.wallet.publicKey;
        const dustVaultPk = await getAssociatedTokenAddress(bank.mint, adminPk);
        const ai = await this.program.provider.connection.getAccountInfo(dustVaultPk);
        const preInstructions = [];
        if (!ai) {
            preInstructions.push(await createAssociatedTokenAccountIdempotentInstruction(adminPk, adminPk, bank.mint));
        }
        const ix = await this.program.methods
            .tokenDeregister()
            .accounts({
            group: group.publicKey,
            admin: adminPk,
            mintInfo: group.mintInfosMapByTokenIndex.get(bank.tokenIndex)
                ?.publicKey,
            dustVault: dustVaultPk,
            solDestination: this.program.provider.wallet
                .publicKey,
        })
            .remainingAccounts([bank.publicKey, bank.vault].map((pk) => ({ pubkey: pk, isWritable: true, isSigner: false })))
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [
            ...preInstructions,
            ix,
        ]);
    }
    async getBanksForGroup(group) {
        return (await this.program.account.bank.all([
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
        ])).map((tuple) => Bank.from(tuple.publicKey, tuple.account));
    }
    async getMintInfosForGroup(group) {
        return (await this.program.account.mintInfo.all([
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
        ])).map((tuple) => {
            return MintInfo.from(tuple.publicKey, tuple.account);
        });
    }
    async getMintInfoForTokenIndex(group, tokenIndex) {
        const tokenIndexBuf = Buffer.alloc(2);
        tokenIndexBuf.writeUInt16LE(tokenIndex);
        return (await this.program.account.mintInfo.all([
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
            {
                memcmp: {
                    bytes: bs58.encode(tokenIndexBuf),
                    offset: 40,
                },
            },
        ])).map((tuple) => {
            return MintInfo.from(tuple.publicKey, tuple.account);
        });
    }
    // Stub Oracle
    async stubOracleCreate(group, mintPk, price) {
        const ix = await this.program.methods
            .stubOracleCreate({ val: I80F48.fromNumber(price).getData() })
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            mint: mintPk,
            payer: this.program.provider.wallet.publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async stubOracleClose(group, oracle) {
        const ix = await this.program.methods
            .stubOracleClose()
            .accounts({
            group: group.publicKey,
            oracle: oracle,
            solDestination: this.program.provider.wallet
                .publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async stubOracleSet(group, oraclePk, price) {
        const ix = await this.program.methods
            .stubOracleSet({ val: I80F48.fromNumber(price).getData() })
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            oracle: oraclePk,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async getStubOracle(group, mintPk) {
        const filters = [
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
        ];
        if (mintPk) {
            filters.push({
                memcmp: {
                    bytes: mintPk.toBase58(),
                    offset: 40,
                },
            });
        }
        return (await this.program.account.stubOracle.all(filters)).map((pa) => StubOracle.from(pa.publicKey, pa.account));
    }
    // MangoAccount
    async getOrCreateMangoAccount(group, loadSerum3Oo = false) {
        const clientOwner = this.program.provider.wallet
            .publicKey;
        let mangoAccounts = await this.getMangoAccountsForOwner(group, this.program.provider.wallet.publicKey, loadSerum3Oo);
        if (mangoAccounts.length === 0) {
            await this.createMangoAccount(group);
            mangoAccounts = await this.getMangoAccountsForOwner(group, clientOwner, loadSerum3Oo);
        }
        return mangoAccounts.sort((a, b) => a.accountNum - b.accountNum)[0];
    }
    async createMangoAccount(group, accountNumber, name, tokenCount, serum3Count, perpCount, perpOoCount) {
        const ix = await this.program.methods
            .accountCreate(accountNumber ?? 0, tokenCount ?? 8, serum3Count ?? 8, perpCount ?? 8, perpOoCount ?? 8, name ?? '')
            .accounts({
            group: group.publicKey,
            owner: this.program.provider.wallet.publicKey,
            payer: this.program.provider.wallet.publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async createAndFetchMangoAccount(group, accountNumber, name, tokenCount, serum3Count, perpCount, perpOoCount, loadSerum3Oo = false) {
        const accNum = accountNumber ?? 0;
        await this.createMangoAccount(group, accNum, name, tokenCount, serum3Count, perpCount, perpOoCount);
        return await this.getMangoAccountForOwner(group, this.program.provider.wallet.publicKey, accNum, loadSerum3Oo);
    }
    async expandMangoAccount(group, account, tokenCount, serum3Count, perpCount, perpOoCount) {
        const ix = await this.program.methods
            .accountExpand(tokenCount, serum3Count, perpCount, perpOoCount)
            .accounts({
            group: group.publicKey,
            account: account.publicKey,
            owner: this.program.provider.wallet.publicKey,
            payer: this.program.provider.wallet.publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async editMangoAccount(group, mangoAccount, name, delegate) {
        const ix = await this.program.methods
            .accountEdit(name ?? null, delegate ?? null)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async computeAccountData(group, mangoAccount) {
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [mangoAccount], [], []);
        const ix = await this.program.methods
            .computeAccountData()
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async toggleMangoAccountFreeze(group, mangoAccount, freeze) {
        const ix = await this.program.methods
            .accountToggleFreeze(freeze)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            admin: this.program.provider.wallet.publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async getMangoAccount(mangoAccount, loadSerum3Oo = false) {
        const mangoAccountPk = mangoAccount instanceof MangoAccount
            ? mangoAccount.publicKey
            : mangoAccount;
        const mangoAccount_ = MangoAccount.from(mangoAccountPk, await this.program.account.mangoAccount.fetch(mangoAccountPk));
        if (loadSerum3Oo) {
            await mangoAccount_?.reloadSerum3OpenOrders(this);
        }
        return mangoAccount_;
    }
    async getMangoAccountWithSlot(mangoAccountPk, loadSerum3Oo = false) {
        const resp = await this.program.provider.connection.getAccountInfoAndContext(mangoAccountPk);
        if (!resp?.value)
            return;
        const decodedMangoAccount = this.program.coder.accounts.decode('mangoAccount', resp.value.data);
        const mangoAccount = MangoAccount.from(mangoAccountPk, decodedMangoAccount);
        if (loadSerum3Oo) {
            await mangoAccount?.reloadSerum3OpenOrders(this);
        }
        return { slot: resp.context.slot, value: mangoAccount };
    }
    async getMangoAccountForOwner(group, ownerPk, accountNumber, loadSerum3Oo = false) {
        const mangoAccounts = await this.getMangoAccountsForOwner(group, ownerPk, loadSerum3Oo);
        const foundMangoAccount = mangoAccounts.find((a) => a.accountNum == accountNumber);
        return foundMangoAccount;
    }
    async getMangoAccountsForOwner(group, ownerPk, loadSerum3Oo = false) {
        const accounts = (await this.program.account.mangoAccount.all([
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
            {
                memcmp: {
                    bytes: ownerPk.toBase58(),
                    offset: 40,
                },
            },
        ])).map((pa) => {
            return MangoAccount.from(pa.publicKey, pa.account);
        });
        if (loadSerum3Oo) {
            await Promise.all(accounts.map(async (a) => await a.reloadSerum3OpenOrders(this)));
        }
        return accounts;
    }
    async getMangoAccountsForDelegate(group, delegate, loadSerum3Oo = false) {
        const accounts = (await this.program.account.mangoAccount.all([
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
            {
                memcmp: {
                    bytes: delegate.toBase58(),
                    offset: 104,
                },
            },
        ])).map((pa) => {
            return MangoAccount.from(pa.publicKey, pa.account);
        });
        if (loadSerum3Oo) {
            await Promise.all(accounts.map(async (a) => await a.reloadSerum3OpenOrders(this)));
        }
        return accounts;
    }
    async getAllMangoAccounts(group, loadSerum3Oo = false) {
        const accounts = (await this.program.account.mangoAccount.all([
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
        ])).map((pa) => {
            return MangoAccount.from(pa.publicKey, pa.account);
        });
        if (loadSerum3Oo) {
            await Promise.all(accounts.map(async (a) => await a.reloadSerum3OpenOrders(this)));
        }
        return accounts;
    }
    /**
     * Note: this ix doesn't settle liabs, reduce open positions, or withdraw tokens to wallet,
     * it simply closes the account. To close successfully ensure all positions are closed, or
     * use forceClose flag
     * @param group
     * @param mangoAccount
     * @param forceClose
     * @returns
     */
    async closeMangoAccount(group, mangoAccount, forceClose = false) {
        const ix = await this.program.methods
            .accountClose(forceClose)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
            solDestination: mangoAccount.owner,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async emptyAndCloseMangoAccount(group, mangoAccount) {
        const instructions = [];
        const healthAccountsToExclude = [];
        for (const serum3Account of mangoAccount.serum3Active()) {
            const serum3Market = group.serum3MarketsMapByMarketIndex.get(serum3Account.marketIndex);
            const closeOOIx = await this.serum3CloseOpenOrdersIx(group, mangoAccount, serum3Market.serumMarketExternal);
            healthAccountsToExclude.push(serum3Account.openOrders);
            instructions.push(closeOOIx);
        }
        for (const perp of mangoAccount.perpActive()) {
            const perpMarketIndex = perp.marketIndex;
            const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
            const deactivatingPositionIx = await this.perpDeactivatePositionIx(group, mangoAccount, perpMarketIndex);
            healthAccountsToExclude.push(perpMarket.publicKey, perpMarket.oracle);
            instructions.push(deactivatingPositionIx);
        }
        for (const index in mangoAccount.tokensActive()) {
            const indexNum = Number(index);
            const accountsToExclude = [...healthAccountsToExclude];
            const token = mangoAccount.tokensActive()[indexNum];
            const bank = group.getFirstBankByTokenIndex(token.tokenIndex);
            //to withdraw from all token accounts we need to exclude previous tokens pubkeys
            //used to build health remaining accounts
            if (indexNum !== 0) {
                for (let i = indexNum; i--; i >= 0) {
                    const prevToken = mangoAccount.tokensActive()[i];
                    const prevBank = group.getFirstBankByTokenIndex(prevToken.tokenIndex);
                    accountsToExclude.push(prevBank.publicKey, prevBank.oracle);
                }
            }
            const withdrawIx = await this.tokenWithdrawNativeIx(group, mangoAccount, bank.mint, U64_MAX_BN, false, [...accountsToExclude]);
            instructions.push(...withdrawIx);
        }
        const closeIx = await this.program.methods
            .accountClose(false)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
            solDestination: mangoAccount.owner,
        })
            .instruction();
        instructions.push(closeIx);
        return await this.sendAndConfirmTransactionForGroup(group, instructions);
    }
    async tokenDeposit(group, mangoAccount, mintPk, amount, reduceOnly = false) {
        const decimals = group.getMintDecimals(mintPk);
        const nativeAmount = toNative(amount, decimals);
        return await this.tokenDepositNative(group, mangoAccount, mintPk, nativeAmount, reduceOnly);
    }
    async tokenDepositNative(group, mangoAccount, mintPk, nativeAmount, reduceOnly = false) {
        const bank = group.getFirstBankByMint(mintPk);
        const tokenAccountPk = await getAssociatedTokenAddress(mintPk, mangoAccount.owner);
        let wrappedSolAccount;
        let preInstructions = [];
        let postInstructions = [];
        const additionalSigners = [];
        if (mintPk.equals(tokenInstructions.WRAPPED_SOL_MINT)) {
            wrappedSolAccount = new web3_js.Keypair();
            const lamports = nativeAmount.add(new anchor.BN(1e7));
            preInstructions = [
                web3_js.SystemProgram.createAccount({
                    fromPubkey: mangoAccount.owner,
                    newAccountPubkey: wrappedSolAccount.publicKey,
                    lamports: lamports.toNumber(),
                    space: 165,
                    programId: splToken.TOKEN_PROGRAM_ID,
                }),
                tokenInstructions.initializeAccount({
                    account: wrappedSolAccount.publicKey,
                    mint: tokenInstructions.WRAPPED_SOL_MINT,
                    owner: mangoAccount.owner,
                }),
            ];
            postInstructions = [
                tokenInstructions.closeAccount({
                    source: wrappedSolAccount.publicKey,
                    destination: mangoAccount.owner,
                    owner: mangoAccount.owner,
                }),
            ];
            additionalSigners.push(wrappedSolAccount);
        }
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [mangoAccount], [bank], []);
        const ix = await this.program.methods
            .tokenDeposit(new anchor.BN(nativeAmount), reduceOnly)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: mangoAccount.owner,
            bank: bank.publicKey,
            vault: bank.vault,
            oracle: bank.oracle,
            tokenAccount: wrappedSolAccount?.publicKey ?? tokenAccountPk,
            tokenAuthority: mangoAccount.owner,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [...preInstructions, ix, ...postInstructions], { additionalSigners });
    }
    async tokenWithdraw(group, mangoAccount, mintPk, amount, allowBorrow) {
        const nativeAmount = toNative(amount, group.getMintDecimals(mintPk));
        const ixes = await this.tokenWithdrawNativeIx(group, mangoAccount, mintPk, nativeAmount, allowBorrow);
        return await this.sendAndConfirmTransactionForGroup(group, ixes);
    }
    async tokenWithdrawNativeIx(group, mangoAccount, mintPk, nativeAmount, allowBorrow, healthAccountsToExclude = []) {
        const bank = group.getFirstBankByMint(mintPk);
        const tokenAccountPk = await getAssociatedTokenAddress(bank.mint, mangoAccount.owner);
        // ensure withdraws don't fail with missing ATAs
        const preInstructions = [
            await createAssociatedTokenAccountIdempotentInstruction(mangoAccount.owner, mangoAccount.owner, bank.mint),
        ];
        const postInstructions = [];
        if (mintPk.equals(tokenInstructions.WRAPPED_SOL_MINT)) {
            postInstructions.push(tokenInstructions.closeAccount({
                source: tokenAccountPk,
                destination: mangoAccount.owner,
                owner: mangoAccount.owner,
            }));
        }
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [mangoAccount], [bank], []);
        const ix = await this.program.methods
            .tokenWithdraw(new anchor.BN(nativeAmount), allowBorrow)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: mangoAccount.owner,
            bank: bank.publicKey,
            vault: bank.vault,
            oracle: bank.oracle,
            tokenAccount: tokenAccountPk,
        })
            .remainingAccounts(healthRemainingAccounts
            .filter((accounts) => !healthAccountsToExclude.find((accountsToExclude) => accounts.equals(accountsToExclude)))
            .map((pk) => ({
            pubkey: pk,
            isWritable: false,
            isSigner: false,
        })))
            .instruction();
        return [...preInstructions, ix, ...postInstructions];
    }
    async tokenWithdrawNative(group, mangoAccount, mintPk, nativeAmount, allowBorrow, healthAccountsToExclude = []) {
        const ixs = await this.tokenWithdrawNativeIx(group, mangoAccount, mintPk, nativeAmount, allowBorrow, healthAccountsToExclude);
        return await this.sendAndConfirmTransactionForGroup(group, ixs);
    }
    // Serum
    async serum3RegisterMarket(group, serum3MarketExternalPk, baseBank, quoteBank, marketIndex, name) {
        const ix = await this.program.methods
            .serum3RegisterMarket(marketIndex, name)
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            serumProgram: OPENBOOK_PROGRAM_ID[this.cluster],
            serumMarketExternal: serum3MarketExternalPk,
            baseBank: baseBank.publicKey,
            quoteBank: quoteBank.publicKey,
            payer: this.program.provider.wallet.publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async serum3deregisterMarket(group, externalMarketPk) {
        const serum3Market = group.serum3MarketsMapByExternal.get(externalMarketPk.toBase58());
        const marketIndexBuf = Buffer.alloc(2);
        marketIndexBuf.writeUInt16LE(serum3Market.marketIndex);
        const [indexReservation] = await web3_js.PublicKey.findProgramAddress([Buffer.from('Serum3Index'), group.publicKey.toBuffer(), marketIndexBuf], this.program.programId);
        const ix = await this.program.methods
            .serum3DeregisterMarket()
            .accounts({
            group: group.publicKey,
            serumMarket: serum3Market.publicKey,
            indexReservation,
            solDestination: this.program.provider.wallet
                .publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async serum3GetMarkets(group, baseTokenIndex, quoteTokenIndex) {
        const bumpfbuf = Buffer.alloc(1);
        bumpfbuf.writeUInt8(255);
        const filters = [
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
        ];
        if (baseTokenIndex) {
            const bbuf = Buffer.alloc(2);
            bbuf.writeUInt16LE(baseTokenIndex);
            filters.push({
                memcmp: {
                    bytes: bs58.encode(bbuf),
                    offset: 40,
                },
            });
        }
        if (quoteTokenIndex) {
            const qbuf = Buffer.alloc(2);
            qbuf.writeUInt16LE(quoteTokenIndex);
            filters.push({
                memcmp: {
                    bytes: bs58.encode(qbuf),
                    offset: 42,
                },
            });
        }
        return (await this.program.account.serum3Market.all(filters)).map((tuple) => Serum3Market.from(tuple.publicKey, tuple.account));
    }
    async serum3CreateOpenOrders(group, mangoAccount, externalMarketPk) {
        const serum3Market = group.serum3MarketsMapByExternal.get(externalMarketPk.toBase58());
        const ix = await this.program.methods
            .serum3CreateOpenOrders()
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            serumMarket: serum3Market.publicKey,
            serumProgram: serum3Market.serumProgram,
            serumMarketExternal: serum3Market.serumMarketExternal,
            owner: this.program.provider.wallet.publicKey,
            payer: this.program.provider.wallet.publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async serum3CreateOpenOrdersIx(group, mangoAccount, externalMarketPk) {
        const serum3Market = group.serum3MarketsMapByExternal.get(externalMarketPk.toBase58());
        const ix = await this.program.methods
            .serum3CreateOpenOrders()
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            serumMarket: serum3Market.publicKey,
            serumProgram: serum3Market.serumProgram,
            serumMarketExternal: serum3Market.serumMarketExternal,
            owner: this.program.provider.wallet.publicKey,
            payer: this.program.provider.wallet.publicKey,
        })
            .instruction();
        return ix;
    }
    async serum3CloseOpenOrdersIx(group, mangoAccount, externalMarketPk) {
        const serum3Market = group.serum3MarketsMapByExternal.get(externalMarketPk.toBase58());
        const openOrders = mangoAccount.serum3.find((account) => account.marketIndex === serum3Market.marketIndex)?.openOrders;
        return await this.program.methods
            .serum3CloseOpenOrders()
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            serumMarket: serum3Market.publicKey,
            serumProgram: serum3Market.serumProgram,
            serumMarketExternal: serum3Market.serumMarketExternal,
            openOrders,
            solDestination: this.program.provider.wallet
                .publicKey,
        })
            .instruction();
    }
    async serum3CloseOpenOrders(group, mangoAccount, externalMarketPk) {
        const ix = await this.serum3CloseOpenOrdersIx(group, mangoAccount, externalMarketPk);
        return await sendTransaction(this.program.provider, [ix], group.addressLookupTablesList, {
            postSendTxCallback: this.postSendTxCallback,
        });
    }
    async serum3PlaceOrderIx(group, mangoAccount, externalMarketPk, side, price, size, selfTradeBehavior, orderType, clientOrderId, limit) {
        const ixs = [];
        const serum3Market = group.serum3MarketsMapByExternal.get(externalMarketPk.toBase58());
        let openOrderPk = undefined;
        const banks = [];
        const openOrdersForMarket = [];
        if (!mangoAccount.getSerum3Account(serum3Market.marketIndex)) {
            const ix = await this.serum3CreateOpenOrdersIx(group, mangoAccount, serum3Market.serumMarketExternal);
            ixs.push(ix);
            openOrderPk = await serum3Market.findOoPda(this.program.programId, mangoAccount.publicKey);
            openOrdersForMarket.push([serum3Market, openOrderPk]);
            const baseTokenIndex = serum3Market.baseTokenIndex;
            const quoteTokenIndex = serum3Market.quoteTokenIndex;
            // only include banks if no deposit has been previously made for same token
            if (!mangoAccount.getToken(baseTokenIndex)?.isActive()) {
                banks.push(group.getFirstBankByTokenIndex(baseTokenIndex));
            }
            if (!mangoAccount.getToken(quoteTokenIndex)?.isActive()) {
                banks.push(group.getFirstBankByTokenIndex(quoteTokenIndex));
            }
        }
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [mangoAccount], banks, [], openOrdersForMarket);
        const serum3MarketExternal = group.serum3ExternalMarketsMap.get(externalMarketPk.toBase58());
        const serum3MarketExternalVaultSigner = await generateSerum3MarketExternalVaultSignerAddress(this.cluster, serum3Market, serum3MarketExternal);
        const limitPrice = serum3MarketExternal.priceNumberToLots(price);
        const maxBaseQuantity = serum3MarketExternal.baseSizeNumberToLots(size);
        const isTaker = orderType !== Serum3OrderType.postOnly;
        const maxQuoteQuantity = new anchor.BN(serum3MarketExternal.decoded.quoteLotSize.toNumber() *
            (1 + Math.max(serum3Market.getFeeRates(isTaker), 0)) *
            serum3MarketExternal.baseSizeNumberToLots(size).toNumber() *
            serum3MarketExternal.priceNumberToLots(price).toNumber());
        const payerTokenIndex = (() => {
            if (side == Serum3Side.bid) {
                return serum3Market.quoteTokenIndex;
            }
            else {
                return serum3Market.baseTokenIndex;
            }
        })();
        const payerBank = group.getFirstBankByTokenIndex(payerTokenIndex);
        const ix = await this.program.methods
            .serum3PlaceOrder(side, limitPrice, maxBaseQuantity, maxQuoteQuantity, selfTradeBehavior, orderType, new anchor.BN(clientOrderId), limit)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
            openOrders: openOrderPk ||
                mangoAccount.getSerum3Account(serum3Market.marketIndex)?.openOrders,
            serumMarket: serum3Market.publicKey,
            serumProgram: OPENBOOK_PROGRAM_ID[this.cluster],
            serumMarketExternal: serum3Market.serumMarketExternal,
            marketBids: serum3MarketExternal.bidsAddress,
            marketAsks: serum3MarketExternal.asksAddress,
            marketEventQueue: serum3MarketExternal.decoded.eventQueue,
            marketRequestQueue: serum3MarketExternal.decoded.requestQueue,
            marketBaseVault: serum3MarketExternal.decoded.baseVault,
            marketQuoteVault: serum3MarketExternal.decoded.quoteVault,
            marketVaultSigner: serum3MarketExternalVaultSigner,
            payerBank: payerBank.publicKey,
            payerVault: payerBank.vault,
            payerOracle: payerBank.oracle,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .instruction();
        ixs.push(ix);
        return ixs;
    }
    async serum3PlaceOrder(group, mangoAccount, externalMarketPk, side, price, size, selfTradeBehavior, orderType, clientOrderId, limit) {
        const placeOrderIxes = await this.serum3PlaceOrderIx(group, mangoAccount, externalMarketPk, side, price, size, selfTradeBehavior, orderType, clientOrderId, limit);
        const settleIx = await this.serum3SettleFundsIx(group, mangoAccount, externalMarketPk);
        return await this.sendAndConfirmTransactionForGroup(group, [
            ...placeOrderIxes,
            settleIx,
        ]);
    }
    async serum3CancelAllOrders(group, mangoAccount, externalMarketPk, limit) {
        const serum3Market = group.serum3MarketsMapByExternal.get(externalMarketPk.toBase58());
        const serum3MarketExternal = group.serum3ExternalMarketsMap.get(externalMarketPk.toBase58());
        const ix = await this.program.methods
            .serum3CancelAllOrders(limit ? limit : 10)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
            openOrders: mangoAccount.getSerum3Account(serum3Market.marketIndex)
                ?.openOrders,
            serumMarket: serum3Market.publicKey,
            serumProgram: OPENBOOK_PROGRAM_ID[this.cluster],
            serumMarketExternal: serum3Market.serumMarketExternal,
            marketBids: serum3MarketExternal.bidsAddress,
            marketAsks: serum3MarketExternal.asksAddress,
            marketEventQueue: serum3MarketExternal.decoded.eventQueue,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async serum3SettleFundsIx(group, mangoAccount, externalMarketPk) {
        const serum3Market = group.serum3MarketsMapByExternal.get(externalMarketPk.toBase58());
        const serum3MarketExternal = group.serum3ExternalMarketsMap.get(externalMarketPk.toBase58());
        const [serum3MarketExternalVaultSigner, openOrderPublicKey] = await Promise.all([
            generateSerum3MarketExternalVaultSignerAddress(this.cluster, serum3Market, serum3MarketExternal),
            serum3Market.findOoPda(this.program.programId, mangoAccount.publicKey),
        ]);
        const ix = await this.program.methods
            .serum3SettleFunds()
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
            openOrders: openOrderPublicKey,
            serumMarket: serum3Market.publicKey,
            serumProgram: OPENBOOK_PROGRAM_ID[this.cluster],
            serumMarketExternal: serum3Market.serumMarketExternal,
            marketBaseVault: serum3MarketExternal.decoded.baseVault,
            marketQuoteVault: serum3MarketExternal.decoded.quoteVault,
            marketVaultSigner: serum3MarketExternalVaultSigner,
            quoteBank: group.getFirstBankByTokenIndex(serum3Market.quoteTokenIndex)
                .publicKey,
            quoteVault: group.getFirstBankByTokenIndex(serum3Market.quoteTokenIndex)
                .vault,
            baseBank: group.getFirstBankByTokenIndex(serum3Market.baseTokenIndex)
                .publicKey,
            baseVault: group.getFirstBankByTokenIndex(serum3Market.baseTokenIndex)
                .vault,
        })
            .instruction();
        return ix;
    }
    async serum3SettleFunds(group, mangoAccount, externalMarketPk) {
        const ix = await this.serum3SettleFundsIx(group, mangoAccount, externalMarketPk);
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async serum3CancelOrderIx(group, mangoAccount, externalMarketPk, side, orderId) {
        const serum3Market = group.serum3MarketsMapByExternal.get(externalMarketPk.toBase58());
        const serum3MarketExternal = group.serum3ExternalMarketsMap.get(externalMarketPk.toBase58());
        const ix = await this.program.methods
            .serum3CancelOrder(side, orderId)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            openOrders: mangoAccount.getSerum3Account(serum3Market.marketIndex)
                ?.openOrders,
            serumMarket: serum3Market.publicKey,
            serumProgram: OPENBOOK_PROGRAM_ID[this.cluster],
            serumMarketExternal: serum3Market.serumMarketExternal,
            marketBids: serum3MarketExternal.bidsAddress,
            marketAsks: serum3MarketExternal.asksAddress,
            marketEventQueue: serum3MarketExternal.decoded.eventQueue,
        })
            .instruction();
        return ix;
    }
    async serum3CancelOrder(group, mangoAccount, externalMarketPk, side, orderId) {
        const ixes = await Promise.all([
            this.serum3CancelOrderIx(group, mangoAccount, externalMarketPk, side, orderId),
            this.serum3SettleFundsIx(group, mangoAccount, externalMarketPk),
        ]);
        return await this.sendAndConfirmTransactionForGroup(group, ixes);
    }
    /// perps
    async perpCreateMarket(group, oraclePk, perpMarketIndex, name, oracleConfig, baseDecimals, quoteLotSize, baseLotSize, maintBaseAssetWeight, initBaseAssetWeight, maintBaseLiabWeight, initBaseLiabWeight, maintOverallAssetWeight, initOverallAssetWeight, baseLiquidationFee, makerFee, takerFee, feePenalty, minFunding, maxFunding, impactQuantity, groupInsuranceFund, settleFeeFlat, settleFeeAmountThreshold, settleFeeFractionLowHealth, settleTokenIndex, settlePnlLimitFactor, settlePnlLimitWindowSize, positivePnlLiquidationFee) {
        const bids = new web3_js.Keypair();
        const asks = new web3_js.Keypair();
        const eventQueue = new web3_js.Keypair();
        const bookSideSize = this.program._coder.accounts.size(this.program.account.bookSide._idlAccount);
        const eventQueueSize = this.program._coder.accounts.size(this.program.account.eventQueue._idlAccount);
        const ix = await this.program.methods
            .perpCreateMarket(perpMarketIndex, name, oracleConfig, baseDecimals, new anchor.BN(quoteLotSize), new anchor.BN(baseLotSize), maintBaseAssetWeight, initBaseAssetWeight, maintBaseLiabWeight, initBaseLiabWeight, maintOverallAssetWeight, initOverallAssetWeight, baseLiquidationFee, makerFee, takerFee, minFunding, maxFunding, new anchor.BN(impactQuantity), groupInsuranceFund, feePenalty, settleFeeFlat, settleFeeAmountThreshold, settleFeeFractionLowHealth, settleTokenIndex, settlePnlLimitFactor, new anchor.BN(settlePnlLimitWindowSize), positivePnlLiquidationFee)
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            oracle: oraclePk,
            bids: bids.publicKey,
            asks: asks.publicKey,
            eventQueue: eventQueue.publicKey,
            payer: this.program.provider.wallet.publicKey,
        })
            .instruction();
        const preInstructions = [
            // book sides
            web3_js.SystemProgram.createAccount({
                programId: this.program.programId,
                space: bookSideSize,
                lamports: await this.program.provider.connection.getMinimumBalanceForRentExemption(bookSideSize),
                fromPubkey: this.program.provider.wallet.publicKey,
                newAccountPubkey: bids.publicKey,
            }),
            web3_js.SystemProgram.createAccount({
                programId: this.program.programId,
                space: bookSideSize,
                lamports: await this.program.provider.connection.getMinimumBalanceForRentExemption(bookSideSize),
                fromPubkey: this.program.provider.wallet.publicKey,
                newAccountPubkey: asks.publicKey,
            }),
            // event queue
            web3_js.SystemProgram.createAccount({
                programId: this.program.programId,
                space: eventQueueSize,
                lamports: await this.program.provider.connection.getMinimumBalanceForRentExemption(eventQueueSize),
                fromPubkey: this.program.provider.wallet.publicKey,
                newAccountPubkey: eventQueue.publicKey,
            }),
        ];
        return await this.sendAndConfirmTransactionForGroup(group, [...preInstructions, ix], {
            additionalSigners: [bids, asks, eventQueue],
        });
    }
    async perpEditMarket(group, perpMarketIndex, params) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const ix = await this.program.methods
            .perpEditMarket(params.oracle, params.oracleConfig, params.baseDecimals, params.maintBaseAssetWeight, params.initBaseAssetWeight, params.maintBaseLiabWeight, params.initBaseLiabWeight, params.maintOverallAssetWeight, params.initOverallAssetWeight, params.baseLiquidationFee, params.makerFee, params.takerFee, params.minFunding, params.maxFunding, params.impactQuantity !== null ? new anchor.BN(params.impactQuantity) : null, params.groupInsuranceFund, params.feePenalty, params.settleFeeFlat, params.settleFeeAmountThreshold, params.settleFeeFractionLowHealth, params.stablePriceDelayIntervalSeconds, params.stablePriceDelayGrowthLimit, params.stablePriceGrowthLimit, params.settlePnlLimitFactor, params.settlePnlLimitWindowSize !== null
            ? new anchor.BN(params.settlePnlLimitWindowSize)
            : null, params.reduceOnly, params.resetStablePrice ?? false, params.positivePnlLiquidationFee)
            .accounts({
            group: group.publicKey,
            oracle: params.oracle ?? perpMarket.oracle,
            admin: this.program.provider.wallet.publicKey,
            perpMarket: perpMarket.publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async perpCloseMarket(group, perpMarketIndex) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const ix = await this.program.methods
            .perpCloseMarket()
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            perpMarket: perpMarket.publicKey,
            bids: perpMarket.bids,
            asks: perpMarket.asks,
            eventQueue: perpMarket.eventQueue,
            solDestination: this.program.provider.wallet
                .publicKey,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async perpGetMarkets(group) {
        const bumpfbuf = Buffer.alloc(1);
        bumpfbuf.writeUInt8(255);
        const filters = [
            {
                memcmp: {
                    bytes: group.publicKey.toBase58(),
                    offset: 8,
                },
            },
        ];
        return (await this.program.account.perpMarket.all(filters)).map((tuple) => PerpMarket.from(tuple.publicKey, tuple.account));
    }
    async perpDeactivatePositionIx(group, mangoAccount, perpMarketIndex) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [mangoAccount], [], []);
        return await this.program.methods
            .perpDeactivatePosition()
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            perpMarket: perpMarket.publicKey,
            owner: this.program.provider.wallet.publicKey,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .instruction();
    }
    async perpDeactivatePosition(group, mangoAccount, perpMarketIndex) {
        const ix = await this.perpDeactivatePositionIx(group, mangoAccount, perpMarketIndex);
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    // perpPlaceOrder ix returns an optional, custom order id,
    // but, since we use a customer tx sender, this method
    // doesn't return it
    async perpPlaceOrder(group, mangoAccount, perpMarketIndex, side, price, quantity, maxQuoteQuantity, clientOrderId, orderType, reduceOnly, expiryTimestamp, limit) {
        const ix = await this.perpPlaceOrderIx(group, mangoAccount, perpMarketIndex, side, price, quantity, maxQuoteQuantity, clientOrderId, orderType, reduceOnly, expiryTimestamp, limit);
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async perpPlaceOrderIx(group, mangoAccount, perpMarketIndex, side, price, quantity, maxQuoteQuantity, clientOrderId, orderType, reduceOnly, expiryTimestamp, limit) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [mangoAccount], 
        // Settlement token bank, because a position for it may be created
        [group.getFirstBankByTokenIndex(0)], [perpMarket]);
        return await this.program.methods
            .perpPlaceOrder(side, perpMarket.uiPriceToLots(price), perpMarket.uiBaseToLots(quantity), maxQuoteQuantity
            ? perpMarket.uiQuoteToLots(maxQuoteQuantity)
            : I64_MAX_BN, new anchor.BN(clientOrderId ? clientOrderId : Date.now()), orderType ? orderType : PerpOrderType.limit, reduceOnly ? reduceOnly : false, new anchor.BN(expiryTimestamp ? expiryTimestamp : 0), limit ? limit : 10)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            perpMarket: perpMarket.publicKey,
            bids: perpMarket.bids,
            asks: perpMarket.asks,
            eventQueue: perpMarket.eventQueue,
            oracle: perpMarket.oracle,
            owner: this.program.provider.wallet.publicKey,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .instruction();
    }
    async perpPlaceOrderPegged(group, mangoAccount, perpMarketIndex, side, priceOffset, pegLimit, quantity, maxQuoteQuantity, clientOrderId, orderType, reduceOnly, expiryTimestamp, limit) {
        const ix = await this.perpPlaceOrderPeggedIx(group, mangoAccount, perpMarketIndex, side, priceOffset, pegLimit, quantity, maxQuoteQuantity, clientOrderId, orderType, reduceOnly, expiryTimestamp, limit);
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async perpPlaceOrderPeggedIx(group, mangoAccount, perpMarketIndex, side, priceOffset, pegLimit, quantity, maxQuoteQuantity, clientOrderId, orderType, reduceOnly, expiryTimestamp, limit) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [mangoAccount], 
        // Settlement token bank, because a position for it may be created
        [group.getFirstBankByTokenIndex(0)], [perpMarket]);
        return await this.program.methods
            .perpPlaceOrderPegged(side, perpMarket.uiPriceToLots(priceOffset), perpMarket.uiPriceToLots(pegLimit), perpMarket.uiBaseToLots(quantity), maxQuoteQuantity
            ? perpMarket.uiQuoteToLots(maxQuoteQuantity)
            : I64_MAX_BN, new anchor.BN(clientOrderId ?? Date.now()), orderType ? orderType : PerpOrderType.limit, reduceOnly ? reduceOnly : false, new anchor.BN(expiryTimestamp ?? 0), limit ? limit : 10, -1)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            perpMarket: perpMarket.publicKey,
            bids: perpMarket.bids,
            asks: perpMarket.asks,
            eventQueue: perpMarket.eventQueue,
            oracle: perpMarket.oracle,
            owner: this.program.provider.wallet.publicKey,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .instruction();
    }
    async perpCancelOrderIx(group, mangoAccount, perpMarketIndex, orderId) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        return await this.program.methods
            .perpCancelOrder(new anchor.BN(orderId))
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
            perpMarket: perpMarket.publicKey,
            bids: perpMarket.bids,
            asks: perpMarket.asks,
        })
            .instruction();
    }
    async perpCancelOrder(group, mangoAccount, perpMarketIndex, orderId) {
        const ix = await this.perpCancelOrderIx(group, mangoAccount, perpMarketIndex, orderId);
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async perpCancelAllOrders(group, mangoAccount, perpMarketIndex, limit) {
        const ix = await this.perpCancelAllOrdersIx(group, mangoAccount, perpMarketIndex, limit);
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async perpCancelAllOrdersIx(group, mangoAccount, perpMarketIndex, limit) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        return await this.program.methods
            .perpCancelAllOrders(limit)
            .accounts({
            group: group.publicKey,
            account: mangoAccount.publicKey,
            perpMarket: perpMarket.publicKey,
            bids: perpMarket.bids,
            asks: perpMarket.asks,
            owner: this.program.provider.wallet.publicKey,
        })
            .instruction();
    }
    async perpSettlePnl(group, profitableAccount, unprofitableAccount, settler, perpMarketIndex) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Scanning, group, [profitableAccount, unprofitableAccount], [group.getFirstBankByTokenIndex(0)], [perpMarket]);
        const bank = group.banksMapByTokenIndex.get(0)[0];
        const ix = await this.program.methods
            .perpSettlePnl()
            .accounts({
            group: group.publicKey,
            accountA: profitableAccount.publicKey,
            accountB: unprofitableAccount.publicKey,
            perpMarket: perpMarket.publicKey,
            oracle: perpMarket.oracle,
            settleOracle: bank.oracle,
            settleBank: bank.publicKey,
            settler: settler.publicKey,
            settlerOwner: this.program.provider.wallet
                .publicKey,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async perpSettleFees(group, account, perpMarketIndex, maxSettleAmount) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [account], // Account must be unprofitable
        [group.getFirstBankByTokenIndex(0)], [perpMarket]);
        const bank = group.banksMapByTokenIndex.get(0)[0];
        const ix = await this.program.methods
            .perpSettleFees(maxSettleAmount)
            .accounts({
            group: group.publicKey,
            account: account.publicKey,
            perpMarket: perpMarket.publicKey,
            oracle: perpMarket.oracle,
            settleOracle: bank.oracle,
            settleBank: bank.publicKey,
        })
            .remainingAccounts(healthRemainingAccounts.map((pk) => ({ pubkey: pk, isWritable: false, isSigner: false })))
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async perpConsumeEvents(group, perpMarketIndex, accounts, limit) {
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const ix = await this.program.methods
            .perpConsumeEvents(new anchor.BN(limit))
            .accounts({
            group: group.publicKey,
            perpMarket: perpMarket.publicKey,
            eventQueue: perpMarket.eventQueue,
        })
            .remainingAccounts(accounts.map((pk) => ({ pubkey: pk, isWritable: true, isSigner: false })))
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async perpConsumeAllEvents(group, perpMarketIndex) {
        const limit = 8;
        const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
        const eventQueue = await perpMarket.loadEventQueue(this);
        const unconsumedEvents = eventQueue.getUnconsumedEvents();
        while (unconsumedEvents.length > 0) {
            const events = unconsumedEvents.splice(0, limit);
            const accounts = events
                .map((ev) => {
                switch (ev.eventType) {
                    case PerpEventQueue.FILL_EVENT_TYPE: {
                        const fill = ev;
                        return [fill.maker, fill.taker];
                    }
                    case PerpEventQueue.OUT_EVENT_TYPE: {
                        const out = ev;
                        return [out.owner];
                    }
                    case PerpEventQueue.LIQUIDATE_EVENT_TYPE:
                        return [];
                    default:
                        throw new Error(`Unknown event with eventType ${ev.eventType}!`);
                }
            })
                .flat();
            await this.perpConsumeEvents(group, perpMarketIndex, accounts, limit);
        }
    }
    async marginTrade({ group, mangoAccount, inputMintPk, amountIn, outputMintPk, userDefinedInstructions, userDefinedAlts = [], 
    // margin trade is a general function
    // set flash_loan_type to FlashLoanType.swap if you desire the transaction to be recorded as a swap
    flashLoanType, }) {
        const inputBank = group.getFirstBankByMint(inputMintPk);
        const outputBank = group.getFirstBankByMint(outputMintPk);
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [mangoAccount], [inputBank, outputBank], []);
        const parsedHealthAccounts = healthRemainingAccounts.map((pk) => ({
            pubkey: pk,
            isWritable: false,
            isSigner: false,
        }));
        /*
         * Find or create associated token accounts
         */
        const inputTokenAccountPk = await getAssociatedTokenAddress(inputBank.mint, mangoAccount.owner);
        const inputTokenAccExists = await this.program.provider.connection.getAccountInfo(inputTokenAccountPk);
        const preInstructions = [];
        if (!inputTokenAccExists) {
            preInstructions.push(await createAssociatedTokenAccountIdempotentInstruction(mangoAccount.owner, mangoAccount.owner, inputBank.mint));
        }
        const outputTokenAccountPk = await getAssociatedTokenAddress(outputBank.mint, mangoAccount.owner);
        const outputTokenAccExists = await this.program.provider.connection.getAccountInfo(outputTokenAccountPk);
        if (!outputTokenAccExists) {
            preInstructions.push(await createAssociatedTokenAccountIdempotentInstruction(mangoAccount.owner, mangoAccount.owner, outputBank.mint));
        }
        const inputBankAccount = {
            pubkey: inputBank.publicKey,
            isWritable: true,
            isSigner: false,
        };
        const outputBankAccount = {
            pubkey: outputBank.publicKey,
            isWritable: true,
            isSigner: false,
        };
        const inputBankVault = {
            pubkey: inputBank.vault,
            isWritable: true,
            isSigner: false,
        };
        const outputBankVault = {
            pubkey: outputBank.vault,
            isWritable: true,
            isSigner: false,
        };
        const inputATA = {
            pubkey: inputTokenAccountPk,
            isWritable: true,
            isSigner: false,
        };
        const outputATA = {
            pubkey: outputTokenAccountPk,
            isWritable: false,
            isSigner: false,
        };
        const groupAM = {
            pubkey: group.publicKey,
            isWritable: false,
            isSigner: false,
        };
        const flashLoanEndIx = await this.program.methods
            .flashLoanEnd(flashLoanType)
            .accounts({
            account: mangoAccount.publicKey,
        })
            .remainingAccounts([
            ...parsedHealthAccounts,
            inputBankVault,
            outputBankVault,
            inputATA,
            {
                isWritable: true,
                pubkey: outputTokenAccountPk,
                isSigner: false,
            },
            groupAM,
        ])
            .instruction();
        const flashLoanBeginIx = await this.program.methods
            .flashLoanBegin([
            toNative(amountIn, inputBank.mintDecimals),
            new anchor.BN(0) /* we don't care about borrowing the target amount, this is just a dummy */,
        ])
            .accounts({
            account: mangoAccount.publicKey,
            owner: this.program.provider.wallet.publicKey,
            instructions: web3_js.SYSVAR_INSTRUCTIONS_PUBKEY,
        })
            .remainingAccounts([
            inputBankAccount,
            outputBankAccount,
            inputBankVault,
            outputBankVault,
            inputATA,
            outputATA,
            groupAM,
        ])
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [
            ...preInstructions,
            flashLoanBeginIx,
            ...userDefinedInstructions.filter((ix) => ix.keys.length > 2),
            flashLoanEndIx,
        ], { alts: [...group.addressLookupTablesList, ...userDefinedAlts] });
    }
    async updateIndexAndRate(group, mintPk) {
        const bank = group.getFirstBankByMint(mintPk);
        const mintInfo = group.mintInfosMapByMint.get(mintPk.toString());
        const ix = await this.program.methods
            .tokenUpdateIndexAndRate()
            .accounts({
            group: group.publicKey,
            mintInfo: mintInfo.publicKey,
            oracle: mintInfo.oracle,
            instructions: web3_js.SYSVAR_INSTRUCTIONS_PUBKEY,
        })
            .remainingAccounts([
            {
                pubkey: bank.publicKey,
                isWritable: true,
                isSigner: false,
            },
        ])
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    /// liquidations
    async liqTokenWithToken(group, liqor, liqee, assetMintPk, liabMintPk, maxLiabTransfer) {
        const assetBank = group.getFirstBankByMint(assetMintPk);
        const liabBank = group.getFirstBankByMint(liabMintPk);
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Scanning, group, [liqor, liqee], [assetBank, liabBank], []);
        const parsedHealthAccounts = healthRemainingAccounts.map((pk) => ({
            pubkey: pk,
            isWritable: pk.equals(assetBank.publicKey) || pk.equals(liabBank.publicKey)
                ? true
                : false,
            isSigner: false,
        }));
        const ix = await this.program.methods
            .liqTokenWithToken(assetBank.tokenIndex, liabBank.tokenIndex, {
            val: I80F48.fromNumber(maxLiabTransfer).getData(),
        })
            .accounts({
            group: group.publicKey,
            liqor: liqor.publicKey,
            liqee: liqee.publicKey,
            liqorOwner: liqor.owner,
        })
            .remainingAccounts(parsedHealthAccounts)
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async altSet(group, addressLookupTable, index) {
        const ix = await this.program.methods
            .altSet(index)
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            addressLookupTable,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async altExtend(group, addressLookupTable, index, pks) {
        const ix = await this.program.methods
            .altExtend(index, pks)
            .accounts({
            group: group.publicKey,
            admin: this.program.provider.wallet.publicKey,
            payer: this.program.provider.wallet.publicKey,
            addressLookupTable,
        })
            .instruction();
        return await this.sendAndConfirmTransactionForGroup(group, [ix]);
    }
    async healthRegionBeginIx(group, account, banks = [], perpMarkets = []) {
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [account], [...banks], [...perpMarkets]);
        const parsedHealthAccounts = healthRemainingAccounts.map((pk) => ({
            pubkey: pk,
            isWritable: false,
            isSigner: false,
        }));
        return await this.program.methods
            .healthRegionBegin()
            .accounts({
            group: group.publicKey,
            account: account.publicKey,
            instructions: web3_js.SYSVAR_INSTRUCTIONS_PUBKEY,
        })
            .remainingAccounts(parsedHealthAccounts)
            .instruction();
    }
    async healthRegionEndIx(group, account, banks = [], perpMarkets = []) {
        const healthRemainingAccounts = this.buildHealthRemainingAccounts(AccountRetriever.Fixed, group, [account], [...banks], [...perpMarkets]);
        const parsedHealthAccounts = healthRemainingAccounts.map((pk) => ({
            pubkey: pk,
            isWritable: false,
            isSigner: false,
        }));
        return await this.program.methods
            .healthRegionEnd()
            .accounts({ account: account.publicKey })
            .remainingAccounts(parsedHealthAccounts)
            .instruction();
    }
    /// static
    static connect(provider, cluster, programId, opts) {
        const idl = IDL;
        return new MangoClient(new anchor.Program(idl, programId, provider), programId, cluster, opts);
    }
    static connectForGroupName(provider, groupName) {
        const idl = IDL;
        const id = Id.fromIdsByName(groupName);
        return new MangoClient(new anchor.Program(idl, new web3_js.PublicKey(id.mangoProgramId), provider), new web3_js.PublicKey(id.mangoProgramId), id.cluster);
    }
    buildHealthRemainingAccounts(retriever, group, mangoAccounts, banks = [], perpMarkets = [], openOrdersForMarket = []) {
        if (retriever === AccountRetriever.Fixed) {
            return this.buildFixedAccountRetrieverHealthAccounts(group, mangoAccounts[0], banks, perpMarkets, openOrdersForMarket);
        }
        else {
            return this.buildScanningAccountRetrieverHealthAccounts(group, mangoAccounts, banks, perpMarkets);
        }
    }
    buildFixedAccountRetrieverHealthAccounts(group, mangoAccount, 
    // Banks and perpMarkets for whom positions don't exist on mango account,
    // but user would potentially open new positions.
    banks, perpMarkets, openOrdersForMarket) {
        const healthRemainingAccounts = [];
        const tokenPositionIndices = mangoAccount.tokens.map((t) => t.tokenIndex);
        for (const bank of banks) {
            const tokenPositionExists = tokenPositionIndices.indexOf(bank.tokenIndex) > -1;
            if (!tokenPositionExists) {
                const inactiveTokenPosition = tokenPositionIndices.findIndex((index) => index === TokenPosition.TokenIndexUnset);
                if (inactiveTokenPosition != -1) {
                    tokenPositionIndices[inactiveTokenPosition] = bank.tokenIndex;
                }
            }
        }
        const mintInfos = tokenPositionIndices
            .filter((tokenIndex) => tokenIndex !== TokenPosition.TokenIndexUnset)
            .map((tokenIndex) => group.mintInfosMapByTokenIndex.get(tokenIndex));
        healthRemainingAccounts.push(...mintInfos.map((mintInfo) => mintInfo.firstBank()));
        healthRemainingAccounts.push(...mintInfos.map((mintInfo) => mintInfo.oracle));
        // insert any extra perp markets in the free perp position slots
        const perpPositionIndices = mangoAccount.perps.map((p) => p.marketIndex);
        for (const perpMarket of perpMarkets) {
            const perpPositionExists = perpPositionIndices.indexOf(perpMarket.perpMarketIndex) > -1;
            if (!perpPositionExists) {
                const inactivePerpPosition = perpPositionIndices.findIndex((perpIdx) => perpIdx === PerpPosition.PerpMarketIndexUnset);
                if (inactivePerpPosition != -1) {
                    perpPositionIndices[inactivePerpPosition] =
                        perpMarket.perpMarketIndex;
                }
            }
        }
        const allPerpMarkets = perpPositionIndices
            .filter((perpIdx) => perpIdx !== PerpPosition.PerpMarketIndexUnset)
            .map((perpIdx) => group.getPerpMarketByMarketIndex(perpIdx));
        healthRemainingAccounts.push(...allPerpMarkets.map((perp) => perp.publicKey));
        healthRemainingAccounts.push(...allPerpMarkets.map((perp) => perp.oracle));
        // insert any extra open orders accounts in the cooresponding free serum market slot
        const serumPositionIndices = mangoAccount.serum3.map((s) => ({
            marketIndex: s.marketIndex,
            openOrders: s.openOrders,
        }));
        for (const [serum3Market, openOrderPk] of openOrdersForMarket) {
            const ooPositionExists = serumPositionIndices.findIndex((i) => i.marketIndex === serum3Market.marketIndex) > -1;
            if (!ooPositionExists) {
                const inactiveSerumPosition = serumPositionIndices.findIndex((serumPos) => serumPos.marketIndex === Serum3Orders.Serum3MarketIndexUnset);
                if (inactiveSerumPosition != -1) {
                    serumPositionIndices[inactiveSerumPosition].marketIndex =
                        serum3Market.marketIndex;
                    serumPositionIndices[inactiveSerumPosition].openOrders = openOrderPk;
                }
            }
        }
        healthRemainingAccounts.push(...serumPositionIndices
            .filter((serumPosition) => serumPosition.marketIndex !== Serum3Orders.Serum3MarketIndexUnset)
            .map((serumPosition) => serumPosition.openOrders));
        // debugHealthAccounts(group, mangoAccount, healthRemainingAccounts);
        return healthRemainingAccounts;
    }
    buildScanningAccountRetrieverHealthAccounts(group, mangoAccounts, banks, perpMarkets) {
        const healthRemainingAccounts = [];
        let tokenIndices = [];
        for (const mangoAccount of mangoAccounts) {
            tokenIndices.push(...mangoAccount.tokens
                .filter((token) => token.tokenIndex !== 65535)
                .map((token) => token.tokenIndex));
        }
        tokenIndices = [...new Set(tokenIndices)];
        if (banks?.length) {
            for (const bank of banks) {
                tokenIndices.push(bank.tokenIndex);
            }
        }
        const mintInfos = [...new Set(tokenIndices)].map((tokenIndex) => group.mintInfosMapByTokenIndex.get(tokenIndex));
        healthRemainingAccounts.push(...mintInfos.map((mintInfo) => mintInfo.firstBank()));
        healthRemainingAccounts.push(...mintInfos.map((mintInfo) => mintInfo.oracle));
        const perpIndices = [];
        for (const mangoAccount of mangoAccounts) {
            perpIndices.push(...mangoAccount.perps
                .filter((perp) => perp.marketIndex !== 65535)
                .map((perp) => perp.marketIndex));
        }
        perpIndices.push(...perpMarkets.map((perp) => perp.perpMarketIndex));
        const allPerpMarkets = [...new Set(perpIndices)].map((marketIndex) => group.findPerpMarket(marketIndex));
        // Add perp accounts
        healthRemainingAccounts.push(...allPerpMarkets.map((p) => p.publicKey));
        // Add oracle for each perp
        healthRemainingAccounts.push(...allPerpMarkets.map((p) => p.oracle));
        for (const mangoAccount of mangoAccounts) {
            healthRemainingAccounts.push(...mangoAccount.serum3
                .filter((serum3Account) => serum3Account.marketIndex !== 65535)
                .map((serum3Account) => serum3Account.openOrders));
        }
        return healthRemainingAccounts;
    }
    async modifyPerpOrder(group, mangoAccount, perpMarketIndex, orderId, side, price, quantity, maxQuoteQuantity, clientOrderId, orderType, reduceOnly, expiryTimestamp, limit) {
        const transactionInstructions = [];
        const [cancelOrderIx, placeOrderIx] = await Promise.all([
            this.perpCancelOrderIx(group, mangoAccount, perpMarketIndex, orderId),
            this.perpPlaceOrderIx(group, mangoAccount, perpMarketIndex, side, price, quantity, maxQuoteQuantity, clientOrderId, orderType, reduceOnly, expiryTimestamp, limit),
        ]);
        transactionInstructions.push(cancelOrderIx, placeOrderIx);
        return await this.sendAndConfirmTransactionForGroup(group, transactionInstructions);
    }
    async modifySerum3Order(group, orderId, mangoAccount, externalMarketPk, side, price, size, selfTradeBehavior, orderType, clientOrderId, limit) {
        const transactionInstructions = [];
        const [cancelOrderIx, settleIx, placeOrderIx] = await Promise.all([
            this.serum3CancelOrderIx(group, mangoAccount, externalMarketPk, side, orderId),
            this.serum3SettleFundsIx(group, mangoAccount, externalMarketPk),
            this.serum3PlaceOrderIx(group, mangoAccount, externalMarketPk, side, price, size, selfTradeBehavior, orderType, clientOrderId, limit),
        ]);
        transactionInstructions.push(cancelOrderIx, settleIx, ...placeOrderIx);
        return await this.sendAndConfirmTransactionForGroup(group, transactionInstructions);
    }
}

exports.Bank = Bank;
exports.BookSide = BookSide;
exports.BookSideType = BookSideType;
exports.Group = Group;
exports.HUNDRED_I80F48 = HUNDRED_I80F48;
exports.HealthType = HealthType;
exports.I64_MAX_BN = I64_MAX_BN;
exports.I80F48 = I80F48;
exports.I80F48Dto = I80F48Dto;
exports.InnerNode = InnerNode;
exports.LeafNode = LeafNode;
exports.MANGO_V4_ID = MANGO_V4_ID;
exports.MAX_I80F48 = MAX_I80F48;
exports.MangoAccount = MangoAccount;
exports.MangoClient = MangoClient;
exports.MintInfo = MintInfo;
exports.NullPerpEditParams = NullPerpEditParams;
exports.NullTokenEditParams = NullTokenEditParams;
exports.ONE_I80F48 = ONE_I80F48;
exports.OPENBOOK_PROGRAM_ID = OPENBOOK_PROGRAM_ID;
exports.PerpEventQueue = PerpEventQueue;
exports.PerpMarket = PerpMarket;
exports.PerpOo = PerpOo;
exports.PerpOoDto = PerpOoDto;
exports.PerpOrder = PerpOrder;
exports.PerpOrderSide = PerpOrderSide;
exports.PerpOrderType = PerpOrderType;
exports.PerpPosition = PerpPosition;
exports.PerpPositionDto = PerpPositionDto;
exports.QUOTE_DECIMALS = QUOTE_DECIMALS;
exports.RUST_I64_MAX = RUST_I64_MAX;
exports.RUST_I64_MIN = RUST_I64_MIN;
exports.RUST_U64_MAX = RUST_U64_MAX;
exports.Serum3Market = Serum3Market;
exports.Serum3OrderType = Serum3OrderType;
exports.Serum3Orders = Serum3Orders;
exports.Serum3PositionDto = Serum3PositionDto;
exports.Serum3SelfTradeBehavior = Serum3SelfTradeBehavior;
exports.Serum3Side = Serum3Side;
exports.StubOracle = StubOracle;
exports.TokenPosition = TokenPosition;
exports.TokenPositionDto = TokenPositionDto;
exports.TrueIxGateParams = TrueIxGateParams;
exports.U64_MAX_BN = U64_MAX_BN;
exports.ZERO_I80F48 = ZERO_I80F48;
exports.buildIxGate = buildIxGate;
exports.buildVersionedTx = buildVersionedTx;
exports.createAssociatedTokenAccountIdempotentInstruction = createAssociatedTokenAccountIdempotentInstruction;
exports.getAssociatedTokenAddress = getAssociatedTokenAddress;
exports.toNative = toNative;
exports.toNativeI80F48 = toNativeI80F48;
exports.toNativeI80F48ForQuote = toNativeI80F48ForQuote;
exports.toUiDecimals = toUiDecimals;
exports.toUiDecimalsForQuote = toUiDecimalsForQuote;
exports.toUiI80F48 = toUiI80F48;
